#!/usr/bin/perl
use warnings;
use strict;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

use Cwd;
use Math::Trig;
use get_info;

# setup_rmc.pl is used to setup the random model calculation (RMC). It
# does not actually launch any of the parallel image processing codes,
# but rather generates the directory structure, auto3dem parameter
# files, random particle orientations, and command script required to
# carry out the RMC

# Input:
#   $rundir - directory containing particle image and parameter files
#   $list - name of file containing list of particle parameter files
#   $nmodels - number of random models (default = 10)

# ----- Directory structure for RMC -----
#
# The user specifies the directory containing the particle image and
# parameter files. Within this directory, subdirectories are then
# created for each random model: RMC1, RMC2, etc. Within the
# subdirectories, symbolic links are set to the image files and new
# parameter files are written which contain the random particle
# orientations.

# ----- Random orientation generation -----
#
# First generate list of (theta,phi) values evenly covering 1/2 of the
# asymmetric unit (ASU), typically done using spacing of 0.5
# degrees. Then, for each particle, randomly choose
#
# (1) (theta,|phi|) from set of orientations
# (2) Sign of phi
# (3) Omega in the range 0 to 360
# 
# The new particle parameter files contain the same header information
# as the original files, but with random orientations and particle
# centers initialized to lie at the center of the box.

# ----- auto3dem parameter files -----
#
# For each random model, an auto3dem parameter file is
# created. Default values are heavily relied upon and only a minimal
# set of parameters are explicitly set. These include the name of the
# final map, the run directory, resolutions, and annulus sizes. The
# initial choices for a number of these parameters depend explicitly
# on the box and pixel sizes of the images.

# ----- Command file -----
#
# The RMC can be broken down into a set of independent auto3dem runs
# each using a different set of random particle orientations. This
# script generates a command file that launches the individual
# auto3dem runs and parses the results to determine which random model
# results in the best resolution.

# Get input from command line
my ($rundir, $piflist, $nmodels, $boxrad_cmdline) = @ARGV;

# Print usage info if no arguments provided
if(@ARGV == 0) {
    print "setup_rmc - script for setting up random model calculation\n\n";
    print "Usage: setup_rmc dir list nmodels [boxrad]\n\n";
    print "dir:         directory containing data\n";
    print "list:        file containing list of parameter files in dir\n";
    print "nmodels:     number of random models to generate\n";
    print "boxrad:      box radius (pixels). Only needed if info missing";
    print " from pif header\n";
    print "\n";
    exit(0);
}

unless(defined $nmodels) {
    $nmodels = 10;
}

# Initialize random number generator to get a reproducible sequence
srand(1);

# Get list of all view in the ASU
my $dang = 0.5;
my @theta = ();
my @phi = ();
get_all_views($dang, \@theta, \@phi);
my $nviews = @theta;

my %ptle;
my %head;
my %numimages;
my %boxrad;
my %pixel_size;
my %defocus;
my ($defocus_maj, $defocus_min);

# For each particle parameter file, get the name of the corresponding
# PIF image file, header information, number of images, box radius,
# and pixel size.

open(PIFLIST, "$piflist");
while(<PIFLIST>) {
    s/^\s+//;
    s/\s+$//;
    my $ptle = $_;
    if(defined $ptle && $ptle =~ /\w/) {

	open(PTLE, "$rundir/$ptle");
	my $pif = <PTLE>;
	$pif =~ s/\s+//g;
	$head{$pif} = <PTLE>;
	chomp($head{$pif});
	(undef, undef, undef, undef, $defocus_maj, $defocus_min, 
	 undef, undef) = split(/,/, $head{$pif});
	$defocus{$pif} = ($defocus_maj + $defocus_min)/2.0;
	close(PTLE);

	$numimages{$pif} = get_numimages("$rundir/$pif");
	$boxrad{$pif} = get_boxrad("$rundir/$pif");
	$pixel_size{$pif} = get_pixel_size("$rundir/$ptle");
	$ptle{$pif} = $ptle;
    }
}
close(PIFLIST);

# Use a total of $limit (default = 150) particles to construct the
# random model. Sort the micrographs by defocus and select particles
# starting with the highest defocus micrographs until the limit is
# reached. The array @pifs will contain the list of micrographs to be
# used and the corresponding values of the hash %numimages are reset
# to the required number of images from each micrograph.

my @pifs = ();
my $running_total = 0;
my $limit = 150;
foreach my $pif (sort { $defocus{$b} <=> $defocus{$a} } keys %defocus) {
    push(@pifs, $pif);
    if($running_total + $numimages{$pif} >= $limit) {
	$numimages{$pif} = $limit - $running_total;
	last;
    }
    $running_total += $numimages{$pif};
}


# Use first file to get boxrad and pixelsize
my $boxrad_from_pif = $boxrad{$pifs[0]};
my $pixel_size = $pixel_size{$pifs[0]};
my $boxrad;

if($boxrad_from_pif == 0) {
    undef $boxrad_from_pif;
}

# The following code is used to allow for the specification of the
# box radius in either the PIF file or the auto3dem parameter
# file. Necessary because some PIF files are missing the box
# radius

# Case 1 - box radius not set in either location
unless(defined $boxrad_cmdline || defined $boxrad_from_pif) {
    print "Box radius not specified in PIF file or command line\n";
    print "Either fix $pifs[0] or manually specify box radius\n";
    die "Terminating execution\n";
}

# Case 2 - box radius specified in both locations
if(defined $boxrad_cmdline && defined $boxrad_from_pif) {
    $boxrad = $boxrad_cmdline;
    print "Box radius specified in both PIF file and command line\n";
    if($boxrad_cmdline == $boxrad_from_pif) {
	print "Box radius ($boxrad) specified consistently\n";
    } else {
	print "\nWarning boxrad from PIF file ($boxrad_from_pif) ";
	print "not equal to boxrad from command line ";
	print "($boxrad_cmdline)\n";
	print "Value in parameter file overides PIF file\n";
    }
}

# Case 3 - box radius specified only in PIF file
if(defined $boxrad_from_pif && !defined $boxrad_cmdline) {
    $boxrad = $boxrad_from_pif;
    print "Box radius from PIF file = $boxrad\n";
}

# Case 4 - box radius specified only in command line
if(!defined $boxrad_from_pif && defined $boxrad_cmdline) {
    $boxrad = $boxrad_cmdline;
    print "Box radius from command_line = $boxrad\n";
}


# Create directory structure, set symbolic links, generate files
# containing random particle orientations, write auto3dem parameter
# files, and create command file for running RMC

my $listing_file = "listing_file";
open(LIST, ">$listing_file");

my $command_file = "run_command";
open(CMD, ">$command_file");
chmod 0755, "$command_file";

my $startdir = getcwd();
chdir($rundir);
my $olddir = getcwd();

for (my $i=1; $i<=$nmodels; $i++) {

    print LIST "RMC${i}_log\n";
    print CMD  "auto3dem.pl NCPU NODEFILE RMC${i}_master\n";

    my $rmcdir = "RMC${i}";
    mkdir $rmcdir;

    open(AUTO, ">${startdir}/RMC${i}_master");

    print AUTO "# Parameter file for random model $i\n\n";

    print AUTO "auto rundir          ${rundir}/RMC${i}\n";
    print AUTO "auto mode            search\n";
    print AUTO "auto switch_mode     0 # Don't allow switch to refine mode\n";
    print AUTO "auto niter           10\n";
    print AUTO "auto have_map        0\n";
    print AUTO "auto outfile         RMC${i}\n";
    print AUTO "auto freeze_annulus  1\n";
    print AUTO "auto freeze_res      1 # Use only for RMC\n";
    print AUTO "auto boxrad          $boxrad\n";
    print AUTO "\n";

    my $in_rad  = sprintf "%0.0f", $boxrad/3.0;
    my $out_rad = sprintf "%0.0f", $boxrad - 10.0;
    print AUTO "pcut in_rad          $in_rad\n";
    print AUTO "pcut out_rad         $out_rad\n";
    print AUTO "\n";

    my $res_min = 29.5;
    my $res_max = 29.0;
    print AUTO "p3dr res_min         $res_min\n";
    print AUTO "p3dr res_max         $res_max\n";
    print AUTO "\n";

    $res_min = 60.0;
    $res_max = 30.0;
    print AUTO "psf  res_min         $res_min\n";
    print AUTO "psf  res_max         $res_max\n";
    print AUTO "\n";

    my $annulus_low  = sprintf "%0.0f", $boxrad/3.0;
    my $annulus_high = sprintf "%0.0f", $boxrad - 10.0;
    my $resolution_low = $boxrad*2.0*$pixel_size/5.0;
    my $resolution_high = 30.0;
    my $pftrad_hi = $boxrad;
    print AUTO "ppft annulus_low     $annulus_low\n";
    print AUTO "ppft annulus_high    $annulus_high\n";
    print AUTO "ppft delta_theta     1 # Use only for RMC\n";
    print AUTO "ppft resolution_low  $resolution_low\n";
    print AUTO "ppft resolution_high $resolution_high\n";
    print AUTO "ppft pftrad_hi       $pftrad_hi\n";
    print AUTO "ppft verbose         -1 # Use only for RMC\n";

    if($pixel_size <= 5.0) {
	print AUTO "ppft bin_factor      2 # Use only for RMC\n";
    }
    print AUTO "\n";

    chdir($rmcdir);
    foreach my $pif (@pifs) {
	print AUTO "data $ptle{$pif}\n";
	symlink "../$pif", "$pif";
	make_random_file($pif, $ptle{$pif}, $head{$pif}, 
			 $numimages{$pif}, $boxrad);
    }
    chdir($olddir);

    close(AUTO);
}

print CMD "findbest.pl $listing_file\n";
close(CMD);

close(LIST);



# --------------------------------------------------------------

sub make_random_file {

    # Purpose: Create a parameter file corresponding to a PIF
    # file. Particle orientations are chosen randomly within the
    # asymmetric unit and coordinates of origin are set equal to the
    # box radius.
    #
    # Input arguments:
    #   $pif - name of PIF file
    #   $ptle - name of particle parameter file
    #   $head - header line (voltage, defocus, etc.)
    #   $numimages - number of images in PIF file
    #   $boxrad - box radius

    my ($pif, $ptle, $head, $numimages, $boxrad) = @_;

    open(PTLE, ">$ptle");
    print PTLE "$pif\n";
    print PTLE "$head\n";

    for (my $i=1; $i<=$numimages; $i++) {

	# Choose random theta, |phi| in ASU
	my $j = int(rand($nviews));
	my $theta = $theta[$j];
	my $phi = $phi[$j];
    
	# Randomly assign sign to phi
	if(int(rand(2)) == 1) {
	    $phi = -$phi;
	}

	# Random select omega in range 0 to 360
	my $omega = rand(360.0);
	$omega = sprintf "%0.2f", $omega;

	printf PTLE "%5d, %7.2f, %7.2f, %7.2f, " ,
	$i, $theta, $phi, $omega;

	printf PTLE "%6.2f, %6.2f, ",
	$boxrad, $boxrad;

	printf PTLE "%5.3f, %5.3f, %5.3f, %5.3f\n",
	1.0, 0.0, 0.0, 0.0;
    }
    close(PTLE);
}

sub get_numimages {
    
    # Purpose: get the number of images contained in a PIF file
    #
    # Input arguments:
    #   $filename - name of PIF file
    #
    # Return values:
    #   $numimages - number of images in PIF file

    my ($filename) = @_;

    open(FH, $filename) || die 
	"Unable to open file $filename in subroutine get_numimages\n";
    binmode FH;

    # Determine the endianness of the machine that created this file
    # based on the 4 bytes offset 28 bytes from the start of the
    # file. According to the PIF standard, a value of zero corresponds
    # to little endian, while a value of 1 corresponds to big
    # endian. Since we don't know the endianness of the file ahead of
    # time, the 4 bytes will arbitrarily be interpreted as a
    # big-endian long int and we'll test whether or not the result is
    # non-zero

    my ($bigend, $numimages);
    seek(FH, 28, 0);
    read FH, $bigend, 4;
    $bigend = unpack "N", $bigend; # # Big endian interpretation

    # numimages variable is offset 24 bytes from start of file
    seek(FH, 24, 0);
    read FH, $numimages, 4;
    
    # Interpret as big- or little-endian based on endianness of file
    if($bigend) {
	$numimages = unpack "N", $numimages;
    } else {
	$numimages = unpack "V", $numimages;
    }

    close(FH);
    return($numimages);
}

sub get_all_views {

    # Purpose: Generate a set of orientations (theta, phi) that evenly
    # cover the icosahedral asymmetric unit. For each value of theta,
    # values of phi are chosen that span the interval
    # 0<=phi<=phimax. In order to evenly sample points on a sphere, a
    # spacing $dang/sin(theta) is used. The maximum value of phi is
    # calculated by a call to subroutine get_phimax()
    #
    # Input parameters: 
    #   $dang - angular spacing for theta
    #   $theta_ref - reference to array of theta's (changed on output) 
    #   $phi_ref - reference to array of phi's (changed on output)
    #
    # Return values:
    #   None

    my ($dang, $theta_ref, $phi_ref) = @_;

    # The first orientation is always chosen to be (69.09, 0)
    my $theta = 69.09;
    my $phi = 0.0;

    push(@{$theta_ref}, $theta);
    push(@{$phi_ref}, $phi);

    # Subsequent orientations start at 69.0 + $dang
    $theta = 69.0 + $dang;

    while ($theta <= 90.0) {
	my $dphi = $dang/sin($theta/180.0*acos(-1.0));
	my $phimax = get_phimax($theta);
	$phi = 0.0;
	while ($phi <= $phimax) {
	    my $theta_rnd = sprintf "%0.2f", $theta;
	    my $phi_rnd   = sprintf "%0.2f", $phi;
	    push(@{$theta_ref}, $theta_rnd);
	    push(@{$phi_ref}, $phi_rnd);
	    $phi = $phi + $dphi;	
	}		
	$theta = $theta + $dang;
    } 
}

sub get_phimax {

    # Purpose: Calculate the maximum allowable value of the angle phi
    # in the icosahedral asymmetic unit for a given value of
    # theta. Calculation is based on spherical triognometry relations
    # described in CRC Standard Mathematical Tables
    # Input parameter:
    #   $theta - inclination angle
    #
    # Return value:
    #   $phimax - maximum allowable azimuthal angle

    my ($theta) = @_;

    my $small_b = 69.09484255211070;
    my $c       = 31.71747441146101;
    my $deg_to_rad = acos(-1.0)/180.0;
    my $rad_to_deg = 1.0/$deg_to_rad;
    my $ninety = 90.0;

    my $temp = (1/tan($c*$deg_to_rad/2))  * 
	(sin(($ninety-$small_b)*$deg_to_rad/2)/ 
	 sin(($ninety+$small_b)*$deg_to_rad/2)) ;
    my $a_minus_b = 2*$rad_to_deg*atan($temp);
    $temp = (1/tan($c*$deg_to_rad/2))  * 
	(cos(($ninety-$small_b)*$deg_to_rad/2)/ 
	 cos(($ninety+$small_b)*$deg_to_rad/2)) ;
    my $a_plus_b = 2*$rad_to_deg*atan($temp) ;
    my $b = ($a_plus_b - $a_minus_b) / 2;
    my $a = $a_minus_b + $b;
    
    $temp    = sin($a*$deg_to_rad)*sin($small_b*$deg_to_rad) / 
	sin($theta*$deg_to_rad) ;
    $b       = $rad_to_deg*asin($temp);
    $temp    = sin(($theta-$small_b)*$deg_to_rad/2)  / 
	( sin(($theta+$small_b)*$deg_to_rad/2) *  
          tan(($a-$b)*$deg_to_rad/2) );
    my $phimax  = 2*$rad_to_deg*atan($temp) ;

    return($phimax);
}
