/* File: libPIF.c  
 * I/O routines  

 * Robert W. Ashmore

 * Mods by James Conway, 23&27 May, 1997, and March 2000.
*/

/*************************************************************
 * A NOTE ABOUT BINNING:
 *  
 * All calls to routines herein assume that the ncol/nrow/nsec
 * values passed to the convenience routines that are called
 * from application programs like robEM, em3dr, pft etc. pass
 * the ORIGINAL value of those dimensions.  The binFactor
 * parameter, then, tells the routines how the data is stored
 * on disk.
 * 
 * For example, if I start out with a 149x149x149 MAP file
 * BUT only put the bin==2 on the disk I would pass these
 * values to the writing routine:
 * 
 *    ncol = 149
 *    nrow = 149
 *    nsec = 149
 *    binFactor = 2
 * 
 * The wriitng routine then would write 75x75x75 data points to 
 * the disk.
 * 
 * Conversely, if I was reading and I got these values as
 * input to the reading routine:
 * 
 *    ncol = 149
 *    nrow = 149
 *    nsec = 149
 *    binFactor = 2
 * 
 * I would only read 75x75x75 data points.  NOTE that if binFactor
 * equals '1' (one) in either of the above examples the original
 * 149x149x149 would be written/read.
 * 
 * All allocation of memory for holding the data points is the
 * responsibility of the calling routine whether reading OR
 * writing.
 * 
 * Rob Ashmore   August 1999
 * James Conway  March  2000 - add big-endian capabilities
 *************************************************************/

#define VERSION "2.10_20030619 libPIF"
#define EM_VERSION "v2.16"

/************
 * INCLUDES *
 ************/
#include "../include/libPIF.h"
#include <sys/types.h>
#include <string.h>

#define MAXLUNS 1000
static FILE *lun[MAXLUNS];
static int   endians[MAXLUNS];


/* ************************* *
 *        GLOBALS            *
 * ************************* */
static fString_t fileName;

static int convTsb= 0;
static int i2Map= TRUE;


/* ******** *
 *  MACROS  *
 * ******** */
#define OKREAD(x,y,z) if(x!=y){printf("[%s] [PIF] Error accessing file.\n",z);\
exit(0);}

void pif_write_map_section4(int *luNum,
                            float *float_buffer,
                            int *secNum,
                            int *numCols,
                            int *numRows,
                            float *floatScaleFactor)
{/* PIF_WRITE_MAP_SECTION4 */
  pif_write_map_section4_(luNum,
                         float_buffer,
                         secNum,
                         numCols,
                         numRows,
                         floatScaleFactor);
}/* PIF_WRITE_MAP_SECTION4 */


/******************************************************************
 * banner_today() - Return todays date and current time in STRING 
 * (static) form.
*/
char *banner_today()
{/* BANNER_TODAY */
  static char s1[132];
  char *ampm;
  int seconds,
      minutes,
      hours,
      hyear,
      hmonth,
      hday;
#ifdef VMS
  tm_t *tim;
#else
  struct tm *tim;
#endif
  time_t tSec;

  tSec= time(NULL); 
  tim= localtime((time_t *)&tSec); 

  seconds= tim->tm_sec;
  minutes= tim->tm_min;
  hours=   tim->tm_hour;
  hday=    tim->tm_mday;
  hmonth=  tim->tm_mon + 1;
  hyear=   tim->tm_year + 1900;

  if(hours > 12)
    {/* PM */
      hours -= 12;
      ampm= "PM";
    }/* PM */
  else if(hours==12)
    ampm= "PM";
  else ampm= "AM";

  sprintf(s1,"%02d/%02d/%04d,  %02d:%02d:%02d%s",
         hmonth, hday, hyear, hours, minutes, seconds, ampm);

  return(s1);
}/* BANNER_TODAY */




char * banner(char *application)
{
  return(banner_(application));
}
/****************************************************************
 * banner() - Print the current version number and today's date
 * to give the executable a time/version stamp.
*/
char * banner_(char *application)
{/* BANNER */
  static char bannerOut[1024];
  pid_t pid;

  pid = getpid();
  sprintf(bannerOut,"%s-%s",application,EM_VERSION);
  /*  OLDWAY   printf("[%s %s]\n",bannerOut,banner_today()); */
  printf("PID[%d][%s %s %s]\n",pid,bannerOut,__DATE__,__TIME__);
  return(bannerOut);

}/* BANNER */



void pif_tsbworld(int setValue)
{
  pif_tsbworld_(setValue);
}
/****************************************************************
 * pif_tsbworld() - Setup for whether we read datat from disk
 * in an XWindows type format where the first row on the disk is 
 * actually the last row in the data unit (convTsb = 1) OR
 * the first row on the disk is the first row in the data unit 
 * (convTsb = 0).
*/
void pif_tsbworld_(int setValue)
{/* PIF_TSBWORLD */
  convTsb = setValue;
}/* PIF_TSBWORLD */



void calcmapstats(float *inbuffer, float *img_bkg, float *map_min,
		  float *map_max, float *img_avg, float *img_std,
		  int *width, int *height, int *nSecs)
{/* calcMapStats */
  calcMapStats_(inbuffer, img_bkg, map_min,
		map_max, img_avg, img_std,
		width, height, nSecs);
}/* calcMapStats */
void calcmapstats_(float *inbuffer, float *img_bkg, float *map_min,
		  float *map_max, float *img_avg, float *img_std,
		  int *width, int *height, int *nSecs)
{/* calcMapStats */
  calcMapStats_(inbuffer, img_bkg, map_min,
		map_max, img_avg, img_std,
		width, height, nSecs);
}/* calcMapStats */
void calcMapStats(float *inbuffer, float *img_bkg, float *map_min,
		  float *map_max, float *img_avg, float *img_std,
		  int *width, int *height, int *nSecs)
{/* calcMapStats */
  calcMapStats_(inbuffer, img_bkg, map_min,
		map_max, img_avg, img_std,
		width, height, nSecs);
}/* calcMapStats */
/********************************************************
 * calcMapStats() - Calculates statistics from the MAP.
*/
void calcMapStats_(float *inbuffer, float *img_bkg, float *map_min,
	  float *map_max, float *img_avg, float *img_std,
	  int *width, int *height, int *nSecs)
{/* calcMapStats */
  int i,j,k,wd,hi,ns;
  int tot;
  float totValues,pixelValue;
  double var,sumsq;

  wd = *width;
  hi = *height;
  ns = *nSecs;
  /* Init all */
  *img_bkg = 0.0;
  *map_min = *map_max = inbuffer[0];
  *img_avg = 0.0;
  *img_std = 0.0;

  /* Accumulate intermediate values */
  totValues = tot = 0;
  var = sumsq = 0.0;
  for(k=0;k<ns;k++)
    {/* Do all secs */
      for(i=0;i<hi;i++)
	{/* height */
	  for(j=0;j<wd;j++)
	    {/* Look at all pixels */
	      pixelValue = inbuffer[((k*hi*wd)+(i*wd)+j)];
	      /* Use non-zeros also */
	      *map_min = MIN(*map_min,pixelValue);
	      *map_max = MAX(*map_max,pixelValue);
	      sumsq += (pixelValue * pixelValue);
	      totValues +=  (float)pixelValue;
	      tot++;
	    }/* Look at all pixels */
	}/* height */
    }/* Do all secs */
  
  /* Calc the background */
  *img_bkg= (float)inbuffer[2];   /* Bkg in the first section */

  /* Calc stats */
  *img_avg = (float)((float)totValues / (float)(wd * hi * ns));
  *img_std = 0.0;
  var = (sumsq / (float)(wd*hi*ns)) - (*img_avg * *img_avg);
  if(var > 0.0)
    *img_std = (float)sqrt(var);

}/* calcMapStats */




void pif_set_i2map(int *value)
{/* PIF_SET_I2MAP */
  pif_set_i2map_(value);
}/* PIF_SET_I2MAP */
/*************************************************************
 * pif_set_i2map() - Setup fo I2 writing from fortran.
*/
void pif_set_i2map_(int *value)
{/* PIF_SET_I2MAP */
  i2Map = *value;
}/* PIF_SET_I2MAP */




void determine_scaling(forPifHead_t *thisHead, float *inbuffer,
			int *ncol, int *nrow, int *nsec)
{/* DETERMINE_SCALING */
  determine_scaling_(thisHead, inbuffer,
		     ncol, nrow, nsec);
}/* DETERMINE_SCALING */
/*************************************************************
 * determine_scaling() - Calculate the factor needed to convert
 * out floatIntegers into floats when we read this file later.
 * Just scan all of our values and pick a value that maintains 
 * our dynamic range.
*/
void determine_scaling_(forPifHead_t *thisHead, float *inbuffer,
			int *ncol, int *nrow, int *nsec)
{/* DETERMINE_SCALING */
  int minVal,maxVal;
  float img_bkg;

  /* calc stats first */
  calcMapStats(inbuffer, 
	       &img_bkg, 
	       &(thisHead->datar4_min),
	       &(thisHead->datar4_max),
	       &(thisHead->datar4_avg),
	       &(thisHead->datar4_std),
	       ncol,nrow,nsec);
  thisHead->datai4_min = (int)thisHead->datar4_min;
  thisHead->datai4_max = (int)thisHead->datar4_max;

  minVal = 99999;
  maxVal = -99999;
  /* Don't forget to account for alp,bet,gam etc... */
  minVal= MIN(minVal,(int)(thisHead->datai4_min));
  minVal= MIN(minVal,(int)(thisHead->datai4_max));
  minVal= MIN(minVal,(int)(thisHead->datar4_min));
  minVal= MIN(minVal,(int)(thisHead->datar4_max));
  minVal= MIN(minVal,(int)(thisHead->datar4_avg));
  minVal= MIN(minVal,(int)(thisHead->datar4_std));
  minVal= MIN(minVal,(int)(thisHead->aoverb));
  minVal= MIN(minVal,(int)(thisHead->map_abang));
  minVal= MIN(minVal,(int)(thisHead->a_cell));
  minVal= MIN(minVal,(int)(thisHead->b_cell));
  minVal= MIN(minVal,(int)(thisHead->c_cell));
  minVal= MIN(minVal,(int)(thisHead->alp));
  minVal= MIN(minVal,(int)(thisHead->bet));
  minVal= MIN(minVal,(int)(thisHead->gam));
  minVal= MIN(minVal,(int)(thisHead->dela));
  minVal= MIN(minVal,(int)(thisHead->delb));
  minVal= MIN(minVal,(int)(thisHead->delc));
  minVal= MIN(minVal,(int)(thisHead->a_star));
  minVal= MIN(minVal,(int)(thisHead->b_star));
  minVal= MIN(minVal,(int)(thisHead->c_star));
  minVal= MIN(minVal,(int)(thisHead->alp_star));
  minVal= MIN(minVal,(int)(thisHead->bet_star));
  minVal= MIN(minVal,(int)(thisHead->gam_star));
  minVal= MIN(minVal,(int)(thisHead->pixelSize));
  minVal= MIN(minVal,(int)(thisHead->dthe));
  minVal= MIN(minVal,(int)(thisHead->dphi_90));
  minVal= MIN(minVal,(int)(thisHead->tempFactor));
  minVal= MIN(minVal,(int)(thisHead->wiener));

  maxVal= MAX(maxVal,(int)(thisHead->datai4_min));
  maxVal= MAX(maxVal,(int)(thisHead->datai4_max));
  maxVal= MAX(maxVal,(int)(thisHead->datar4_min));
  maxVal= MAX(maxVal,(int)(thisHead->datar4_max));
  maxVal= MAX(maxVal,(int)(thisHead->datar4_avg));
  maxVal= MAX(maxVal,(int)(thisHead->datar4_std));
  maxVal= MAX(maxVal,(int)(thisHead->aoverb));
  maxVal= MAX(maxVal,(int)(thisHead->map_abang));
  maxVal= MAX(maxVal,(int)(thisHead->a_cell));
  maxVal= MAX(maxVal,(int)(thisHead->b_cell));
  maxVal= MAX(maxVal,(int)(thisHead->c_cell));
  maxVal= MAX(maxVal,(int)(thisHead->alp));
  maxVal= MAX(maxVal,(int)(thisHead->bet));
  maxVal= MAX(maxVal,(int)(thisHead->gam));
  maxVal= MAX(maxVal,(int)(thisHead->dela));
  maxVal= MAX(maxVal,(int)(thisHead->delb));
  maxVal= MAX(maxVal,(int)(thisHead->delc));
  maxVal= MAX(maxVal,(int)(thisHead->a_star));
  maxVal= MAX(maxVal,(int)(thisHead->b_star));
  maxVal= MAX(maxVal,(int)(thisHead->c_star));
  maxVal= MAX(maxVal,(int)(thisHead->alp_star));
  maxVal= MAX(maxVal,(int)(thisHead->bet_star));
  maxVal= MAX(maxVal,(int)(thisHead->gam_star));
  maxVal= MAX(maxVal,(int)(thisHead->pixelSize));
  maxVal= MAX(maxVal,(int)(thisHead->dthe));
  maxVal= MAX(maxVal,(int)(thisHead->dphi_90));
  maxVal= MAX(maxVal,(int)(thisHead->tempFactor));
  maxVal= MAX(maxVal,(int)(thisHead->wiener));

  /* Get greatest magnitude */
  if(minVal < 0)
    minVal = 0 - minVal;
  maxVal = MAX(minVal,maxVal);
  /* Now let's give it a little extra so we're guaranteed NOT
   * to run over the top */
  maxVal = (int)((float)maxVal*1.05);


  /* We're doing MAPs as shorts so make it smaller */
  if(i2Map)
    (thisHead->floatScaleFactor) = (double)((double) MAXSHORT / 
					     (double) maxVal);
  else
    (thisHead->floatScaleFactor) = (double)((double) MAXINT / 
					     (double) maxVal);

  (thisHead->int4scaleF) = (double)((double) MAXINT / 
					     (double) maxVal);

  (thisHead->recipFloatScaleFactor) = (double)((double)1.0 / 
				(double)(thisHead->floatScaleFactor)); 

}/* DETERMINE_SCALING */




/***************************************************************
 * In UNIX the fortran compiler adds an underbar for every
 * routine here called from FORTRAN.  So let's just have a dummy 
 * routine that calls the appropriate routine.
 ***************************************************************/

void makefv_fname(char *inFname, char *out1, char *out2, char *out3)
{/* makefv_FNAME */
  makefv_fname_(inFname, out1, out2, out3);
}/* makefv_FNAME */

void increment_fname(fString_t *inFname, fString_t *outFname)
{/* INCREMENT_FNAME */
  increment_fname_(inFname, outFname);
}/* INCREMENT_FNAME */

void pif_close(int *luNum)
{/* PIF_CLOSE */
  pif_close_(luNum);
}/* PIF_CLOSE */

void pif_open(char *fName, int *luNum, char *readWrite)
{/* PIF_OPEN */
  pif_open_(fName, luNum, readWrite);
}/* PIF_OPEN */

void pif_open_vms(fString_t *fName, int *luNum, char *readWrite)
{/* PIF_OPEN_VMS */
  pif_open_vms_(fName, luNum, readWrite);
}/* PIF_OPEN_VMS */

int pif_read_gh(forPifHead_t *thisHead)
{/* PIF_READ_GH */
  return(pif_read_gh_(thisHead));
}/* PIF_READ_GH */

int pif_read_dh(forPifHead_t *thisHead)
{/* PIF_READ_DH */
  return(pif_read_dh_(thisHead));
}/* PIF_READ_DH */

int pif_read_header(int *luNum,
		    int *nimg, 
		    float *recipFloatScaleFactor,
		    float *int4scaleF,
		    int *htype, int *numCols, int *numRows, int *numSecs,
		    int *mode, int *endianNess, char *versionStr, 
		    float *pifVersion, 
		    int *even, int *mrcX, int *mrcY, int *mrcZ)
{/* PIF_READ_HEADER */
  return(pif_read_header_(luNum,
			  nimg, 
			  recipFloatScaleFactor,
			  int4scaleF,
			  htype, numCols, numRows, numSecs,
			  mode, endianNess, versionStr, pifVersion, 
			  even, mrcX, mrcY, mrcZ
			  ));
}/* PIF_READ_HEADER */

int pif_read_image_header(int *luNum,
			   int *imgNum,
			   float *recipFloatScaleFactor,
			   float *img_bkg, int *img_min, int *img_max,
			   float *img_avg, float *img_std,
			   int *numCols, int *numRows, 
			   int *numSecs,
			   int *imgOffset, int *pifMode, int *htype,
			   fString_t *descriptionStr, 
			   fString_t *timeStr,       /*RSS*/
			   fString_t *micrographStr, /*RSS*/
                           fString_t *scanNumberStr, /*RSS*/
			   float *pifVersion)
{/* PIF_READ_IMAGE_HEADER */
  return(pif_read_image_header_(luNum,
			   imgNum,
			   recipFloatScaleFactor,
			   img_bkg, img_min, img_max,
			   img_avg, img_std,
			   numCols, numRows, 
			   numSecs,
			   imgOffset, pifMode, htype,
			   descriptionStr, 
                           timeStr,       /*RSS*/
                           micrographStr, /*RSS*/
			   scanNumberStr, /*RSS*/
			   pifVersion));
}/* PIF_READ_IMAGE_HEADER */

int pif_read_data_header(int *luNum,
			 int *imgNum,
			 float *recipFloatScaleFactor,
			 float *int4scaleF,
			 float *img_bkg, 
			 int *img_min, int *img_max,
			 float *img_avg, float *img_std,
			 int *numCols, int *numRows, 
			 int *numSecs,
			 int *imgOffset, int *pifMode, int *htype,
			 char *descriptionStr,
			 char *timeStr,       /*RSS*/
			 char *micrographStr, /*RSS*/
			 char *scanNumberStr, /*RSS*/
			 float *aoverb, float *map_abang,
			 float *a_cell,float *b_cell, float *c_cell, 
			 float *alp, float *bet, float *gam, 
			 float *dela, float *delb, float *delc,
			 float *a_star, float *b_star, float *c_star,
			 float *alp_star, float *bet_star, float *gam_star,
			 float *pixelSize,
			 int *t_matrix, int *binFactor,
			 float *dthe, float *dphi_90, int *symmetry,
			 int *pixSizUnits, float *res_out,
			 int *ctfMode, float *tempFactor,
			 float *wiener, int *packRadius,
			 int *borderType, int *borderWidth,
			 float *r4_min, float *r4_max,
			 int *rad_lo, int *rad_hi,
			 float *pifVersion,
			 int *transX, int *transY, int *transZ, int *isTrunc,
			 int *origX, int *origY, int *origZ)
{/* PIF_READ_DATA_HEADER */
  return(pif_read_data_header_(luNum,
			       imgNum,
			       recipFloatScaleFactor,
			       int4scaleF,
			       img_bkg, img_min, img_max,
			       img_avg, img_std,
			       numCols, numRows, 
			       numSecs,
			       imgOffset, pifMode, htype,
			       descriptionStr,
			       timeStr,       /*RSS*/
			       micrographStr, /*RSS*/
			       scanNumberStr, /*RSS*/
			       aoverb, map_abang,
			       a_cell,b_cell, c_cell, 
			       alp, bet, gam, 
			       dela, delb, delc,
			       a_star,b_star,c_star,
			       alp_star,bet_star,gam_star,
			       pixelSize,
			       t_matrix, binFactor,
			       dthe, dphi_90, symmetry,
			       pixSizUnits, res_out,
			       ctfMode, tempFactor,
			       wiener, packRadius,
			       borderType, borderWidth,
			       r4_min, r4_max,
			       rad_lo, rad_hi,
			       pifVersion,
			       transX, transY, transZ, isTrunc,
			       origX, origY, origZ));
}/* PIF_READ_DATA_HEADER */

void pif_read_byte_image(int *luNum,
			 BYTE *byte_buffer,
			 int *imgOffset,
			 int *numCols, int *numRows, int *numSecs)
{/* PIF_READ_BYTE_IMAGE */
 pif_read_byte_image_(luNum,
			 byte_buffer,
			 imgOffset,
			 numCols, numRows, numSecs);
}/* PIF_READ_BYTE_IMAGE */

int pif_read_float_image(int *luNum,
			 float *float_buffer,
			 int *imgOffset,
			 int *numCols, int *numRows, int *numSecs,
			 int *binSwitch,
			 float *recipScaleF)
{/* PIF_READ_FLOAT_IMAGE */
  return(pif_read_float_image_(luNum,
			       float_buffer,
			       imgOffset,
			       numCols, numRows, numSecs,
			       binSwitch, recipScaleF));
}/* PIF_READ_FLOAT_IMAGE */

int pif_read_byteshort_image(int *luNum,
			      short *short_buffer,
			      int *imgOffset,
			      int *numCols, int *numRows, int *numSecs,
			      int *binSwitch)
{/* PIF_READ_BYTESHORT_IMAGE */
  return(pif_read_byteshort_image_(luNum,
			      short_buffer,
			      imgOffset,
			      numCols, numRows, numSecs,
			      binSwitch));
}/* PIF_READ_BYTESHORT_IMAGE */

void pif_read_short_image(int *luNum,
			  unsigned short *short_buffer,
			  int *imgOffset,
			  int *numCols, int *numRows, int *numSecs)
{/* PIF_READ_SHORT_IMAGE */
 pif_read_short_image_(luNum,
			  short_buffer,
			  imgOffset,
			  numCols, numRows, numSecs);
}/* PIF_READ_SHORT_IMAGE */


void pif_read_map_short_image(int *luNum,
			      float *float_buffer,
			      int *imgOffset,
			      int *numCols, int *numRows, int *numSecs,
			      float *recipFloatScaleFactor,
			      float *minMapPixel, float *maxMapPixel,
			      int *binSwitch,
			      int *strtCol, int *strtRow, int *strtSec,
			      int *ncol, int *nrow, int *nsec )
{/* PIF_READ_MAP_SHORT_IMAGE */
 pif_read_map_short_image_(luNum,
			   float_buffer,
			   imgOffset,
			   numCols, numRows, numSecs,
			   recipFloatScaleFactor,
			   minMapPixel, maxMapPixel,
			   binSwitch,strtCol,strtRow,strtSec,
			   ncol,nrow,nsec);
}/* PIF_READ_MAP_SHORT_IMAGE */


void pif_read_imgi4(forPifHead_t *thisHead,
		    float *float_buffer,
		    int * ncol,
		    int * nrow,
		    int * thisImg)
{
  pif_read_imgi4_(thisHead, float_buffer, ncol, nrow, thisImg);
}

void pif_read_map_int_image(int *luNum,
			     float *float_buffer,
			     int *imgOffset,
			     int *numCols, int *numRows, int *numSecs,
			     float *recipFloatScaleFactor,
			     float *minMapPixel, float *maxMapPixel,
			     int *binSwitch, int *mode)
{/* PIF_READ_MAP_INT_IMAGE */
 pif_read_map_int_image_(luNum,
			     float_buffer,
			     imgOffset,
			     numCols, numRows, numSecs,
			     recipFloatScaleFactor,
			     minMapPixel, maxMapPixel,
			     binSwitch, mode);
}/* PIF_READ_MAP_INT_IMAGE */


/***********
 * WRITING *
 ***********/
void pif_list_head(forPifHead_t *thisHead)
{
  pif_list_head_(thisHead);
}

int pif_write_gh(forPifHead_t *thisHead)
{
  return(pif_write_gh_(thisHead));
}

int pif_write_dh(forPifHead_t *thisHead)
{
  return(pif_write_dh_(thisHead));
}

int pif_write_header(int *luNum,
		     int *nimg, 
		     float *recipFloatScaleFactor,
		     float *int4scaleF,
		     int *htype, int *numCols, int *numRows, int *numSecs,
		     int *mode, int *endianNess, char *versionStr, 
		     int *even, int *mrcX, int *mrcY, int *mrcZ)
{/* PIF_WRITE_HEADER */
  return(pif_write_header_(luNum,
			   nimg, 
			   recipFloatScaleFactor,
			   int4scaleF,
			   htype, numCols, numRows, numSecs,
			   mode, endianNess, versionStr, even, mrcX, mrcY, mrcZ
			   ));
}/* PIF_WRITE_HEADER */

void pif_write_image_header(int *luNum,
			    int *imgNum,
			    float *floatScaleFactor,
			    float *img_bkg, int *img_min, int *img_max,
			    float *img_avg, float *img_std,
			    int *numCols, int *numRows, int *numSecs,
			    int *imgOffset, int *mode,
			    char *descriptionStr, char *timeStr,
			    char *micrographStr, char *scanNumberStr)
{/* PIF_WRITE_IMAGE_HEADER */
  pif_write_image_header_(luNum,
			    imgNum,
			    floatScaleFactor,
			    img_bkg, img_min, img_max,
			    img_avg, img_std,
			    numCols, numRows, numSecs,
			    imgOffset, mode,
			    descriptionStr, timeStr,
			    micrographStr, scanNumberStr);
}/* PIF_WRITE_IMAGE_HEADER */

int pif_write_data_header(int *luNum,
			  float *floatScaleFactor,
			  float *int4scaleF,
			  float *img_bkg, int *img_min, int *img_max,
			  float *img_avg, float *img_std,
			  int *numCols, int *numRows, 
			  int *numSecs,
			  int *imgOffset, int *pifMode, int *htype,
			  char *descriptionStr,
			  char *timeStr,       /*RSS*/
			  char *micrographStr, /*RSS*/
			  char *scanNumberStr, /*RSS*/
			  float *aoverb, float *map_abang,
			  float *a_cell,float *b_cell,
			  float *c_cell, float *alp, float *bet,
			  float *gam, float *dela, float *delb, float *delc,
			  float *a_star,float *b_star,float *c_star,
			  float *alp_star,float *bet_star,float *gam_star,
			  float *pixelSize,
			  int *t_matrix, int *packRadius, int *binFactor,
			  float *dthe, float *dphi_90, int *symmetry,
			  int *pixSizUnits, float *res_out,
			  int *ctfMode, float *tempFactor,
			  float *wiener,
			  int *borderType, int *borderWidth,
			  float *r4_min, float *r4_max,
			  int *rad_lo, int *rad_hi,
			  int *transX, int *transY, int *transZ, int *isTrunc,
			  int *origX, int *origY, int *origZ)
{/* PIF_WRITE_DATA_HEADER */
  return(pif_write_data_header_(luNum,
				floatScaleFactor,
				int4scaleF,
				img_bkg, img_min, img_max,
				img_avg, img_std,
				numCols, numRows, 
				numSecs,
				imgOffset, pifMode, htype,
				descriptionStr,
				timeStr,       /*RSS*/
				micrographStr, /*RSS*/
				scanNumberStr, /*RSS*/
				aoverb, map_abang,
				a_cell,b_cell,
				c_cell, alp, bet,
				gam, dela, delb, delc,
				a_star,b_star,c_star,
				alp_star,bet_star,gam_star,
				pixelSize,
				t_matrix, packRadius, binFactor,
				dthe, dphi_90, symmetry,
				pixSizUnits, res_out,
				ctfMode, tempFactor,
				wiener,
				borderType, borderWidth,
				r4_min, r4_max,
				rad_lo,rad_hi,
				transX, transY, transZ, isTrunc,
				origX, origY, origZ));
}/* PIF_WRITE_DATA_HEADER */

void pif_write_prji1(forPifHead_t *prjHead,
		      float *float_buffer,
		      int *id,
		      int *numCols, int *numRows,
		     float *sFac, float *min)
{/* PIF_WRITE_PRJI1 */
  pif_write_prji1_(prjHead,
		    float_buffer,
		    id,
		    numCols, numRows,
		   sFac, min);
}/* PIF_WRITE_PRJI1 */


void pif_write_byteshort_image(int *luNum,
			  short *short_buffer,
			  int *imgOffset,
			  int *numCols, int *numRows, int *numSecs)
{/* PIF_WRITE_BYTESHORT_IMAGE */
 pif_write_byteshort_image_(luNum,
			    short_buffer,
			    imgOffset,
			    numCols, numRows, numSecs);
}/* PIF_WRITE_BYTESHORT_IMAGE */

void pif_write_byte_image(int *luNum,
			  BYTE *byte_buffer,
			  int *imgOffset,
			  int *numCols, int *numRows, int *numSecs)
{/* PIF_WRITE_BYTE_IMAGE */
 pif_write_byte_image_(luNum,
			  byte_buffer,
			  imgOffset,
			  numCols, numRows, numSecs);
}/* PIF_WRITE_BYTE_IMAGE */

void pif_write_short_image(int *luNum,
			   unsigned short *short_buffer,
			   int *imgOffset,
			   int *numCols, int *numRows, int *numSecs)
{/* PIF_WRITE_SHORT_IMAGE */
 pif_write_short_image_(luNum,
			short_buffer,
			imgOffset,
			numCols, numRows, numSecs);
}/* PIF_WRITE_SHORT_IMAGE */


void pif_write_mapi2(int *luNum,  float *float_buffer, 
		     int *numCols, int *numRows, 
		     int *numSecs, float *floatScaleFactor, int *bin)
{/* PIF_WRITE_MAPI2 */
  pif_write_mapi2_(luNum, float_buffer, 
		   numCols, numRows, 
		   numSecs, floatScaleFactor, bin);
}/* PIF_WRITE_MAPI2 */


void pif_write_map_short_image(int *luNum,
				short *short_buffer,
				int *mapOffset,
				int *numCols)
{/* PIF_WRITE_MAP_SHORT_IMAGE */
  pif_write_map_short_image_(luNum,
			     short_buffer,
			     mapOffset,
			     numCols);
}/* PIF_WRITE_MAP_SHORT_IMAGE */

void pif_write_map_float_image(int *luNum,
				float *float_buffer,
				int *mapOffset,
				int *numCols,
				int *numRows,
				int *numSecs,
				float *floatScaleFactor)
{/* PIF_WRITE_MAP_FLOAT_IMAGE */
  pif_write_map_float_image_(luNum,
			     float_buffer,
			     mapOffset,
			     numCols,
			     numRows,
			     numSecs,
			     floatScaleFactor);
}/* PIF_WRITE_MAP_FLOAT_IMAGE */

void pif_write_map_section(int *luNum,
			    float *float_buffer,
			    int *secNum,
			    int *numCols,
			    int *numRows,
			    float *floatScaleFactor)
{/* PIF_WRITE_MAP_SECTION */
  pif_write_map_section_(luNum,
			 float_buffer,
			 secNum,
			 numCols,
			 numRows,
			 floatScaleFactor);
}/* PIF_WRITE_MAP_SECTION */


void pif_write_map_fltint(int *luNum,
			   float *float_buffer,
			   int *mapOffset,
			   int *numCols,
			   int *numRows,
			   int *numSecs,
			   float *floatScaleFactor)
{/* PIF_WRITE_MAP_FLTINT */
  pif_write_map_fltint_(luNum,
			float_buffer,
			mapOffset,
			numCols,
			numRows,
			numSecs,
			floatScaleFactor);
}/* PIF_WRITE_MAP_FLTINT */




/************************************************************
  ==========================================================
  Now the real code ...
  ==========================================================
  ***********************************************************/
/*********************************************************
 * makefv_fname() - Make appropriate names for emicofv.
*/
void makefv_fname_(char *inFname, char *out1, char *out2, char *out3)
{/* MAKEFV_FNAME */
  int i,j,incrVal;
  char *s1,*s2,*s3;
  static char tmpStr[512];

  /* Init */
  sprintf(tmpStr,"%s",inFname);
  s1 = strchr(tmpStr,'.');
  if(s1)
    *s1 = '\0';
  sprintf(out1,"%s%s",tmpStr,".tmpgrad2"); 
  sprintf(out2,"%s%s",tmpStr,".tmpfv"); 
  sprintf(out3,"%s%s",tmpStr,"_chart.pif"); 

}/* MAKEFV_FNAME */





/*********************************************************
 * increment_fname_unix() - Increment the number after the
 * tilde '~' in the extension.  USed by pft.for to 
 * calculate the out filename.  YOU MUST HAVE 256 BYTES for
 * the OUTFNAME and it's recommended to have 256 for the 
 * inFname. UNIX is (char *) VMS is (fString_t *).
*/
void increment_fname_unix_(char *inFname, char *outFname)
{/* INCREMENT_FNAME_UNIX */
  int i,j,incrVal;
  char *s1,*s2,*s3;
  char tmpStr[512];

  /* Init */
  for(i=0;i<256;i++)
    tmpStr[i]= '\0';
  for(i=0;i<256;i++)
    outFname[i]= '\0';
  /* FORTRAN doesn't always NULL terminate its strings esp.
   * if they're part of an array of strings.  Let's just get 
   * what we need here */
  for(i=0;i<256;i++)
    tmpStr[i]= inFname[i];
  tmpStr[255]= '\0';
  s1= strchr(tmpStr,' ');
  if(s1)
    *s1= '\0';       /* Get rid of trailing blanks */
  if(strlen(tmpStr) > 250)
    {/* Error */
      printf("Error - Filename too long [%s]\n",inFname);
      exit(0);
    }/* Error */
  s1= strrchr(tmpStr,'_');
  s2= strrchr(tmpStr,'.');
  incrVal = 1;
  if(s1 && (s1 > s2))
    {/* Find out the last value */
      incrVal = (int)strtod((s1+1),0);
      incrVal++;
      *(s1)= '\0';   /* Get rid of everything at and after the underscore */
    }/* Find out the last value */

  sprintf(outFname,"%s_%03d",tmpStr,incrVal); 

}/* INCREMENT_FNAME_UNIX */




/*********************************************************
 * increment_fname_() - Increment the number after the
 * tilde '~' in the extension.  USed by pft.for to 
 * calculate the out filename.  YOU MUST HAVE 256 BYTES for
 * the OUTFNAME and it's recommended to have 256 for the 
 * inFname.  VMS sends us this funky string descriptor.
 * UNIX uses char *.
*/
void increment_fname_(fString_t *inFname, fString_t *outFname)
{/* INCREMENT_FNAME_ */
  int i,j,incrVal;
  char *s1,*s2,*s3;
  char tmpStr[512];

#ifdef VMS
  /* Init */
  for(i=0;i<256;i++)
    tmpStr[i]= '\0';
  for(i=0;i<256;i++)
    (outFname->addr)[i]= '\0';
  /* FORTRAN doesn't always NULL terminate its strings esp.
   * if they're part of an array of strings.  Let's just get 
   * what we need here */
  for(i=0;i<256;i++)
    tmpStr[i]= (inFname->addr)[i];
  tmpStr[255]= '\0';
  s1= strchr(tmpStr,' ');
  if(s1)
    *s1= '\0';       /* Get rid of trailing blanks */
  if(strlen(tmpStr) > 250)
    {/* Error */
      printf("Error - Filename too long [%s]\n",inFname->addr);
      exit(0);
    }/* Error */
  s1= strrchr(tmpStr,'_');
  s2= strrchr(tmpStr,'.');
  incrVal = 1;
  if(s1 && (s1 > s2))
    {/* Find out the last value */
      incrVal = (int)strtod((s1+1),0);
      incrVal++;
      *(s1)= '\0';   /* Get rid of everything at and after the underscore */
    }/* Find out the last value */

  sprintf(outFname->addr,"%s_%03d",tmpStr,incrVal); 
#else
  increment_fname_unix_((char *)inFname,(char *)outFname);
#endif

}/* INCREMENT_FNAME_ */



/*****************************************************************************
** Util_IAmBigEndian - Written by James Conway.  Auto-detect whether we're
** BIG or little ENDIAN.
** =================
Return TRUE if this is a big-endian machine (byte order is hi/lo), else FALSE
(byte order is lo/hi).
*/
static
int Util_IAmBigEndian(void)
{/* Util_IAmBigEndian */
  static int firstTime = TRUE;
  static int bigEndian;

  if(firstTime)  {
    union {
      unsigned char  bytes[2];
      unsigned short word;
    } byteWord;
    firstTime = FALSE;
    byteWord.word = 1;
    bigEndian = (byteWord.bytes[1] == 1);
  }
  return(bigEndian);
}/* Util_IAmBigEndian */



/*********************************************************
 * differentEndian() - Convert this to a different ENDIAN
 * bit of memory. 
*/
void differentEndian_(BYTE *inStr, int length, int endianNess )
{/* DIFFERENTENDIAN */
  int i,lm1;
  BYTE thisStr[8];

  if((endianNess == 0) != (Util_IAmBigEndian() == 0)) 
    {
      for(i=0;i<length;i++)
	*(thisStr + i)= *(inStr + i);
      lm1= length - 1;
      for(i=0;i<length;i++)
	*(inStr + i) = (*(thisStr + (lm1-i)) & 0377);
    }
}/* DIFFERENTENDIAN */




/*******************************************************************
 * convertBackFloat() - Convert the int passed in to the float
 * passed.  Use the recipFloatScaleFactor read from the file.
*/
void convertBackFloat_(int thisInt, float *thisFloat, 
		      float recipFloatScaleFactor)
{/* CONVERTBACKFLOAT */
  int tempI,tempI2;
  int tempF;

#ifdef OLDWAY
  *thisFloat = (float)(thisInt * recipFloatScaleFactor);
#else
  /* We should be using 32768 here instead of 30000 - it's 2 raised to
   * the 15th power == 32768  */
#ifdef USETHISKLUDGE
  if(i2Map && (thisInt >= 32768 || thisInt <= -32768))
    {/* We've got to bump this up to 4 bytes */
      *thisFloat = (float)(((float)thisInt) * recipFloatScaleFactor)/64000.0;
    }/* We've got to bump this up to 4 bytes */
  else
#endif
    *thisFloat = (float)(thisInt * recipFloatScaleFactor);
#endif

}/* CONVERTBACKFLOAT */





/************************************************************
 * convertFloat() - Convert the float passed in to the int
 * passed.  Use the floatScaleFactor calculated by calcScaleFactor()
 * below.
*/
void convertFloat_(int *thisInt, float thisFloat, float floatScaleFactor)
{/* CONVERTFLOAT */
  int tempI,tempI2;
  float tempF;

#ifdef ODLWAY
  *thisInt = thisFloat * floatScaleFactor;
#else
#ifdef USETHISKLUDGE
  if(i2Map)
    {/* We've got to bump this up to 4 bytes */
      /* We should be using 32768 here instead of 30000 - it's 2 raised to
       * the 15th power == 32768  */
      *thisInt = (int)((thisFloat * 64000.0) * floatScaleFactor);
    }/* We've got to bump this up to 4 bytes */
  else
#endif
  *thisInt = thisFloat * floatScaleFactor;
#endif

}/* CONVERTFLOAT */




/************************************************************
 * pif_close() - Close up the PIF file.
*/
void pif_close_(int *luNum)
{/* PIF_CLOSE */

  fclose(lun[*luNum]);

}/* PIF_CLOSE */







/************************************************************
 * pif_open() - Open up the PIF file.
*/
void pif_open_(char *fName, int *luNum, char *readWrite)
{/* PIF_OPEN */
  BYTE *b1;
  int i,leadingWS,thisPos;
  char *s, fin_name[210], r_w[4];

#ifdef VMS

  /* VMS does something funky with the strings */
  pif_open_vms_((fString_t *)fName, luNum, readWrite);

#else

  /* Init */
  s = (fName);
  fin_name[0] = '\0';
  /****change the Fortran filename to C string **/
  leadingWS = TRUE;
  thisPos = 0;
  for(i=0; i < 200 ; i++) {
        if(*s != ' ' && leadingWS) leadingWS = FALSE;
        if(*s == ' ' && !leadingWS) break;
        fin_name[thisPos] = *s;
	if(*s != ' ') /* Then we've got a good one */
	  thisPos++;
        s++;
  }
  fin_name[thisPos] = '\0';
  r_w[0] = *(readWrite);
  r_w[1] = *(readWrite+1);
  r_w[2] = '\0';

  if(*luNum >= MAXLUNS)
    {/* Error */
      printf("PIF_OPEN Maximum LUN exceeded [%d]\n",*luNum);
      exit(0);
    }/* Error */

  lun[*luNum] = fopen( fin_name, r_w );
  if((lun[*luNum])==0)
    {/* Error */
      printf("****************************************\n");
      printf("* [PIF_OPEN] Error opening [%s] for [%s]\n\n",
             fin_name, r_w);
      printf("****************************************\n");
      exit(0);
    }/* Error */
#endif

}/* PIF_OPEN */




/************************************************************
 * pif_open_vms_() - Open up the PIF file.
*/
void pif_open_vms_(fString_t *fName, int *luNum, char *readWrite)
{/* PIF_OPEN_VMS */
  BYTE *b1;
  int i,firstChar,lastChar;
  char *s, fin_name[200], r_w[4];

  /* Init */
  s = (fName->addr);
  fin_name[0] = '\0';
  /****change the Fortran filename to C string **/
  firstChar = FALSE;
  for(i=0; i < 200 ; i++) 
    {
      if(*s != ' ')
	firstChar = TRUE;
      lastChar = i;
      if(firstChar && *s == ' ') break;
      fin_name[i] = *s;
      s++;
    }
  fin_name[lastChar] = '\0';
  r_w[0] = *(readWrite);
  r_w[1] = *(readWrite+1);
  r_w[2] = '\0';

  if(*luNum >= MAXLUNS)
    {/* Error */
      printf("PIF_OPEN Maximum LUN exceeded [%d]\n",*luNum);
      exit(0);
    }/* Error */

  lun[*luNum] = fopen(fin_name, r_w );
  if((lun[*luNum])==0)
    {/* Error */
      printf("****************************************\n");
      printf("* [PIF_OPEN] Error opening [%s] for [%s]\n\n",
             fin_name, r_w);
      printf("****************************************\n");
      exit(0);
    }/* Error */

}/* PIF_OPEN_VMS */



/*************************************************************
 * pif_read_gh() - Read the PIF file's global header.
 * Changed reading in of FloatScaleFactor, R. Ashmore, D. Belnap, 30 Sep 1997
*/
int pif_read_gh_(forPifHead_t *thisHead)
{/* PIF_READ_GH */
  int status;

  status = pif_read_header_(&(thisHead->luNum),
			    &(thisHead->nimg),
			    &(thisHead->recipFloatScaleFactor),
			    &(thisHead->int4scaleF),
			    &(thisHead->htype),
			    &(thisHead->numCols),
			    &(thisHead->numRows),
			    &(thisHead->numSecs),
			    &(thisHead->mode),
			    &(thisHead->endianNess),
			    (thisHead->versionStr),
			    &(thisHead->pifVersion),
			    &(thisHead->even),
			    &(thisHead->mrcX),
			    &(thisHead->mrcY),
			    &(thisHead->mrcZ)
			    );

  /* We will need thisis we ever want to write this again */
  thisHead->floatScaleFactor = 1.0 / thisHead->recipFloatScaleFactor;

  return(status);

}/* PIF_READ_GH */



/***************************************************************
 * pif_read_dh() - Read the PIF file's image header.
 * Read the data desired into a struct to help the application
 * programmer better deal with the mountain of data this 
 * entails.
*/
int pif_read_dh_(forPifHead_t *thisHead)
{/* PIF_READ_DH */

  return(pif_read_data_header_(&(thisHead->luNum),
			       &(thisHead->imgNum),
			       &(thisHead->recipFloatScaleFactor),
			       &(thisHead->int4scaleF),
			       &(thisHead->img_bkg),
			       &(thisHead->datai4_min),
			       &(thisHead->datai4_max),
			       &(thisHead->datar4_avg),
			       &(thisHead->datar4_std),
			       &(thisHead->numCols),
			       &(thisHead->numRows),
			       &(thisHead->numSecs),
			       &(thisHead->imgOffset),
			       &(thisHead->mode),
			       &(thisHead->htype),
			       (thisHead->descriptionStr),
			       (thisHead->timeStr),       /*RSS*/
			       (thisHead->micrographStr), /*RSS*/
			       (thisHead->scanNumberStr), /*RSS*/
			       &(thisHead->aoverb),
			       &(thisHead->map_abang),
			       &(thisHead->a_cell),
			       &(thisHead->b_cell),
			       &(thisHead->c_cell),
			       &(thisHead->alp),
			       &(thisHead->bet),
			       &(thisHead->gam),
			       &(thisHead->dela),
			       &(thisHead->delb),
			       &(thisHead->delc),
			       &(thisHead->a_star),
			       &(thisHead->b_star),
			       &(thisHead->c_star),
			       &(thisHead->alp_star),
			       &(thisHead->bet_star),
			       &(thisHead->gam_star),
			       &(thisHead->pixelSize),
			       (thisHead->t_matrix),
			       &(thisHead->binFactor),
			       &(thisHead->dthe),
			       &(thisHead->dphi_90),
			       &(thisHead->symmetry),
			       &(thisHead->pixSizUnits),
			       &(thisHead->res_out),
			       &(thisHead->ctfMode),
			       &(thisHead->tempFactor),
			       &(thisHead->wiener),
			       &(thisHead->packRadius),
			       &(thisHead->borderType),
			       &(thisHead->borderWidth),
			       &(thisHead->datar4_min),
			       &(thisHead->datar4_max),
			       &(thisHead->rad_lo),
			       &(thisHead->rad_hi),
			       &(thisHead->pifVersion),
			       &(thisHead->transX),
			       &(thisHead->transY),
			       &(thisHead->transZ),
			       &(thisHead->isTrunc),
			       &(thisHead->origX),
			       &(thisHead->origY),
			       &(thisHead->origZ)
			       ));

}/* PIF_READ_DH */



/*************************************************************
 * pif_read_header() - Read the PIF file's global header.
 * Changed reading in of FloatScaleFactor, R. Ashmore, D. Belnap, 30 Sep 1997
*/
int pif_read_header_(int *luNum,
		     int *nimg, 
		     float *recipFloatScaleFactor,
		     float *int4scaleF,
		     int *htype, int *numCols, int *numRows, int *numSecs,
		     int *mode, int *endianNess, char *versionStr,
		     float *pifVersion, 
		     int *even, int *mrcX, int *mrcY, int *mrcZ)
{/* PIF_READ_HEADER */
  int  j,i;
  char floatScaleStr[17];
  char file_id[4];
  double thisDouble;
  char *s1;
  FILE *fIn;

  fIn = lun[*luNum];
  
  /* [1] Read global file header */
  if(fseek(fIn,0,0) != 0)
    {/* Can't seek */
      printf("[PIF_READ_HEADER] Error seeking file.\n");
      exit(0);
    }/* Can't seek */

  /* Read the file type header - its there twice, check the 2nd one */
  if(fread(&file_id[0],sizeof(file_id),1,fIn) != 1) /*ROK*/
    {/* Can't read file */
      printf("[PIF_READ_HEADER] Error reading file.\n");
      exit(0);
    }/* Can't read file */
  if(fread(&file_id[0],sizeof(file_id),1,fIn) != 1) /*ROK*/
    {/* Can't read file */
      printf("[PIF_READ_HEADER] Error reading file.\n");
      exit(0);
    }/* Can't read file */

  if(file_id[0]!=8 && file_id[1]!=0 && file_id[2]!=0 && file_id[3]!=0)
    {/* printf */
      printf("[PIF_READ_HEADER] Don't understand file type [%d,%d,%d,%d]\n"
                        ,file_id[0] ,file_id[1] ,file_id[2] ,file_id[3] );
      return(FALSE);
    }/* printf */

  OKREAD(fread(floatScaleStr,sizeof(char),16,fIn),16,"PIF_READ_HEADER"); /*ROK*/
  floatScaleStr[16]= '\0';
  s1 = floatScaleStr;
  /*  printf("floatScaleStr = %s\n",floatScaleStr); */
  while(*s1==' ')                                 /* Get rid of white space */
    s1++;
  *recipFloatScaleFactor = (float)strtod(s1,0);
  fflush(stdout);
/* 24 */

  if(*recipFloatScaleFactor==0.0)
    {/* Error */
      printf("Invalid scaling factor[%f]\n",*recipFloatScaleFactor);
      *recipFloatScaleFactor= 1.0;
      /* exit(0); */
    }/* Error */

  OKREAD(fread(nimg,sizeof(int),1,fIn),1,"PIF_READ_HEADER"); /*ROK*/
  OKREAD(fread(endianNess,sizeof(int),1,fIn),1,"PIF_READ_HEADER"); /*ROK*/
  endians[*luNum] = (*endianNess) != 0;
/* 32 */

  differentEndian_((BYTE *)nimg, 4, endians[*luNum]);

  fread(versionStr,sizeof(char),32,fIn); /*ROK*/
  *pifVersion = (float)strtod(versionStr,0);
/* 64 */

  fread(htype,sizeof(int),1,fIn);   /*ROK*/
  fread(numCols,sizeof(int),1,fIn); /*ROK*/
  fread(numRows,sizeof(int),1,fIn); /*ROK*/
  fread(numSecs,sizeof(int),1,fIn); /*ROK*/
  differentEndian_((BYTE *)htype, 4, endians[*luNum]);
  differentEndian_((BYTE *)numCols, 4, endians[*luNum]);
  differentEndian_((BYTE *)numRows, 4, endians[*luNum]);
  differentEndian_((BYTE *)numSecs, 4, endians[*luNum]);
  if(*htype!=1)	/* htype==1 means all images have the same size */
    {/* can\'t goon */
	printf("[PIF_READ_HEADER] Don't understand this type [%d]\n",*htype); /*RSS*/
      return(FALSE);
      /* exit(0); */
    }/* can\'t goon */
/* 80 */


  fread(mode,sizeof(int),1,fIn); /*ROK*/
  differentEndian_((BYTE *)mode, 4, endians[*luNum]);
  if(*mode==20)
    i2Map = TRUE;
  else i2Map = FALSE;
/* 84 */

  fread(even,sizeof(int),1,fIn); /*ROK*/
  differentEndian_((BYTE *)even, 4, endians[*luNum]);
  fread(mrcX,sizeof(int),1,fIn); /*ROK*/
  differentEndian_((BYTE *)mrcX, 4, endians[*luNum]);
  fread(mrcY,sizeof(int),1,fIn); /*ROK*/
  differentEndian_((BYTE *)mrcY, 4, endians[*luNum]);
  fread(mrcZ,sizeof(int),1,fIn); /*ROK*/
  differentEndian_((BYTE *)mrcZ, 4, endians[*luNum]);
/* 100 */


  OKREAD(fread(floatScaleStr,sizeof(char),16,fIn),16,"PIF_READ_HEADER");/*ROK*/
  floatScaleStr[16]= '\0';
  s1 = floatScaleStr;
  /*  printf("floatScaleStr = %s\n",floatScaleStr); */
  while(*s1==' ')                                 /* Get rid of white space */
    s1++;
  *int4scaleF = (float)strtod(s1,0);
  fflush(stdout);
/* 116 */

  /* We have a historical bug whereby old PIF files had the
   * numberOfImages variable stuffed here.  Fixed 3/97 */
  if(*mode==6) /* No sections for packed boxes */
    {/* Packed boxes */
      *numSecs = 1;
    }/* Packed boxes */

  return(TRUE);

}/* PIF_READ_HEADER */





/*************************************************************
 * pif_read_image_header() - Read the PIF file's header for a 
 * specific image.
 * JFC.2 23-May-97 Fast seeking if htype=1
 * JFC.3 27-May-97 Added error return and error checking.
*/
int pif_read_image_header_(int *luNum,
			   int *imgNum,
			   float *recipFloatScaleFactor,
			   float *img_bkg, int *img_min, int *img_max,
			   float *img_avg, float *img_std,
			   int *numCols, int *numRows, 
			   int *numSecs,
			   int *imgOffset, int *pifMode, int *htype,
			   fString_t *descriptionStr, 
			   fString_t *timeStr,       /*RSS*/
			   fString_t *micrographStr, /*RSS*/
			   fString_t *scanNumberStr, /*RSS*/
			   float *pifVersion)
{/* PIF_READ_IMAGE_HEADER */
  int  i,j,k;
  long seekPos, imageBytes;  
  int  thisNcols,thisNrows,thisNsecs;
  char floatScaleStr[57];
  char versionStr[33];
  /*char timeStr[36];*/         /*RSS*/
  /*char micrographStr[200];*/  /*RSS*/
  int  scanNum;
  /*char scanNumberStr[210];*/  /*RSS*/
  /*char descStr[100];*/        /*RSS*/
  char *s1;
  float junkFloat;
  int  junkInt,localInt;
  int  numRead;
  int mrc_nxstart,mrc_nystart,mrc_nzstart,
      mrc_mx,mrc_my,mrc_mz,mrc_xlength,mrc_ylength,
      mrc_zlength,mrc_alpha,mrc_beta,mrc_gamma,
      mrc_mapc,mrc_mapr,mrc_maps,mrc_min,mrc_max,
      mrc_mean,mrc_std,mrc_ispg,mrc_nsymbt,mrc_xorigin,
      mrc_yorigin,packRadius;
  FILE *fIn;

  fIn = lun[*luNum];    /* a */

  if((*htype)==1)  
    {/* all images have same size, we can fseek directly */
      
      /* We have a historical bug whereby old PIF files had the
       * numberOfImages variable stuffed here.  Fixed 3/97 */
      if((*pifMode)==6) 	/* No sections for packed boxes */
	(*numSecs) = (int)1;

      imageBytes =  (*numCols) * (*numRows) * (*numSecs);
      seekPos = 512 + ((*imgNum)-1) * (512 + imageBytes);
      
      fseek(fIn,seekPos,0);
      OKREAD(fread(&thisNcols, sizeof(int),1,fIn),1,"PIF_READ_IMAGE_HEADER");/*ROK*/
      fread(&thisNrows, sizeof(int),1,fIn);/*ROK*/
      fread(&thisNsecs, sizeof(int),1,fIn);/*ROK*/
      differentEndian_((BYTE *)&thisNcols, 4, endians[*luNum]);
      differentEndian_((BYTE *)&thisNrows, 4, endians[*luNum]);
      differentEndian_((BYTE *)&thisNsecs, 4, endians[*luNum]);

      /* We have a historical bug whereby old PIF files had the
       * numberOfImages variable stuffed here.  Fixed 3/97 */
      if((*pifMode)==6) 	/* No sections for packed boxes */
	thisNsecs = 1;

      /* Check that the size is what we expect (since htype==1) */
      if((*pifMode != 31) &&
	 (thisNcols != *numCols ||
	  thisNrows != *numRows ||
	  thisNsecs != *numSecs)  
	 )
	{
	  printf("Error in pif_read_image_header for image %d\n", *imgNum);
	  printf("  - htype=1 implies all images should have the same size\n");
	  printf("  - expected size is (%d,%d,%d)\n", 
		 *numCols, *numRows, *numSecs);
	  printf("  - found size is    (%d,%d,%d)\n", 
		 thisNcols, thisNrows, thisNsecs);
	  return(1);	/* error */
	}
    }/* all images have same size, we can fseek directly */
  else 
    {/* htype != 1, we must work through all the images */
      seekPos = 512;            /* global header */
      fseek(fIn,seekPos,0);
      
      for(i=0;i<(*imgNum);i++)
	{/* Seek the one we want */
	    fread(&thisNcols, sizeof(int),1,fIn);/*ROK*/
	    fread(&thisNrows, sizeof(int),1,fIn);/*ROK*/
	    fread(&thisNsecs, sizeof(int),1,fIn);/*ROK*/
	  differentEndian_((BYTE *)&thisNcols, 4, endians[*luNum]);
	  differentEndian_((BYTE *)&thisNrows, 4, endians[*luNum]);
	  differentEndian_((BYTE *)&thisNsecs, 4, endians[*luNum]);
	  
	  /* We have a historical bug whereby old PIF files had the
	   * numberOfImages variable stuffed here.  Fixed 3/97 */
	  if(*pifMode==6) /* No sections for packed boxes */
	    {/* Packed boxes */
	      thisNsecs = 1;
	    }/* Packed boxes */
	  
	  if((i+1) < *imgNum)  
	    {
	      seekPos += 512 + (thisNcols*thisNrows*thisNsecs); 
	      /* Header + image*/
	      fseek(fIn,seekPos,0);
	    }
	}/* Seek the one we want */
      *numCols = thisNcols;
      *numRows = thisNrows;
      *numSecs = thisNsecs;
    }/* htype != 1, we must work through all the images */
  
  *imgOffset = seekPos + 512;  /* image starts after image header */
  
  if(fIn != NULL)
    {/* It is readable */
      
      /* [2] Read header for each image */
	fread(pifMode,sizeof(int),1,fIn); /*ROK*/
	differentEndian_((BYTE *)pifMode, 4, endians[*luNum]);
	fread(&localInt,sizeof(int),1,fIn); /*ROK*/       
	differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      if(*pifVersion >= 2.09)
	convertBackFloat_(localInt,img_bkg,*recipFloatScaleFactor);
      else
	*img_bkg = (float)localInt;

      fread(&packRadius,sizeof(int),1,fIn); /*ROK*/
      differentEndian_((BYTE *)&packRadius, 4, endians[*luNum]);

      fread(&mrc_nxstart,sizeof(int),1,fIn); /*ROK*/
      differentEndian_((BYTE *)&mrc_nxstart, 4, endians[*luNum]);
      fread(&mrc_nystart,sizeof(int),1,fIn); /*ROK*/       
      differentEndian_((BYTE *)&mrc_nystart, 4, endians[*luNum]);
      fread(&mrc_nzstart,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&mrc_nzstart, 4, endians[*luNum]);
      fread(&mrc_mx,sizeof(int),1,fIn); /*ROK*/                    
      differentEndian_((BYTE *)&mrc_mx, 4, endians[*luNum]);
      fread(&mrc_my,sizeof(int),1,fIn); /*ROK*/                    
      differentEndian_((BYTE *)&mrc_my, 4, endians[*luNum]);
      fread(&mrc_mz,sizeof(int),1,fIn); /*ROK*/                    
      differentEndian_((BYTE *)&mrc_mz, 4, endians[*luNum]);
      fread(&mrc_xlength,sizeof(int),1,fIn); /*ROK*/             
      differentEndian_((BYTE *)&mrc_xlength, 4, endians[*luNum]);
      fread(&mrc_ylength,sizeof(int),1,fIn); /*ROK*/             
      differentEndian_((BYTE *)&mrc_ylength, 4, endians[*luNum]);
      fread(&mrc_zlength,sizeof(int),1,fIn); /*ROK*/         
      differentEndian_((BYTE *)&mrc_zlength, 4, endians[*luNum]);
    
      convertBackFloat_(mrc_xlength,&junkFloat,*recipFloatScaleFactor);
      convertBackFloat_(mrc_ylength,&junkFloat,*recipFloatScaleFactor);
      convertBackFloat_(mrc_zlength,&junkFloat,*recipFloatScaleFactor);

      fread(&mrc_alpha,sizeof(int),1,fIn); /*ROK*/
      differentEndian_((BYTE *)&mrc_alpha, 4, endians[*luNum]);
      fread(&mrc_beta,sizeof(int),1,fIn);  /*ROK*/        
      differentEndian_((BYTE *)&mrc_beta, 4, endians[*luNum]);
      fread(&mrc_gamma,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&mrc_gamma, 4, endians[*luNum]);

      convertBackFloat_(mrc_alpha,&junkFloat,*recipFloatScaleFactor);
      convertBackFloat_(mrc_beta,&junkFloat,*recipFloatScaleFactor);
      convertBackFloat_(mrc_gamma,&junkFloat,*recipFloatScaleFactor);

      fread(&mrc_mapc,sizeof(int),1,fIn); /*ROK*/
      differentEndian_((BYTE *)&mrc_mapc, 4, endians[*luNum]);
      fread(&mrc_mapr,sizeof(int),1,fIn); /*ROK*/          
      differentEndian_((BYTE *)&mrc_mapr, 4, endians[*luNum]);
      fread(&mrc_maps,sizeof(int),1,fIn); /*ROK*/           
      differentEndian_((BYTE *)&mrc_maps, 4, endians[*luNum]);
      fread(&mrc_min,sizeof(int),1,fIn);  /*ROK*/          
      differentEndian_((BYTE *)&mrc_min, 4, endians[*luNum]);
      fread(&mrc_max,sizeof(int),1,fIn);  /*ROK*/         
      differentEndian_((BYTE *)&mrc_max, 4, endians[*luNum]);
      fread(&mrc_mean,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&mrc_mean, 4, endians[*luNum]);
      fread(&mrc_std,sizeof(int),1,fIn);  /*ROK*/    
      differentEndian_((BYTE *)&mrc_std, 4, endians[*luNum]);

      *img_min= mrc_min;
      *img_max= mrc_max;
      convertBackFloat_(mrc_mean,img_avg,*recipFloatScaleFactor);
      convertBackFloat_(mrc_std,img_std,*recipFloatScaleFactor);

      fread(&mrc_ispg,sizeof(int),1,fIn);       /*ROK*/         
      differentEndian_((BYTE *)&mrc_ispg, 4, endians[*luNum]);
      fread(&mrc_nsymbt,sizeof(int),1,fIn);     /*ROK*/       
      differentEndian_((BYTE *)&mrc_nsymbt, 4, endians[*luNum]);
      fread(&mrc_xorigin,sizeof(float),1,fIn);  /*ROK*/    
      differentEndian_((BYTE *)&mrc_xorigin, 4, endians[*luNum]);
      fread(&mrc_yorigin,sizeof(float),1,fIn);  /*ROK*/    
      differentEndian_((BYTE *)&mrc_yorigin, 4, endians[*luNum]);
      
      /*fread(descStr,sizeof(char),80,fIn); */ /*RSS*/ /*ROK*/
      /*      sprintf(descriptionStr->addr,"%s",descStr); */
      fread(descriptionStr,sizeof(char),80,fIn); /*RSS*/ /*ROK*/
      fread(timeStr,sizeof(char),32,fIn);                /*ROK*/
      fread(micrographStr,sizeof(char),16,fIn);          /*ROK*/
      fread(scanNumberStr,sizeof(char),8,fIn);           /*ROK*/
      
    }/* It is readable */
  else
    {/* error */
	printf("[PIF_READ_IMAGE_HEADER] Image [%d] unreadable.\n",*imgNum);/*RSS*/
      return(1);	/* error */
    }/* error */

  return(0); 		/* no error	   */
}/* PIF_READ_IMAGE_HEADER */



/*************************************************************
 * pif_read_data_header() - Read the PIF file's header for a 
 * specific image.
*/
int pif_read_data_header_(int *luNum,
			  int *imgNum,
			  float *recipFloatScaleFactor,
			  float *int4scaleF,
			  float *img_bkg, int *img_min, int *img_max,
			  float *img_avg, float *img_std,
			  int *numCols, int *numRows, 
			  int *numSecs,
			  int *imgOffset, int *pifMode, int *htype,
			  char *descriptionStr,
			  char *timeStr,       /*RSS*/
			  char *micrographStr, /*RSS*/
			  char *scanNumberStr, /*RSS*/
			  float *aoverb, float *map_abang,
			  float *a_cell,float *b_cell, float *c_cell, 
			  float *alp, float *bet, float *gam, 
			  float *dela, float *delb, float *delc,
			  float *a_star,float *b_star,float *c_star,
			  float *alp_star,float *bet_star,float *gam_star,
			  float *pixelSize,
			  int *t_matrix, int *binFactor,
			  float *dthe, float *dphi_90, int *symmetry,
			  int *pixSizUnits, float *res_out,
			  int *ctfMode, float *tempFactor,
			  float *wiener, int *packRadius,
			  int *borderType, int *borderWidth,
			  float *r4_min, float *r4_max,
			  int *rad_lo, int *rad_hi,
			  float *pifVersion,
			  int *transX, int *transY, int *transZ, int *isTrunc,
			  int *origX, int *origY, int *origZ
			  )
{/* PIF_READ_DATA_HEADER */
  int  i,j,k;
  long seekPos, imageBytes;  
  int  thisNcols,thisNrows,thisNsecs,localInt;
  char floatScaleStr[57];
  char versionStr[33];
  /*char timeStr[36];*/        /*RSS*/
  /*char micrographStr[200];*/ /*RSS*/
  int  scanNum;
  /*char scanNumberStr[210];*/ /*RSS*/
  char descStr[100];
  char *s1;
  float junkFloat,thisFSF;
  int  junkInt,dataSize;
  int  numRead;
  int mrc_nxstart,mrc_nystart,mrc_nzstart,
      mrc_mx,mrc_my,mrc_mz,mrc_xlength,mrc_ylength,
      mrc_zlength,mrc_alpha,mrc_beta,mrc_gamma,
      mrc_mapc,mrc_mapr,mrc_maps,mrc_min,mrc_max,
      mrc_mean,mrc_std,mrc_ispg,mrc_nsymbt,mrc_xorigin,
      mrc_yorigin;
  int aoverbI, map_abangI, delaI, delbI, delcI;
  int dtheI, dphi_90I, tmpI;
  FILE *fIn;

  fIn = lun[*luNum];    /* a */

  /* We've got to account for larger data e.g. i4 versus byte */
  if(*pifMode == 46)
    dataSize = 4;
  else if(*pifMode==40) /* used for 4 byte 3D density map */
    dataSize = 4;
  else if(*pifMode==30) /* used for 4 byte 3D fft map */
                        /* we save the fft map using the same methods as mode=40 
                           by double the column size in user program*/
    dataSize = 4;
  else if(*pifMode==20) /* used for 2 byte 3D density map */
    dataSize = 2;
  else if(*pifMode==10) /* used for 2 byte 3D fft map */
                        /* we save the fft map using the same methods as mode=20 
                           by double the column size in user program*/
    dataSize = 2;
  else if(*pifMode == 1)
    dataSize = 2;
  else if(*pifMode == 7)
    dataSize = 2;
  else dataSize = 1;

  thisFSF = *recipFloatScaleFactor;
  if(*pifMode == 20)
    {/* We've got to get header values like pixelSize in greater precision */
      thisFSF = 1.0 / *int4scaleF;
    }/* We've got to get header values like pixelSize in greater precision */

  if((*htype)==1)  
    {/* all images have same size, we can fseek directly */
      
      /* We have a historical bug whereby old PIF files had the
       * numberOfImages variable stuffed here.  Fixed 3/97 */
      if((*pifMode)==6) 	/* No sections for packed boxes */
	(*numSecs) = (int)1;

      imageBytes =  (*numCols) * (*numRows) * (*numSecs) * dataSize;
      seekPos = 512 + ((*imgNum)-1) * (512 + imageBytes);
      
      fseek(fIn,seekPos,0);
      OKREAD(fread(&thisNcols, sizeof(int),1,fIn),1,"PIF_READ_DATA_HEADER");/*ROK*/
      fread(&thisNrows, sizeof(int),1,fIn);/*ROK*/
      fread(&thisNsecs, sizeof(int),1,fIn);/*ROK*/
      differentEndian_((BYTE *)&thisNcols, 4, endians[*luNum]);
      differentEndian_((BYTE *)&thisNrows, 4, endians[*luNum]);
      differentEndian_((BYTE *)&thisNsecs, 4, endians[*luNum]);

      /* We have a historical bug whereby old PIF files had the
       * numberOfImages variable stuffed here.  Fixed 3/97 */
      if((*pifMode)==6) 	/* No sections for packed boxes */
	thisNsecs = 1;

      /* Check that the size is what we expect (since htype==1) */
      if((*pifMode!=31) &&
	 (thisNcols != *numCols ||
	  thisNrows != *numRows ||
	  thisNsecs != *numSecs)  
	 )
	{
	  printf("Error in pif_read_data_header for image %d\n", *imgNum);
	  printf("  - htype=1 implies all images should have the same size\n");
	  printf("  - expected size is (%d,%d,%d)\n", 
		 *numCols, *numRows, *numSecs);
	  printf("  - found size is    (%d,%d,%d) \n", 
		 thisNcols, thisNrows, thisNsecs);
	  *imgOffset = seekPos + 512;  /* image starts after image header */
	  return(1);	/* error */
	}
    }/* all images have same size, we can fseek directly */
  else 
    {/* htype != 1, we must work through all the images */
      seekPos = 512;            /* global header */
      fseek(fIn,seekPos,0);
      
      for(i=0;i<(*imgNum);i++)
	{/* Seek the one we want */
	    fread(&thisNcols, sizeof(int),1,fIn);/*ROK*/
	    fread(&thisNrows, sizeof(int),1,fIn); /*ROK*/
	    fread(&thisNsecs, sizeof(int),1,fIn); /*ROK*/
	  differentEndian_((BYTE *)&thisNcols, 4, endians[*luNum]);
	  differentEndian_((BYTE *)&thisNrows, 4, endians[*luNum]);
	  differentEndian_((BYTE *)&thisNsecs, 4, endians[*luNum]);
	  
	  /* We have a historical bug whereby old PIF files had the
	   * numberOfImages variable stuffed here.  Fixed 3/97 */
	  if(*pifMode==6) /* No sections for packed boxes */
	    {/* Packed boxes */
	      thisNsecs = 1;
	    }/* Packed boxes */
	  
	  if((i+1) < *imgNum)  
	    {
	      seekPos += 512 + (thisNcols*thisNrows*thisNsecs*dataSize); 
	      /* Header + image*/
	      fseek(fIn,seekPos,0);
	    }
	}/* Seek the one we want */
      *numCols = thisNcols;
      *numRows = thisNrows;
      *numSecs = thisNsecs;
    }/* htype != 1, we must work through all the images */
  
  *imgOffset = seekPos + 512;  /* image starts after image header */
  
  if(fIn != NULL)
    {/* It is readable */
      
      /* [2] Read header for each image */
	fread(pifMode,sizeof(int),1,fIn); /*ROK*/
      differentEndian_((BYTE *)pifMode, 4, endians[*luNum]);
      fread(&localInt,sizeof(int),1,fIn); /*ROK*/       
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      if(*pifVersion >= 2.09)
	convertBackFloat_(localInt,img_bkg,thisFSF);
      else
	*img_bkg = (float)localInt;

      fread(packRadius,sizeof(int),1,fIn); /*ROK*/       
      differentEndian_((BYTE *)packRadius, 4, endians[*luNum]);

      fread(&mrc_nxstart,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&mrc_nxstart, 4, endians[*luNum]);
      fread(&mrc_nystart,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&mrc_nystart, 4, endians[*luNum]);
      fread(&mrc_nzstart,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&mrc_nzstart, 4, endians[*luNum]);
      fread(&mrc_mx,sizeof(int),1,fIn); /*ROK*/             
      differentEndian_((BYTE *)&mrc_mx, 4, endians[*luNum]);
      fread(&mrc_my,sizeof(int),1,fIn); /*ROK*/             
      differentEndian_((BYTE *)&mrc_my, 4, endians[*luNum]);
      fread(&mrc_mz,sizeof(int),1,fIn); /*ROK*/             
      differentEndian_((BYTE *)&mrc_mz, 4, endians[*luNum]);
      fread(&mrc_xlength,sizeof(int),1,fIn); /*ROK*/      
      differentEndian_((BYTE *)&mrc_xlength, 4, endians[*luNum]);
      fread(&mrc_ylength,sizeof(int),1,fIn); /*ROK*/      
      differentEndian_((BYTE *)&mrc_ylength, 4, endians[*luNum]);
      fread(&mrc_zlength,sizeof(int),1,fIn); /*ROK*/
      differentEndian_((BYTE *)&mrc_zlength, 4, endians[*luNum]);
      
      convertBackFloat_(mrc_xlength,a_cell,thisFSF);
      convertBackFloat_(mrc_ylength,b_cell,thisFSF);
      convertBackFloat_(mrc_zlength,c_cell,thisFSF);

      fread(&mrc_alpha,sizeof(int),1,fIn); /*ROK*/
      differentEndian_((BYTE *)&mrc_alpha, 4, endians[*luNum]);
      fread(&mrc_beta,sizeof(int),1,fIn);  /*ROK*/        
      differentEndian_((BYTE *)&mrc_beta, 4, endians[*luNum]);
      fread(&mrc_gamma,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&mrc_gamma, 4, endians[*luNum]);

      convertBackFloat_(mrc_alpha,alp,thisFSF);
      convertBackFloat_(mrc_beta,bet,thisFSF);
      convertBackFloat_(mrc_gamma,gam,thisFSF);

      fread(&mrc_mapc,sizeof(int),1,fIn); /*ROK*/           
      differentEndian_((BYTE *)&mrc_mapc, 4, endians[*luNum]);
      fread(&mrc_mapr,sizeof(int),1,fIn); /*ROK*/           
      differentEndian_((BYTE *)&mrc_mapr, 4, endians[*luNum]);
      fread(&mrc_maps,sizeof(int),1,fIn); /*ROK*/           
      differentEndian_((BYTE *)&mrc_maps, 4, endians[*luNum]);
      fread(&mrc_min,sizeof(int),1,fIn); /*ROK*/          
      differentEndian_((BYTE *)&mrc_min, 4, endians[*luNum]);
      fread(&mrc_max,sizeof(int),1,fIn); /*ROK*/          
      differentEndian_((BYTE *)&mrc_max, 4, endians[*luNum]);
      fread(&mrc_mean,sizeof(int),1,fIn); /*ROK*/         
      differentEndian_((BYTE *)&mrc_mean, 4, endians[*luNum]);
      fread(&mrc_std,sizeof(int),1,fIn); /*ROK*/      
      differentEndian_((BYTE *)&mrc_std, 4, endians[*luNum]);

      *img_min= mrc_min;
      *img_max= mrc_max;

      convertBackFloat_(mrc_mean,img_avg,thisFSF);
      convertBackFloat_(mrc_std,img_std,thisFSF);

      fread(&mrc_ispg,sizeof(int),1,fIn);      /*ROK*/         
      differentEndian_((BYTE *)&mrc_ispg, 4, endians[*luNum]);
      fread(&mrc_nsymbt,sizeof(int),1,fIn);    /*ROK*/        
      differentEndian_((BYTE *)&mrc_nsymbt, 4, endians[*luNum]);
      fread(&mrc_xorigin,sizeof(float),1,fIn); /*ROK*/     
      differentEndian_((BYTE *)&mrc_xorigin, 4, endians[*luNum]);
      fread(&mrc_yorigin,sizeof(float),1,fIn); /*ROK*/     
      differentEndian_((BYTE *)&mrc_yorigin, 4, endians[*luNum]);
      
      fread(descriptionStr,sizeof(char),80,fIn); /*ROK*/
      if(EQU(descriptionStr,""))
	{/* Init it out ... */
	  sprintf(descriptionStr,
"                                                                           ");
	}/* Init it out ... */
      fread(timeStr,sizeof(char),32,fIn);       /*ROK*/
      fread(micrographStr,sizeof(char),16,fIn); /*ROK*/
      fread(scanNumberStr,sizeof(char),8,fIn);  /*ROK*/
      
      fread(&aoverbI,sizeof(int),1,fIn);        /*ROK*/
      fread(&map_abangI,sizeof(int),1,fIn);     /*ROK*/    
      fread(&delaI,sizeof(int),1,fIn);          /*ROK*/ 
      fread(&delbI,sizeof(int),1,fIn);          /*ROK*/
      fread(&delcI,sizeof(int),1,fIn);          /*ROK*/
      differentEndian_((BYTE *)&aoverbI, 4, endians[*luNum]);
      differentEndian_((BYTE *)&map_abangI, 4, endians[*luNum]);
      differentEndian_((BYTE *)&delaI, 4, endians[*luNum]);
      differentEndian_((BYTE *)&delbI, 4, endians[*luNum]);
      differentEndian_((BYTE *)&delcI, 4, endians[*luNum]);

      fread(t_matrix,sizeof(int),6,fIn);        /*ROK*/
      for(i=0;i<6;i++)
	differentEndian_((BYTE *)&t_matrix[i], 4, endians[*luNum]);
      convertBackFloat_(aoverbI,aoverb,thisFSF);
      convertBackFloat_(map_abangI,map_abang,thisFSF);
      convertBackFloat_(delaI, dela,thisFSF);
      convertBackFloat_(delbI, delb,thisFSF);
      convertBackFloat_(delcI, delc,thisFSF);

      fread(&dtheI,sizeof(int),1,fIn);        /*ROK*/
      fread(&dphi_90I,sizeof(int),1,fIn);     /*ROK*/    
      differentEndian_((BYTE *)&dtheI, 4, endians[*luNum]);
      differentEndian_((BYTE *)&dphi_90I, 4, endians[*luNum]);
      convertBackFloat_(dtheI,dthe,thisFSF);
      convertBackFloat_(dphi_90I,dphi_90,thisFSF);

      fread(symmetry,sizeof(int),1,fIn);     /*ROK*/   
      differentEndian_((BYTE *)symmetry, 4, endians[*luNum]);
      fread(binFactor,sizeof(int),1,fIn);    /*ROK*/     
      differentEndian_((BYTE *)binFactor, 4, endians[*luNum]);

      fread(&tmpI,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&tmpI, 4, endians[*luNum]);
      convertBackFloat_(tmpI,a_star,thisFSF);
      fread(&tmpI,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&tmpI, 4, endians[*luNum]);
      convertBackFloat_(tmpI,b_star,thisFSF);
      fread(&tmpI,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&tmpI, 4, endians[*luNum]);
      convertBackFloat_(tmpI,c_star,thisFSF);
      fread(&tmpI,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&tmpI, 4, endians[*luNum]);
      convertBackFloat_(tmpI,alp_star,thisFSF);
      fread(&tmpI,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&tmpI, 4, endians[*luNum]);
      convertBackFloat_(tmpI,bet_star,thisFSF);
      fread(&tmpI,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&tmpI, 4, endians[*luNum]);
      convertBackFloat_(tmpI,gam_star,thisFSF);
      fread(&tmpI,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&tmpI, 4, endians[*luNum]);
      convertBackFloat_(tmpI,pixelSize,thisFSF);

      fread(pixSizUnits,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)pixSizUnits, 4, endians[*luNum]);
      fread(&tmpI,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&tmpI, 4, endians[*luNum]);
      convertBackFloat_(tmpI,res_out,thisFSF);
      fread(ctfMode,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)ctfMode, 4, endians[*luNum]);
      fread(&tmpI,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&tmpI, 4, endians[*luNum]);
      convertBackFloat_(tmpI,tempFactor,thisFSF);
      fread(&tmpI,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&tmpI, 4, endians[*luNum]);
      convertBackFloat_(tmpI,wiener,thisFSF);
      fread(borderType,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)borderType, 4, endians[*luNum]);
      fread(borderWidth,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)borderWidth, 4, endians[*luNum]);

      fread(&tmpI,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&tmpI, 4, endians[*luNum]);
      convertBackFloat_(tmpI,r4_min,thisFSF);
      fread(&tmpI,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)&tmpI, 4, endians[*luNum]);
      convertBackFloat_(tmpI,r4_max,thisFSF);

      fread(rad_lo,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)rad_lo, 4, endians[*luNum]);
      fread(rad_hi,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)rad_hi, 4, endians[*luNum]);

      fread(transX,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)transX, 4, endians[*luNum]);
      fread(transY,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)transY, 4, endians[*luNum]);
      fread(transZ,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)transZ, 4, endians[*luNum]);
      fread(isTrunc,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)isTrunc, 4, endians[*luNum]);
      fread(origX,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)origX, 4, endians[*luNum]);
      fread(origY,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)origY, 4, endians[*luNum]);
      fread(origZ,sizeof(int),1,fIn); /*ROK*/        
      differentEndian_((BYTE *)origZ, 4, endians[*luNum]);

    }/* It is readable */
  else
    {/* error */
	printf("[PIF_READ_DATA_HEADER] Image [%d] unreadable.\n",*imgNum);/*RSS*/
      return(1);	/* error */
    }/* error */

  return(0); 		/* no error	   */

}/* PIF_READ_DATA_HEADER */




/*************************************************************
 * pif_read_byte_image() - Read the PIF file's BYTE image for a 
 * specific image.
*/
void pif_read_byte_image_(int *luNum,
			 BYTE *byte_buffer,
			 int *imgOffset,
			 int *numCols, int *numRows, int *numSecs)
{/* PIF_READ_BYTE_IMAGE */
  FILE *fIn;
  int i;

  fIn= lun[*luNum];

  fseek(fIn,*imgOffset,0);
#define NORMAL_WAY 1
#ifdef NORMAL_WAY
  OKREAD(fread(byte_buffer,sizeof(BYTE), /*ROK*/
	       ((*numCols)*(*numRows)*(*numSecs)),fIn),
	 ((*numCols)*(*numRows)*(*numSecs)),
	 "PIF_READ_BYTE_IMAGE");
#else
  fread(byte_buffer,sizeof(BYTE), /*ROK*/
	       ((*numCols)*(*numRows)*(*numSecs)),fIn);
#endif

}/* PIF_READ_BYTE_IMAGE */




/*************************************************************
 * pif_read_float_image() - Read the PIF file's float image for a 
 * specific image and stuff it in the float array passed in.
 * JFC.3 27-May-97 Added error return (0=ok, else error).
*/
int pif_read_float_image_(int *luNum,
			  float *in_float_buffer,
			  int *imgOffset,
			  int *numCols, int *numRows, int *numSecs,
			  int *binSwitch,
			  float *recipScaleF)
{/* PIF_READ_FLOAT_IMAGE */
  FILE *fIn;
  int   i, j,k,l,m,n, totBuffer, binStep, binSection;
  int   iM1,i0,iP1,jM1,jP1,oddFactor;
  int   lastSec,lastRow,lastCol;
  int   totPnts,lastBinSec,binCols,binRows;
  float totVals,tmpVals[9],backGround;
  int   *int_buffer[2],*thisIntBuffer, *bBuf;
  float binVals[9] = {0.0625,0.125,0.0625,0.125,0.25,
			0.125,0.0625,0.125,0.0625};

  fIn= lun[*luNum];

  totBuffer = (*numCols)*(*numRows);
  
  if(*binSwitch==2)
    {/* They want to "sample" average */
      binStep = 2;
      binCols = (*numCols/2)+1;
      binRows = (*numRows/2)+1;
      oddFactor = 0;
      if((binCols % 2)==0)
	{/* Make odd */
	  oddFactor = 1;
	  binCols--;
	}/* Make odd */
      if((binRows % 2)==0)
	binRows--;
      binVals[0] = 0.0625;
      binVals[1] = 0.125;
      binVals[2] = 0.0625;
      binVals[3] = 0.125;
      binVals[4] = 0.25;
      binVals[5] = 0.125;
      binVals[6] = 0.0625;
      binVals[7] = 0.125;
      binVals[8] = 0.0625;
    }/* They want to "sample" average */
  else
    {/* They DON'T want to bin the data */
      binSection = totBuffer;
      binStep = 1;
      lastSec = *numSecs;
      lastRow = *numRows;
      lastCol = *numCols;
      lastBinSec = 1;
    }/* They DON'T want to bin the data */

  int_buffer[0] = (int *)malloc(sizeof(int)*totBuffer);
  if(int_buffer[0] == NULL)  {
    printf("ERROR in pif_read_float_image - can't allocate int_buffer\n");
    return(1);	/* error */
  }
  int_buffer[1] = (int *)malloc(sizeof(int)*totBuffer);
  if(int_buffer[1] == NULL)  {
    printf("ERROR in pif_read_float_image - can't allocate int_buffer\n");
    return(1);	/* error */
  }

  if(*binSwitch == 2)
    {/* DO the "sample" average */

      /* Seek the image - for 2D images we just do this once. */
      if(fseek(fIn,(*imgOffset),0) != 0)  
	{/* failed to find position */
	  perror("ERROR in pif_read_float_image fseek");
	  printf("ERROR seeking position %d\n", *imgOffset);
	  return(1);	/* error */
	}/* failed to find position */
      
      if(fread(int_buffer[0],sizeof(int), /*ROK*/
	       totBuffer, fIn) != totBuffer)
	{
	  perror("ERROR in pif_read_float_image fseek");
	  printf("ERROR reading float image\n");
	  return(1);	/* error */
	}
      /* Init out buffer */
      for(i=0;i<totBuffer;i++) {                            /*RSSSS*/
	  differentEndian_((BYTE *)&int_buffer[0][i],       /*RSSSS*/
			   4, endians[*luNum]);             /*RSSSS*/
      }                                                     /*RSSSS*/
      bBuf = int_buffer[0];        /* Just use this one */
      backGround = (float)bBuf[0] * (*recipScaleF);
      for(i=0;i<(binRows);i++)
	{/* Do all rows */
	  for(j=0;j<(binCols);j++)
	    {/* DO all cols */
	      in_float_buffer[((i*binCols)+j)] = (float)bBuf[0] *
		(*recipScaleF);
	    }/* DO all cols */
	}/* Do all rows */

      /* Now sample average */
      for(i=oddFactor;i<(*numRows-oddFactor);i+=binStep)
	{/* Do all rows */
	  if(i==0)
	    iM1 = (i)*(*numCols);
	  else iM1 = (i-1)*(*numCols);
	  i0  = (i)  *(*numCols);
	  if((i+1)==*numRows)
	    iP1 = (i)*(*numCols);
	  else iP1 = (i+1)*(*numCols);
	  for(j=oddFactor;j<(*numCols-oddFactor);j+=binStep)
	    {/* DO all cols */
	      if(j==0)
		jM1 = j;
	      else jM1 = j - 1;
	      jP1 = j + 1;
	      tmpVals[0] = bBuf[(iM1+jM1)]*(*recipScaleF);
	      tmpVals[1] = bBuf[(iM1+j)]*(*recipScaleF);
	      tmpVals[2] = bBuf[(iM1+jP1)]*(*recipScaleF);
	      tmpVals[3] = bBuf[(i0+jM1)]*(*recipScaleF);
	      tmpVals[4] = bBuf[(i0+j)]*(*recipScaleF);
	      tmpVals[5] = bBuf[(i0+jP1)]*(*recipScaleF);
	      tmpVals[6] = bBuf[(iP1+jM1)]*(*recipScaleF);
	      tmpVals[7] = bBuf[(iP1+j)]*(*recipScaleF);
	      tmpVals[8] = bBuf[(iP1+jP1)]*(*recipScaleF);
	      if((i)== 0)
		{/* Past rows */
		  tmpVals[0] = backGround;
		  tmpVals[1] = backGround;
		  tmpVals[2] = backGround;
		}/* Past rows */
	      if((i)>=(*numRows-1))
		{/* Past rows */
		  tmpVals[6] = backGround;
		  tmpVals[7] = backGround;
		  tmpVals[8] = backGround;
		}/* Past rows */
	      if((jM1)< 0)
		{/* Past cols */
		  tmpVals[0] = backGround;
		  tmpVals[3] = backGround;
		  tmpVals[6] = backGround;
		}/* Past rows */
	      if((jP1)>=*numCols)
		{/* Past cols */
		  tmpVals[2] = backGround;
		  tmpVals[5] = backGround;
		  tmpVals[8] = backGround;
		}/* Past rows */

	      totVals = ((float)tmpVals[0] * binVals[0]) +
		        ((float)tmpVals[1] * binVals[1]) +
		        ((float)tmpVals[2] * binVals[2]) +
		        ((float)tmpVals[3] * binVals[3]) +
		        ((float)tmpVals[4] * binVals[4]) +
		        ((float)tmpVals[5] * binVals[5]) +
		        ((float)tmpVals[6] * binVals[6]) +
		        ((float)tmpVals[7] * binVals[7]) +
		        ((float)tmpVals[8] * binVals[8]) ;

	      in_float_buffer[((i/binStep)*(binCols))+
			   (j/binStep)] =
		(float)totVals;
		

	    }/* DO all cols */
	  
	}/* Do all rows */

    }/* DO the "sample" average */
  else
    {/* Normal */
      for(i=0;i<(lastSec);i+= binStep)
	{/* Read all secs */
	  
	  /* Seek the image - for 2D images we just do this once and
	   * i==0 there. */
	  if(fseek(fIn,(*imgOffset+(i*totBuffer)),0) != 0)  
	    {/* failed to find position */
	      perror("ERROR in pif_read_float_image fseek");
	      printf("ERROR seeking position %d\n", *imgOffset);
	      return(1);	/* error */
	    }/* failed to find position */
	  
	  if(fread(int_buffer[0],sizeof(int), totBuffer, fIn) /*ROK*/
	     != totBuffer)
	    {
	      perror("ERROR in pif_read_float_image fseek");
	      printf("ERROR reading section %d\n", i+1);
	      return(1);	/* error */
	    }

	  for(k=0;k<totBuffer;k++) {                       /*RSSSS*/
	    differentEndian_((BYTE *)&int_buffer[0][k], 4, /*RSSSS*/
			     endians[*luNum]);             /*RSSSS*/
	  }                                                /*RSSSS*/

	  if(lastBinSec > 1)
	    {/* Bin it */
		if(fread(int_buffer[1],sizeof(int), /*ROK*/
			 totBuffer, fIn) != totBuffer)  
		{
		    perror("ERROR in pif_read_float_image fseek");
		    printf("ERROR reading section %d\n", i+1);
		    return(1);	/* error */
		}

		for(k=0;k<totBuffer;k++) {                       /*RSSSS*/
		  differentEndian_((BYTE *)&int_buffer[1][k], 4, /*RSSSS*/
				   endians[*luNum]);             /*RSSSS*/
		}                                                /*RSSSS*/

	    }/* Bin it */
	  for(j=0;j<(lastRow);j+= binStep)
	    {/* Rows */
	      for(k=0;k<(lastCol);k+= binStep)
		{/* Do all */
		  totVals = 0.0;
		  totPnts = 0;
		  
		  for(n=0;n<lastBinSec;n++)
		    {/* secs */
		      thisIntBuffer = int_buffer[n];
		      for(l=0;l<binStep;l++)
			for(m=0;m<binStep;m++)
			  {/* Do local neighbors */
			    totVals += 
			      (float)thisIntBuffer[(((j+l)*
						      (*numCols))+(k+m))] *
			      (*recipScaleF);
			    totPnts ++;
			  }/* Do local neighbors */
		    }/* secs */
		  
		  in_float_buffer[(((i/binStep)*binSection)+
				((j/binStep)*(*numCols/binStep))+
				(k/binStep))]= (float)(totVals / 
						       (float)totPnts);
		}/* Do all */
	    }/* Rows */
	  
	}/* Read all secs */
    }/* Normal */
    
  free(int_buffer[0]);
  free(int_buffer[1]);

  return(0);	/*  no error */

}/* PIF_READ_FLOAT_IMAGE */




/*************************************************************
 * pif_read_byteshort_image() - Read the PIF file's BYTE image for a 
 * specific image and stuff it in the short array passed in.
 * JFC.3 27-May-97 Added error return (0=ok, else error).
*/
int pif_read_byteshort_image_(int *luNum,
			      short *short_buffer,
			      int *imgOffset,
			      int *numCols, int *numRows, int *numSecs,
			      int *binSwitch)
{/* PIF_READ_BYTESHORT_IMAGE */
  FILE *fIn;
  int   i, j,k,l,m,n, totBuffer, binStep, binSection;
  int   iM1,i0,iP1,jM1,jP1,oddFactor;
  int   lastSec,lastRow,lastCol;
  int   totPnts,lastBinSec,binCols,binRows;
  float totVals,tmpVals[9],backGround;
  BYTE  *byte_buffer[2],*thisByteBuffer, *bBuf;
  float binVals[9] = {0.0625,0.125,0.0625,0.125,0.25,
			0.125,0.0625,0.125,0.0625};

  fIn= lun[*luNum];

  totBuffer = (*numCols)*(*numRows);
  
  if(*binSwitch==2)
    {/* They want to "sample" average */
      binStep = 2;
      binCols = (*numCols/2)+1;
      binRows = (*numRows/2)+1;
      oddFactor = 0;
      if((binCols % 2)==0)
	{/* Make odd */
	  oddFactor = 1;
	  binCols--;
	}/* Make odd */
      if((binRows % 2)==0)
	binRows--;
      binVals[0] = 0.0625;
      binVals[1] = 0.125;
      binVals[2] = 0.0625;
      binVals[3] = 0.125;
      binVals[4] = 0.25;
      binVals[5] = 0.125;
      binVals[6] = 0.0625;
      binVals[7] = 0.125;
      binVals[8] = 0.0625;
    }/* They want to "sample" average */
  else
    {/* They DON'T want to bin the data */
      binSection = totBuffer;
      binStep = 1;
      lastSec = *numSecs;
      lastRow = *numRows;
      lastCol = *numCols;
      lastBinSec = 1;
    }/* They DON'T want to bin the data */

  byte_buffer[0] = (BYTE *)malloc(sizeof(BYTE)*totBuffer);
  if(byte_buffer[0] == NULL)  {
    printf("ERROR in pif_read_byteShort_image - can't allocate byte_buffer\n");
    return(1);	/* error */
  }
  byte_buffer[1] = (BYTE *)malloc(sizeof(BYTE)*totBuffer);
  if(byte_buffer[1] == NULL)  {
    printf("ERROR in pif_read_byteShort_image - can't allocate byte_buffer\n");
    return(1);	/* error */
  }


  if(*binSwitch == 2)
    {/* DO the "sample" average */

      /* Seek the image - for 2D images we just do this once. */
      if(fseek(fIn,(*imgOffset),0) != 0)  
	{/* failed to find position */
	  perror("ERROR in pif_read_byteShort_image fseek");
	  printf("ERROR seeking position %d\n", *imgOffset);
	  return(1);	/* error */
	}/* failed to find position */
      
      if(fread(byte_buffer[0],sizeof(BYTE), /*ROK*/
	       totBuffer, fIn) != totBuffer)
	{
	  perror("ERROR in pif_read_byteShort_image fseek");
	  printf("ERROR reading byteshort image\n");
	  return(1);	/* error */
	}
      /* Init out buffer */
      bBuf = byte_buffer[0];        /* Just use this one */
      backGround = (float)bBuf[0];
      for(i=0;i<(binRows);i++)
	{/* Do all rows */
	  for(j=0;j<(binCols);j++)
	    {/* DO all cols */
	      short_buffer[((i*binCols)+j)] = (short)bBuf[0];
	    }/* DO all cols */
	}/* Do all rows */

      /* Now sample average */
      for(i=oddFactor;i<(*numRows-oddFactor);i+=binStep)
	{/* Do all rows */
	  if(i==0)
	    iM1 = (i)*(*numCols);
	  else iM1 = (i-1)*(*numCols);
	  i0  = (i)  *(*numCols);
	  if((i+1)==*numRows)
	    iP1 = (i)*(*numCols);
	  else iP1 = (i+1)*(*numCols);
	  for(j=oddFactor;j<(*numCols-oddFactor);j+=binStep)
	    {/* DO all cols */
	      if(j==0)
		jM1 = j;
	      else jM1 = j - 1;
	      jP1 = j + 1;
	      tmpVals[0] = bBuf[(iM1+jM1)];
	      tmpVals[1] = bBuf[(iM1+j)];
	      tmpVals[2] = bBuf[(iM1+jP1)];
	      tmpVals[3] = bBuf[(i0+jM1)];
	      tmpVals[4] = bBuf[(i0+j)];
	      tmpVals[5] = bBuf[(i0+jP1)];
	      tmpVals[6] = bBuf[(iP1+jM1)];
	      tmpVals[7] = bBuf[(iP1+j)];
	      tmpVals[8] = bBuf[(iP1+jP1)];
	      if((i)== 0)
		{/* Past rows */
		  tmpVals[0] = backGround;
		  tmpVals[1] = backGround;
		  tmpVals[2] = backGround;
		}/* Past rows */
	      if((i)>=(*numRows-1))
		{/* Past rows */
		  tmpVals[6] = backGround;
		  tmpVals[7] = backGround;
		  tmpVals[8] = backGround;
		}/* Past rows */
	      if((jM1)< 0)
		{/* Past cols */
		  tmpVals[0] = backGround;
		  tmpVals[3] = backGround;
		  tmpVals[6] = backGround;
		}/* Past rows */
	      if((jP1)>=*numCols)
		{/* Past cols */
		  tmpVals[2] = backGround;
		  tmpVals[5] = backGround;
		  tmpVals[8] = backGround;
		}/* Past rows */

	      totVals = ((float)tmpVals[0] * binVals[0]) +
		        ((float)tmpVals[1] * binVals[1]) +
		        ((float)tmpVals[2] * binVals[2]) +
		        ((float)tmpVals[3] * binVals[3]) +
		        ((float)tmpVals[4] * binVals[4]) +
		        ((float)tmpVals[5] * binVals[5]) +
		        ((float)tmpVals[6] * binVals[6]) +
		        ((float)tmpVals[7] * binVals[7]) +
		        ((float)tmpVals[8] * binVals[8]) ;

	      short_buffer[((i/binStep)*(binCols))+
			   (j/binStep)] =
		(short)totVals;
	    }/* DO all cols */
	  
	}/* Do all rows */

    }/* DO the "sample" average */
  else
    {/* Normal */
      for(i=0;i<(lastSec);i+= binStep)
	{/* Read all secs */
	  
	  /* Seek the image - for 2D images we just do this once and
	   * i==0 there. */
	  if(fseek(fIn,(*imgOffset+(i*totBuffer)),0) != 0)  
	    {/* failed to find position */
	      perror("ERROR in pif_read_byteShort_image fseek");
	      printf("ERROR seeking position %d\n", *imgOffset);
	      return(1);	/* error */
	    }/* failed to find position */
	  
	  if(fread(byte_buffer[0],sizeof(BYTE), totBuffer, fIn) /*ROK*/
	     != totBuffer)  
	    {
	      perror("ERROR in pif_read_byteShort_image fseek");
	      printf("ERROR reading section %d\n", i+1);
	      return(1);	/* error */
	    }
	  if(lastBinSec > 1)
	    {/* Bin it */
		if(fread(byte_buffer[1],sizeof(BYTE), /*ROK*/
		       totBuffer, fIn) != totBuffer)  
		{
		  perror("ERROR in pif_read_byteShort_image fseek");
		  printf("ERROR reading section %d\n", i+1);
		  return(1);	/* error */
		}
	    }/* Bin it */
	  for(j=0;j<(lastRow);j+= binStep)
	    {/* Rows */
	      for(k=0;k<(lastCol);k+= binStep)
		{/* Do all */
		  totVals = 0.0;
		  totPnts = 0;
		  
		  for(n=0;n<lastBinSec;n++)
		    {/* secs */
		      thisByteBuffer = byte_buffer[n];
		      for(l=0;l<binStep;l++)
			for(m=0;m<binStep;m++)
			  {/* Do local neighbors */
			    totVals += 
			      (float)thisByteBuffer[(((j+l)*
						      (*numCols))+(k+m))];
			    totPnts ++;
			  }/* Do local neighbors */
		    }/* secs */
		  
		  short_buffer[(((i/binStep)*binSection)+
				((j/binStep)*(*numCols/binStep))+
				(k/binStep))]= (short)(totVals / 
						       (float)totPnts);
		}/* Do all */
	    }/* Rows */
	  
	}/* Read all secs */
    }/* Normal */
    
  free(byte_buffer[0]);
  free(byte_buffer[1]);

  return(0);	/*  no error */

}/* PIF_READ_BYTESHORT_IMAGE */





/*************************************************************
 * pif_read_short_image() - Read the PIF file's SHORT image for a 
 * specific image.
*/
void pif_read_short_image_(int *luNum,
			   unsigned short *short_buffer,
			   int *imgOffset,
			   int *numCols, int *numRows, int *numSecs)
{/* PIF_READ_SHORT_IMAGE */
  int i,j,k;
  FILE *fIn;

  fIn = lun[*luNum];

  fseek(fIn,*imgOffset,0);
  OKREAD(fread(short_buffer,sizeof(unsigned short), /*ROK*/
	       ((*numCols)*(*numRows)*(*numSecs)),
	       fIn),
	 ((*numCols)*(*numRows)*(*numSecs)),
	 "PIF_READ_SHORT_IMAGE");
  for(i=0;i<(*numSecs);i++)
    for(j=0;j<(*numRows);j++)
      for(k=0;k<(*numCols);k++)
	differentEndian_((BYTE *)&short_buffer[((i*(*numRows)*(*numCols))+
					       (j*(*numCols))+k)], 2, endians[*luNum]);

}/* PIF_READ_SHORT_IMAGE */




void pif_read_imgi2(forPifHead_t *thisHead,
		    unsigned short *short_buffer,
		    int * ncol,
		    int * nrow,
		    int * thisImg)
{
  pif_read_imgi2_(thisHead, short_buffer, ncol, nrow, thisImg);
}

/***************************************************************
 * pif_read_imgi2_() - Read the PIF file's SHORT image 
 * for a floating point MAP.
*/
void pif_read_imgi2_(forPifHead_t *thisHead,
		     unsigned short *short_buffer,
		     int * ncol,
		     int * nrow,
		     int * thisImg
		     )
{/* PIF_READ_IMGI2 */
  int tmpThisImg, tmpSec2, numSecs, binCol, binRow, thisBin;

  if(thisHead->recipFloatScaleFactor == 0.0)
    thisHead->recipFloatScaleFactor =  1.0 /  thisHead->floatScaleFactor ;
  if(thisHead->floatScaleFactor == 0.0)
    thisHead->floatScaleFactor =  1.0 /  thisHead->recipFloatScaleFactor ;

  tmpThisImg = *thisImg;

  if(tmpThisImg > thisHead->nimg ||
     tmpThisImg < 1 )
    {/* Fatal error */
      printf("[PIF_READ_IMGI2() Invalid image [%d]\n",
	     tmpThisImg);
      exit(0);
    }/* Fatal error */

  numSecs = 1;
  
  binCol = thisHead->numCols;
  binRow = thisHead->numRows;

  if(*ncol == thisHead->numCols &&
     *nrow == thisHead->numRows )
    thisBin = 1;
  else thisBin = 2;

  if(thisHead->binFactor == 2)
    {/* It's already binned on disk */
      binRow  = (thisHead->numRows)/2+1;
      binCol  = (thisHead->numCols)/2+1;
      if((binRow%2)==0)
	binRow--;
      if((binCol%2)==0)
	binCol--;
      if(thisHead->binFactor == 2 &&
	 thisBin == 1)
	{/* Error */
	  printf("***********************************************\n");
	  printf("PIF_READ_IMGI2()\n");
	  printf("Can't expand BIN! DiskBIN_STORE==[%d] You want[%d]\n",
		 thisHead->binFactor,
		 thisBin);
	  printf("***********************************************\n");
	  exit(0);
	}/* Error */

      /* We've already taken care of thisBin with binCol/binRow etc... */
      thisBin = 1;

    }/* It's already binned on disk */

  /* Times 1 because we're BYTE here */
  thisHead->imgOffset = 1024 + (512 * (tmpThisImg-1)) +
    (int)((tmpThisImg - 1) * (binCol * binRow * 1));


  pif_read_short_image_(&(thisHead->luNum),
			short_buffer,
			&(thisHead->imgOffset),
			&binCol,
			&binRow,
			&numSecs
			);

}/* PIF_READ_IMGI2 */





void pif_read_imgi1(forPifHead_t *thisHead,
		    short *short_buffer,
		    int * ncol,
		    int * nrow,
		    int * thisImg)
{
  pif_read_imgi1_(thisHead, short_buffer, ncol, nrow, thisImg);
}

/***************************************************************
 * pif_read_imgi1_() - Read the PIF file's SHORT image 
 * for a floating point MAP.
*/
void pif_read_imgi1_(forPifHead_t *thisHead,
		     short *short_buffer,
		     int * ncol,
		     int * nrow,
		     int * thisImg
		     )
{/* PIF_READ_IMGI1 */
  int tmpThisImg, tmpSec2, numSecs, binCol, binRow, thisBin;

  if(thisHead->recipFloatScaleFactor == 0.0)
    thisHead->recipFloatScaleFactor =  1.0 /  thisHead->floatScaleFactor ;
  if(thisHead->floatScaleFactor == 0.0)
    thisHead->floatScaleFactor =  1.0 /  thisHead->recipFloatScaleFactor ;

  tmpThisImg = *thisImg;

  if(tmpThisImg > thisHead->nimg ||
     tmpThisImg < 1 )
    {/* Fatal error */
      printf("[PIF_READ_IMGI1() Invalid image [%d]\n",
	     tmpThisImg);
      exit(0);
    }/* Fatal error */

  numSecs = 1;
  
  binCol = thisHead->numCols;
  binRow = thisHead->numRows;

  if(*ncol == thisHead->numCols &&
     *nrow == thisHead->numRows )
    thisBin = 1;
  else thisBin = 2;

  if(thisHead->binFactor == 2)
    {/* It's already binned on disk */
      binRow  = (thisHead->numRows)/2+1;
      binCol  = (thisHead->numCols)/2+1;
      if((binRow%2)==0)
	binRow--;
      if((binCol%2)==0)
	binCol--;
      if(thisHead->binFactor == 2 &&
	 thisBin == 1)
	{/* Error */
	  printf("***********************************************\n");
	  printf("PIF_READ_IMGI1()\n");
	  printf("Can't expand BIN! DiskBIN_STORE==[%d] You want[%d]\n",
		 thisHead->binFactor,
		 thisBin);
	  printf("***********************************************\n");
	  exit(0);
	}/* Error */

      /* We've already taken care of thisBin with binCol/binRow etc... */
      thisBin = 1;

    }/* It's already binned on disk */

  /* Times 1 because we're BYTE here */
  thisHead->imgOffset = 1024 + (512 * (tmpThisImg-1)) +
    (int)((tmpThisImg - 1) * (binCol * binRow * 1));


  pif_read_byteshort_image_(&(thisHead->luNum),
			    short_buffer,
			    &(thisHead->imgOffset),
			    &(thisHead->numCols),
			    &(thisHead->numRows),
			    &numSecs,
			    &thisBin);

}/* PIF_READ_IMGI1 */





void pif_read_mapi2(forPifHead_t *thisHead,
		    float *float_buffer,
		    int * sec1,
		    int * sec2,
		    int * ncol,
		    int * nrow,
		    int * nsec)
{
  pif_read_mapi2_(thisHead, float_buffer, sec1, sec2, ncol, nrow, nsec);
}
void pif_read_wmapi2(forPifHead_t *thisHead,
		     float *float_buffer,
		     int *ncol,
		     int *nrow,
		     int *nsec)
{
  pif_read_wmapi2_(thisHead, float_buffer, ncol, nrow, nsec);
}


/*******************************************************************
 * pif_read_wmapi2() - Read entire Whole mapi2 into buffer.
*/
void pif_read_wmapi2_(forPifHead_t *thisHead,
		     float *float_buffer,
		     int *ncol,
		     int *nrow,
		     int *nsec
		     )
{/* PIF_READ_WMAPI2 */
  int startSec,endSec,i,j,k;
  startSec = 1;
  endSec = thisHead->numSecs;
  if(thisHead->binFactor==2)
    {/* It's binned down */
      endSec  = (thisHead->numSecs)/2+1;
      if((endSec%2)==0)
	endSec--;
    }/* It's binned down */
  pif_read_mapi2_(thisHead,
		  float_buffer,
		  &startSec,
		  &endSec,
		  ncol,nrow,nsec
		  );
}/* PIF_READ_WMAPI2 */



/***************************************************************
 * pif_read_mapi2_() - Read the PIF file's SHORT image 
 * for a floating point MAP.
*/
void pif_read_mapi2_(forPifHead_t *thisHead,
		     float *float_buffer,
		     int * sec1,
		     int * sec2,
		     int * ncol,
		     int * nrow,
		     int * nsec
		     )
{/* PIF_READ_MAPI2 */
  int tmpSec1, tmpSec2, numSecs;
  int binCol,binRow,binSec,thisBin;
  int strtCol,strtRow,strtSec;
  int endCol,endRow,endSec;

  if(thisHead->binFactor < 1 ||
     thisHead->binFactor > 2)
    thisHead->binFactor = 1;

  if(thisHead->recipFloatScaleFactor == 0.0)
    thisHead->recipFloatScaleFactor =  1.0 /  thisHead->floatScaleFactor ;
  if(thisHead->floatScaleFactor == 0.0)
    thisHead->floatScaleFactor =  1.0 /  thisHead->recipFloatScaleFactor ;

  if(*sec1 > *sec2)
    {/* Switch them */
      tmpSec1 = *sec2;
      tmpSec2 = *sec1;
    }/* Switch them */
  else
    {/* Normal */
      tmpSec1 = *sec1;
      tmpSec2 = *sec2;      
    }/* Normal */

  if(tmpSec1 > thisHead->numSecs ||
     tmpSec2 > thisHead->numSecs ||
     tmpSec1 < 1 ||
     tmpSec2 < 1 )
    {/* Fatal error */
      printf("[PIF_READ_MAPI2() Invalid section range [%d][%d]\n",
	     tmpSec1,tmpSec2);
      exit(0);
    }/* Fatal error */

  numSecs = tmpSec2 - tmpSec1;
  
  numSecs++;  /* Always get the last one */

  binCol = thisHead->numCols;
  binRow = thisHead->numRows;
  binSec = thisHead->numSecs;

  if((*ncol == binCol ||
     *nrow == binRow ||
     *nsec == binSec )
     )
    thisBin = 1;
  else thisBin = 2;

  if(thisHead->binFactor==2)
    {/* They want to sample average bin the data */
      binRow  = (thisHead->numRows)/2+1;
      binCol  = (thisHead->numCols)/2+1;
      binSec  = (thisHead->numSecs)/2+1;
      if((binRow%2)==0)
	binRow--;
      if((binCol%2)==0)
	binCol--;
      if((binSec%2)==0)
	binSec--;
      if(thisHead->mode==22)
	{/* PFT file is special */
	  binCol  = (thisHead->numCols)/2;
	  binRow  = (thisHead->numRows)/2;
	}/* PFT file is special */
      if(thisHead->binFactor == 2 &&
	 thisBin == 1)
	{/* Error */
	  printf("***********************************************\n");
	  printf("PIF_READ_MAPI2()\n");
	  printf("Can't expand BIN! DiskBIN_STORE==[%d] You want[%d]\n",
		 thisHead->binFactor,
		 thisBin);
	  printf("***********************************************\n");
	  exit(0);
	}/* Error */

      /* We've already taken care of thisBin with binCol/binRow etc... */
      thisBin = 1;

    }/* They want to sample average bin the data */


  /* Times 2 because we're short here */
  thisHead->imgOffset = 1024 +
    (int)((tmpSec1-1) * (binCol * binRow * 2));

  /*
  printf("Calling pif_rmshort[%d][%d][%d]  bin_store[%d]thisBin[%d]\n",
       binCol,binRow,numSecs,thisHead->binFactor,thisBin);
       */

  strtCol = strtRow = strtSec = 1;
  endCol = binCol;
  endRow = binRow;
  endSec = numSecs;
  pif_read_map_short_image_(&(thisHead->luNum),
			    float_buffer,
			    &(thisHead->imgOffset),
			    &binCol,
			    &binRow,
			    &numSecs,
			    &(thisHead->recipFloatScaleFactor),
			    &(thisHead->datar4_min),
			    &(thisHead->datar4_max),
			    &thisBin,
			    &strtCol, &strtRow, &strtSec,
			    &endCol,&endRow,&endSec);


}/* PIF_READ_MAPI2 */





void pif_read_tmapi2(forPifHead_t *thisHead,
		     float *float_buffer,
		     int * ncol,
		     int * nrow,
		     int * nsec,
		     int * col1,
		     int * row1,
		     int * sec1 )
{
  pif_read_tmapi2_(thisHead, float_buffer, ncol, nrow, nsec,
		   col1, row1, sec1);
}
/**************************************************************************
 * pif_read_tmapi2_() - Read the PIF file's SHORT image for a floating 
 * point MAP and truncate it accordingly.
 * Args are:
 *     
 *     forPifHead_t *thisHead,    PIF header
 *     float *float_buffer,       The real*4 buffer that will hold the 
 *                                truncated MAP
 *     int * ncol,                The dimensions of float_buffer
 *     int * nrow,
 *     int * nsec,
 *     int * col1,                What col/row/sec to START with from the disk
 *     int * row1,
 *     int * sec1
 *
 *     
 *     
*/
void pif_read_tmapi2_(forPifHead_t *thisHead,
		      float *float_buffer,
		      int * ncol,
		      int * nrow,
		      int * nsec,
		      int * col1,
		      int * row1,
		      int * sec1
		      )
{/* PIF_READ_TMAPI2 */
  int tmpSec1, tmpSec2, numSecs;
  int binCol,binRow,binSec,thisBin;

  if(thisHead->binFactor < 1 ||
     thisHead->binFactor > 2)
    thisHead->binFactor = 1;

  if(thisHead->recipFloatScaleFactor == 0.0)
    thisHead->recipFloatScaleFactor =  1.0 /  thisHead->floatScaleFactor ;
  if(thisHead->floatScaleFactor == 0.0)
    thisHead->floatScaleFactor =  1.0 /  thisHead->recipFloatScaleFactor ;

  if(*sec1 > thisHead->numSecs ||
     *row1 > thisHead->numRows ||
     *col1 > thisHead->numCols )
    {/* Fatal error */
      printf("[PIF_READ_TMAPI2() Invalid range [%d][%d][%d]\n",
	     *col1,*row1,*sec1);/*RSS*/
      exit(0);
    }/* Fatal error */

  binCol = thisHead->numCols;
  binRow = thisHead->numRows;
  binSec = thisHead->numSecs;

  thisBin = 1;

  if(thisHead->binFactor==2)
    {/* They want to sample average bin the data */
      binRow  = (thisHead->numRows)/2+1;
      binCol  = (thisHead->numCols)/2+1;
      binSec  = (thisHead->numSecs)/2+1;
      if((binRow%2)==0)
	binRow--;
      if((binCol%2)==0)
	binCol--;
      if((binSec%2)==0)
	binSec--;
      if(thisHead->mode==22)
	{/* PFT file is special */
	  binCol  = (thisHead->numCols)/2;
	  binRow  = (thisHead->numRows)/2;
	}/* PFT file is special */
      if(thisHead->binFactor == 2 &&
	 thisBin == 1)
	{/* Error */
	  printf("***********************************************\n");
	  printf("PIF_READ_MAPI2()\n");
	  printf("Can't expand BIN! DiskBIN_STORE==[%d] You want[%d]\n",
		 thisHead->binFactor,
		 thisBin);
	  printf("***********************************************\n");
	  exit(0);
	}/* Error */

      /* We've already taken care of thisBin with binCol/binRow etc... */
      thisBin = 1;

    }/* They want to sample average bin the data */


  thisHead->imgOffset = 1024 ;

  pif_read_map_short_image_(&(thisHead->luNum),
			    float_buffer,
			    &(thisHead->imgOffset),
			    &binCol,
			    &binRow,
			    &binSec,
			    &(thisHead->recipFloatScaleFactor),
			    &(thisHead->datar4_min),
			    &(thisHead->datar4_max),
			    &thisBin,
			    col1,
			    row1,
			    sec1,
			    ncol,
			    nrow,
			    nsec);


}/* PIF_READ_TMAPI2 */





/****************************************************************
 * pif_read_seci2() - Read just one section of floatInt*2 map.
*/
void pif_read_seci2(forPifHead_t *thisHead,
		    float *float_buffer,
		    int * ncol,
		    int * nrow,
		    int * sec1
		    )
{/* PIF_READ_SECI2 */
  int sec2,nsec;
  sec2 = *sec1;
  nsec = *ncol;
  pif_read_mapi2_(thisHead, float_buffer, sec1, &sec2, ncol, nrow, &nsec);
}/* PIF_READ_SECI2 */

void pif_read_seci2_(forPifHead_t *thisHead,
		     float *float_buffer,
		     int * ncol,
		     int * nrow,
		     int * sec1)
{
  pif_read_seci2(thisHead, float_buffer, ncol, nrow, sec1);
}

/****************************************************************
 * pif_read_seci4() - Read just one section of floatInt*4 map.
*/
void pif_read_seci4(forPifHead_t *thisHead,
		     float *float_buffer,
		     int * ncol,
		     int * nrow,
		     int * sec1)
{
  int sec2,nsec;
  sec2 = *sec1;
  nsec = *ncol;
  pif_read_mapi4_(thisHead, float_buffer, sec1, &sec2, ncol, nrow, &nsec);
}

void pif_read_seci4_(forPifHead_t *thisHead,
		     float *float_buffer,
		     int * ncol,
		     int * nrow,
		     int * sec1)
{
  pif_read_seci4(thisHead, float_buffer, ncol, nrow, sec1);
}

void pif_read_wmapi4(forPifHead_t *thisHead,
		     float *float_buffer,
		     int * ncol,
		     int * nrow,
		     int * nsec)
{
  pif_read_wmapi4_(thisHead, float_buffer, ncol, nrow, nsec);
}

void pif_read_wmapi4_(forPifHead_t *thisHead,
		     float *float_buffer,
		     int * ncol,
		     int * nrow,
		     int * nsec)
{
  int sec1,sec2;
  sec1 = 1;
  sec2 = *nsec;
  pif_read_mapi4_(thisHead, float_buffer, &sec1, &sec2, ncol, nrow, nsec);
}

/***************************************************************
 * pif_read_mapi4_() - Read the PIF file's INT image 
 * for a floating point MAP.
*/
void pif_read_mapi4(forPifHead_t *thisHead,
                     float *float_buffer,
                     int *sec1,
                     int *sec2,
                     int *ncol,
                     int *nrow,
                     int *nsec
                     )
{
   pif_read_mapi4_(thisHead, float_buffer, sec1, sec2, ncol, nrow,  
nsec);
}

void pif_read_mapi4_(forPifHead_t *thisHead,
		     float *float_buffer,
		     int *sec1,
		     int *sec2,
		     int *ncol,
		     int *nrow,
		     int *nsec
		     )
{/* PIF_READ_MAPI4 */
  int tmpSec1, tmpSec2, numSecs,
      binCol, binRow, binSec, thisBin;

  if(thisHead->recipFloatScaleFactor == 0.0)
    thisHead->recipFloatScaleFactor =  1.0 /  thisHead->floatScaleFactor ;
  if(thisHead->floatScaleFactor == 0.0)
    thisHead->floatScaleFactor =  1.0 /  thisHead->recipFloatScaleFactor ;

  if(*sec1 > *sec2)
    {/* Switch them */
      tmpSec1 = *sec2;
      tmpSec2 = *sec1;
    }/* Switch them */
  else
    {/* Normal */
      tmpSec1 = *sec1;
      tmpSec2 = *sec2;      
    }/* Normal */

  if(tmpSec1 > thisHead->numSecs ||
     tmpSec2 > thisHead->numSecs ||
     tmpSec1 < 1 ||
     tmpSec2 < 1 )
    {/* Fatal error */
      printf("[PIF_READ_WMAPI4() Invalid section range [%d][%d]\n",
	     tmpSec1,tmpSec2);
      exit(0);
    }/* Fatal error */

  numSecs = tmpSec2 - tmpSec1;
  
  binCol = thisHead->numCols;
  binRow = thisHead->numRows;
  binSec = thisHead->numSecs;

  if(*ncol == binCol &&
     *nrow == binRow)
    thisBin = 1;
  else thisBin = 2;

  if((thisBin != thisHead->binFactor) &&
     (thisHead->mode==21 || thisHead->mode==22))
    {/* Error */
      printf ("PIF_READ_WMAPI4()  - PRJS/PFTS data\n");
      printf ("Incompatible bin factors BIN_STORED[%d] BIN_EXEC[%d]\n",
	      thisHead->binFactor,thisBin);
      exit(0);
    }/* Error */

  if(thisHead->binFactor==2)
    {/* They want to sample average bin the data */
      binRow  = (thisHead->numRows)/2+1;
      binCol  = (thisHead->numCols)/2+1;
      binSec  = (thisHead->numSecs)/2+1;
      if((binRow%2)==0)
	binRow--;
      if((binCol%2)==0)
	binCol--;
      if((binSec%2)==0)
	binSec--;
      if(thisHead->mode==22)
	{/* PFT file is special */
	  binCol  = (thisHead->numCols)/2;
	  binRow  = (thisHead->numRows)/2;
	}/* PFT file is special */
      if(thisHead->mode==22 ||thisHead->mode==21 )
	{/* PFT/PRJ file is special */
	  binSec  = (thisHead->numSecs);
	}/* PFT/PRJ file is special */

      if(thisBin == 1)
	{/* Error */
	  printf("***********************************************\n");
	  printf("PIF_READ_WMAPI4()\n");
	  printf("Can't expand BIN! DiskBIN_STORE==[%d] You want[%d]\n",
		 thisHead->binFactor,
		 thisBin);
	  printf("***********************************************\n");
	  exit(0);
	}/* Error */

      /* We've already taken into account
       * above what the binCol/binRow are etc... */
      thisBin = 1;

    }/* They want to sample average bin the data */


  /* Times 4 because we're INT here */
  thisHead->imgOffset = 1024 +
    (int)((tmpSec1-1) * (binCol * binRow * 4));

  numSecs++;  /* Always get the last one */


  /*
  printf("\n\nsec1[%d] sec2[%d] convtsb[%d] thisBin[%d]tH->bF[%d]\n",
	 tmpSec1, tmpSec2,convTsb,thisBin,thisHead->binFactor);
  printf("lunum[%d] offset[%d],numSecs[%d]\n",
	 thisHead->luNum,thisHead->imgOffset,numSecs);
  printf("ncol[%d]nrow[%d]\n",thisHead->numCols,thisHead->numRows);
  printf("binCol[%d]binRow[%d]binSec[%d]\n",binCol,binRow,binSec);
  printf("Before call float_buff[%d]\n",float_buffer);
  */

  pif_read_map_int_image_(&(thisHead->luNum),
			  float_buffer,
			  &(thisHead->imgOffset),
			  &binCol,
			  &binRow,
#ifdef OLDWAY
			  &binSec,
#else
			  &numSecs,
#endif
			  &(thisHead->recipFloatScaleFactor),
			  &(thisHead->datar4_min),
			  &(thisHead->datar4_max),
			  &thisBin,
			  &(thisHead->mode)
			  );

}/* PIF_READ_MAPI4 */



/***************************************************************
 * pif_read_map_short_image() - Read the PIF file's SHORT image 
 * for a floating point MAP.
*/
void pif_read_map_short_image_(int *luNum,
			       float *float_buffer,
			       int *imgOffset,
			       int *numCols, int *numRows, int *numSecs,
			       float *recipFloatScaleFactor,
			       float *minMapPixel, float *maxMapPixel,
			       int *binSwitch,
			       int *strtCol,
			       int *strtRow,
			       int *strtSec,
			       int *ncol,
			       int *nrow,
			       int *nsec)
{/* PIF_READ_MAP_SHORT_IMAGE */
  int i,j,k,l,m,n,rowCol,rowColBin;
  short *short_buffer[4];         /* Hard code this for only 2x2 binning now */
  short *sec1,*sec2,*sec3;
  short *thisShortBuffer;
  float thisFloat,totFloats;
  FILE *fIn;
  int totPnts, thisOffset,binCol,binRow,binSec,oddFactor;
  int binStep,secOut,rowOut,colOut,lastSec,lastRow,lastCol;
  int jM1,j0,jP1,kM1,kP1;
  float binVals[27] = {0.015625, 
			 0.03125, 
			 0.015625, 
			 0.03125, 
			 0.0625,   
			 0.03125,
			 0.015625, 
			 0.03125, 
			 0.015625, 
			 0.03125, 
			 0.0625,   
			 0.03125,
			 0.0625,   
			 0.125,
			 0.0625,   
			 0.03125,
			 0.0625,   
			 0.03125,
			 0.015625, 
			 0.03125, 
			 0.015625, 
			 0.03125, 
			 0.0625,   
			 0.03125,
			 0.015625, 
			 0.03125, 
			 0.015625 };
  float tmpVals[27];
		       
  if(*binSwitch > 2)
    {/* Error */
      printf("PIF_READ_MAP_SHORT_IMAGE - Cannot bin data like this[%d]\n",
	     *binSwitch);
      exit(0);
    }/* Error */
  else if(*binSwitch==2)
    {/* They want to sample average bin the data */
      binStep = 2;
      lastSec = *numSecs - 2;
      lastRow = *numRows - 2;
      lastCol = *numCols - 2;
      binSec  = *numSecs/2+1;
      binRow  = *numRows/2+1;
      binCol  = *numCols/2+1;
      oddFactor = 0;
      if((binSec%2)==0)
	binSec--;
      if((binRow%2)==0)
	binRow--;
      if((binCol%2)==0)
	{/* Make Odd */
	  oddFactor = 1;
	  binCol--;
	}/* Make Odd */
      rowColBin = (binCol) * (binRow);
      binVals[0] = 0.015625;
      binVals[1] = 0.03125; 
      binVals[2] = 0.015625; 
      binVals[3] = 0.03125; 
      binVals[4] = 0.0625;   
      binVals[5] = 0.03125;
      binVals[6] = 0.015625; 
      binVals[7] = 0.03125; 
      binVals[8] = 0.015625; 
      binVals[9] = 0.03125; 
      binVals[10] = 0.0625;   
      binVals[11] = 0.03125;
      binVals[12] = 0.0625;   
      binVals[13] = 0.125;
      binVals[14] = 0.0625;   
      binVals[15] = 0.03125;
      binVals[16] = 0.0625;   
      binVals[17] = 0.03125;
      binVals[18] = 0.015625; 
      binVals[19] = 0.03125; 
      binVals[20] = 0.015625; 
      binVals[21] = 0.03125; 
      binVals[22] = 0.0625;   
      binVals[23] = 0.03125;
      binVals[24] = 0.015625; 
      binVals[25] = 0.03125; 
      binVals[26] = 0.015625; 

    }/* They want to sample average bin the data */
  else
    {/* They DON'T want to bin the data */
      binStep = 1;
      lastSec = *numSecs;
      lastRow = *numRows;
      lastCol = *numCols;
      rowColBin = (*numCols/binStep) * (*numRows/binStep);

    }/* They DON'T want to bin the data */

  rowCol    = (*numCols) * (*numRows);
  sec1 = short_buffer[0] = (short *)malloc(sizeof(short)*rowCol);
  sec2 = short_buffer[1] = (short *)malloc(sizeof(short)*rowCol);
  sec3 = short_buffer[2] = (short *)malloc(sizeof(short)*rowCol);

  fIn = lun[*luNum];

  fseek(fIn,*imgOffset,0);

  if(*binSwitch == 2)
    {/*  Sample Average binning */

      if(*strtCol != 1 ||
	 *strtRow != 1 ||
	 *strtSec != 1 ||
	 *ncol    != *numCols ||
	 *nrow    != *numRows ||
	 *nsec    != *numSecs 
	 )
	{/* Can't do */
	  printf("Don't know how to truncate the MAP with binned data.\n");
	  printf("[P_R_M_S_I] [%d][%d][%d]-[%d][%d][%d]\n",
		 *strtCol,*strtRow,*strtSec,
		 *ncol,*nrow,*nsec);
	  exit(0);
	}/* Can't do */

      for(i=oddFactor;i<(*numSecs-oddFactor-1);i+=binStep)
	{/* Do these sections */
	  if(i==oddFactor)
	    {/* Read all three */
	      if(oddFactor==0)
		{/* Don't read first section */
		}/* Don't read first section */
	      else
		{/* Read it */
		    OKREAD(fread(short_buffer[0],sizeof(short), /*ROK*/
				 (rowCol),fIn),rowCol,
			 "PIF_READ_MAP_SHORT_IMAGE");
		  for(m=0;m<rowCol;m++)
		    differentEndian_((BYTE *)&sec1[m],2, endians[*luNum]);
		}/* Read it */
	      OKREAD(fread(short_buffer[1],sizeof(short),(rowCol),fIn),/*ROK*/
		     rowCol,
		     "PIF_READ_MAP_SHORT_IMAGE");
	      for(m=0;m<rowCol;m++)
		differentEndian_((BYTE *)&sec2[m],2, endians[*luNum]);
	      OKREAD(fread(short_buffer[2],sizeof(short),(rowCol),fIn),/*ROK*/
		     rowCol,
		     "PIF_READ_MAP_SHORT_IMAGE");
	      for(m=0;m<rowCol;m++)
		differentEndian_((BYTE *)&sec3[m],2, endians[*luNum]);
	    }/* Read all three */
	  else
	    {/* Just read two */
	      /* Stuff from last time */
	      for(m=0;m<rowCol;m++)
		sec1[m]= sec3[m];
	      OKREAD(fread(short_buffer[1],sizeof(short),(rowCol),fIn),/*ROK*/
		     rowCol,
		     "PIF_READ_MAP_SHORT_IMAGE");
	      for(m=0;m<rowCol;m++)
		differentEndian_((BYTE *)&sec2[m],2, endians[*luNum]);
	      OKREAD(fread(short_buffer[2],sizeof(short),(rowCol),fIn),/*ROK*/
		     rowCol,
		     "PIF_READ_MAP_SHORT_IMAGE");
	      for(m=0;m<rowCol;m++)
		differentEndian_((BYTE *)&sec3[m],2, endians[*luNum]);
	    }/* Just read two */

	  for(j=oddFactor;j<(*numRows-oddFactor-binStep);j+=binStep)
	    {/* Do rows */
	      jM1 = (j-1)*(*numCols);
	      if(jM1 < 0)
		jM1 = 0;
	      j0  = (j)*(*numCols);
	      jP1 = (j+1)*(*numCols);
	      for(k=oddFactor;k<(*numCols-oddFactor-binStep);k+=binStep)
		{/* Do columns */
		  kM1 = k - 1;
		  if(kM1 < 0)
		    kM1 = 0;
		  kP1 = k + 1;
		  
		  tmpVals[0] = ((float)sec1[jM1+kM1] );
		  tmpVals[1] = ((float)sec1[jM1+k]   );
		  tmpVals[2] = ((float)sec1[jM1+kP1] );
		  tmpVals[3] = ((float)sec1[j0+kM1]  );
		  tmpVals[4] = ((float)sec1[j0+k]    );
		  tmpVals[5] = ((float)sec1[j0+kP1]  );
		  tmpVals[6] = ((float)sec1[jP1+kM1] );
		  tmpVals[7] = ((float)sec1[jP1+k]   );
		  tmpVals[8] = ((float)sec1[jP1+kP1] );
		  tmpVals[9] = ((float)sec2[jM1+kM1] );
		  tmpVals[10] = ((float)sec2[jM1+k]   );
		  tmpVals[11] = ((float)sec2[jM1+kP1] );
		  tmpVals[12] = ((float)sec2[j0+kM1]  );
		  tmpVals[13] = ((float)sec2[j0+k]    );
		  tmpVals[14] = ((float)sec2[j0+kP1]  );
		  tmpVals[15] = ((float)sec2[jP1+kM1] );
		  tmpVals[16] = ((float)sec2[jP1+k]   );
		  tmpVals[17] = ((float)sec2[jP1+kP1] );
		  tmpVals[18] = ((float)sec3[jM1+kM1] );
		  tmpVals[19] = ((float)sec3[jM1+k]   );
		  tmpVals[20] = ((float)sec3[jM1+kP1] );
		  tmpVals[21] = ((float)sec3[j0+kM1]  );
		  tmpVals[22] = ((float)sec3[j0+k]    );
		  tmpVals[23] = ((float)sec3[j0+kP1]  );
		  tmpVals[24] = ((float)sec3[jP1+kM1] );
		  tmpVals[25] = ((float)sec3[jP1+k]   );
		  tmpVals[26] = ((float)sec3[jP1+kP1] );

		  if(j==0)
		    {/* Off the bottom */
		      tmpVals[0] = 
			tmpVals[1] = 
			  tmpVals[2] = 
			    tmpVals[9] = 
			      tmpVals[10] = 
				tmpVals[11] = 
				  tmpVals[18] = 
				    tmpVals[19] = 
				      tmpVals[20] = 0.0;
		    }/* Off the bottom */
		  if(k==0)
		    {/* Off the bottom */
		      tmpVals[0] = 
			tmpVals[3] = 
			  tmpVals[6] = 
			    tmpVals[9] = 
			      tmpVals[12] = 
				tmpVals[15] = 
				  tmpVals[18] = 
				    tmpVals[21] = 
				      tmpVals[24] = 0.0;
		    }/* Off the bottom */
		  if(j==(*numRows-1))
		    {/* Off the Top */
		      tmpVals[6] = 
			tmpVals[7] = 
			  tmpVals[8] = 
			    tmpVals[15] = 
			      tmpVals[16] = 
				tmpVals[17] = 
				  tmpVals[24] = 
				    tmpVals[25] = 
				      tmpVals[26] = 0.0;
		    }/* Off the Top */
		  if(k==(*numCols-1))
		    {/* Off the Top */
		      tmpVals[2] = 
			tmpVals[5] = 
			  tmpVals[8] = 
			    tmpVals[11] = 
			      tmpVals[14] = 
				tmpVals[17] = 
				  tmpVals[20] = 
				    tmpVals[23] = 
				      tmpVals[26] = 0.0;
		    }/* Off the Top */

		  totFloats = (
			       ((float)tmpVals[0] * binVals[0]) +
			       ((float)tmpVals[1] * binVals[1]) +
			       ((float)tmpVals[2] * binVals[2]) +
			       ((float)tmpVals[3] * binVals[3]) +
			       ((float)tmpVals[4] * binVals[4]) +
			       ((float)tmpVals[5] * binVals[5]) +
			       ((float)tmpVals[6] * binVals[6]) +
			       ((float)tmpVals[7] * binVals[7]) +
			       ((float)tmpVals[8] * binVals[8]) +
			       ((float)tmpVals[9] * binVals[9]) +
			       ((float)tmpVals[10] * binVals[10]) +
			       ((float)tmpVals[11] * binVals[11]) +
			       ((float)tmpVals[12] * binVals[12]) +
			       ((float)tmpVals[13] * binVals[13]) +
			       ((float)tmpVals[14] * binVals[14]) +
			       ((float)tmpVals[15] * binVals[15]) +
			       ((float)tmpVals[16] * binVals[16]) +
			       ((float)tmpVals[17] * binVals[17]) +
			       ((float)tmpVals[18] * binVals[18]) +
			       ((float)tmpVals[19] * binVals[19]) +
			       ((float)tmpVals[20] * binVals[20]) +
			       ((float)tmpVals[21] * binVals[21]) +
			       ((float)tmpVals[22] * binVals[22]) +
			       ((float)tmpVals[23] * binVals[23]) +
			       ((float)tmpVals[24] * binVals[24]) +
			       ((float)tmpVals[25] * binVals[25]) +
			       ((float)tmpVals[26] * binVals[26]) 
			       );

#ifdef DEBUG4
		  /* e.g. Just sample it don't average yet */
		  totFloats = ((float)sec2[j0+k]    );
#endif

		  /* Do we want to convert from TSB format for maps to robem
		     format?? */
		  if(convTsb)
		    rowOut = ((lastRow - j - 1) / binStep)*(binCol);
		  else
		    rowOut = (j / binStep)*(binCol);
		  
		  colOut = (k) / binStep;
		  
		  thisFloat= 
		    float_buffer[((i/binStep)*rowColBin)+
				 rowOut+colOut]= (float)
				   (((float)(totFloats)) *
				    (*recipFloatScaleFactor));
		  
		  *minMapPixel = MIN(*minMapPixel,thisFloat);
		  *maxMapPixel = MAX(*maxMapPixel,thisFloat);
		}/* Do columns */

	    }/* Do rows */

	}/* Do these sections */

    }/*  Sample Average binning */
  else
    {/* Normal */
      int maxColOut,maxRowOut,maxSecOut;

      rowColBin = (*ncol) * (*nrow);
      maxColOut = (*ncol);
      maxRowOut = ((*nrow - 1) / binStep)*(*ncol/binStep);
      maxSecOut = (int)((int)((*nsec-1)/binStep)*rowColBin);

      /* Only do what we need for the truncate buffer case */
      lastSec = *nsec + (*strtSec-1);
      lastRow = *nrow + (*strtRow-1);
      lastCol = *ncol + (*strtCol-1);

      /* Don't read past what's on the disk */
      lastSec = MIN(lastSec,*numSecs);
      lastRow = MIN(lastRow,*numRows);
      lastCol = MIN(lastCol,*numCols);

      for(i=0;i<(lastSec);i+=binStep)
	{/* Do all sections */
	  /* printf("Sec[%d]\n",i); */
	  for(l=0;l<binStep;l++)
	    {/* REad all sections */
		OKREAD(fread(short_buffer[l],sizeof(short), /*ROK*/
			     (rowCol),fIn),rowCol,
		     "PIF_READ_MAP_SHORT_IMAGE");
	    }/* REad all sections */
	  /* Account for where they want to start on the disk */
	  secOut = (int)((int)((i-(*strtSec-1))/binStep)*rowColBin);
	  /* Make sure we don't go negative */
	  secOut = MIN(MAX(0,secOut),maxSecOut);
	  
	  for(j=0;j<(lastRow);j+=binStep)
	    {/* Rows */
	      /* Do we want to convert from TSB format for maps to robem
		 format?? */
	      /* Account for where they want to start on the disk */
	      if(convTsb)
		{/* Flip Y */
		  rowOut = (((lastRow+(*strtRow-1)) - j - 1) / binStep)*
		    (*ncol/binStep);
		}/* Flip Y */
	      else
		{/* "normal */
		  rowOut = ((j-(*strtRow-1)) / binStep)*(*ncol/binStep);
		}/* "normal */

	      /* Make sure we don't overflow */
	      rowOut = MIN(MAX(0,rowOut),maxRowOut);
	      
	      for(k=0;k<(lastCol);k+=binStep)
		{/* Stuff the float buffer */
		  
		  colOut = (k) / binStep;
		  /* Account for where they want to start on the disk */
		  colOut -= (*strtCol-1);
		  /* Make sure we don't go negative */
		  colOut = MIN(MAX(0,colOut),maxColOut);

		  if(binStep == 1)
		    {/* Just do every pixel */
		      thisShortBuffer = short_buffer[0];

		      thisOffset = (int)((j*(*numCols))+k);
		      
		      differentEndian_((BYTE *)&thisShortBuffer[thisOffset], 
				       2, endians[*luNum]);
		      thisFloat =
			float_buffer[(secOut+rowOut+colOut)]= 
				       (float)
				     (((float)(thisShortBuffer[thisOffset])) *
					  (*recipFloatScaleFactor));
		      
		    }/* Just do every pixel */
		  else
		    {/* Average pixels to "bin" them */
		      totFloats = 0.0;
		      totPnts = 0;
		      for(l=0;l<binStep;l++)
			{/* Do both sections */
			  for(m=0;m<binStep;m++)
			    {/* Do both rows */
			      for(n=0;n<binStep;n++)
				{/* Do both cols */
				  
				  thisOffset = (((j+m)*(*numCols))+(k+n));
				  thisShortBuffer = short_buffer[l];
				  
				  differentEndian_((BYTE *)&thisShortBuffer[
							   thisOffset], 
						   2, endians[*luNum]);
				  
				  totFloats +=
				    (float)
				      (((float)(thisShortBuffer[thisOffset])) *
				       (*recipFloatScaleFactor));
				  totPnts++;
				  
				}/* Do both cols */
			    }/* Do both rows */
			}/* Do both sections */
		      
		      /* Debugging stuff:
			 printf("i[%d]bin[%d]rcb[%d]ro[%d]co[%d]\n",
			 i,binStep,rowColBin,
			 rowOut,colOut);
			 */
		      thisFloat = 
			float_buffer[((i/binStep)*rowColBin)+
				     rowOut+colOut]= (float)
				       (totFloats/(float)totPnts);
		    }/* Average pixels to "bin" them */
		  
		  *minMapPixel = MIN(*minMapPixel,thisFloat);
		  *maxMapPixel = MAX(*maxMapPixel,thisFloat);
		}/* Stuff the float buffer */
	      
	    }/* Rows */
	}/* Do all sections */
      
    }/* Normal */
  
  free(short_buffer[0]);
  free(short_buffer[1]);
  free(short_buffer[2]);

}/* PIF_READ_MAP_SHORT_IMAGE */

/***************************************************************
 * pif_read_imgi4_() - Read the PIF file's SHORT image 
 * for a floating point MAP.
*/
void pif_read_imgi4_(forPifHead_t *thisHead,
		     float *float_buffer,
		     int * ncol,
		     int * nrow,
		     int * thisImg
		     )
{/* PIF_READ_IMGI4 */
  int tmpThisImg, tmpSec2, numSecs, binCol, binRow, thisBin;

  if(thisHead->recipFloatScaleFactor == 0.0)
    thisHead->recipFloatScaleFactor =  1.0 /  thisHead->floatScaleFactor ;
  if(thisHead->floatScaleFactor == 0.0)
    thisHead->floatScaleFactor =  1.0 /  thisHead->recipFloatScaleFactor ;

  tmpThisImg = *thisImg;

  if(tmpThisImg > thisHead->nimg ||
     tmpThisImg < 1 )
    {/* Fatal error */
      printf("[PIF_READ_IMGI4() Invalid image [%d]\n",
	     tmpThisImg);
      exit(0);
    }/* Fatal error */

  numSecs = 1;
  
  binCol = thisHead->numCols;
  binRow = thisHead->numRows;

  if(*ncol == thisHead->numCols &&
     *nrow == thisHead->numRows )
    thisBin = 1;
  else thisBin = 2;

  if(thisHead->binFactor == 2)
    {/* It's already binned on disk */
      binRow  = (thisHead->numRows)/2+1;
      binCol  = (thisHead->numCols)/2+1;
      if((binRow%2)==0)
	binRow--;
      if((binCol%2)==0)
	binCol--;
      if(thisHead->binFactor == 2 &&
	 thisBin == 1)
	{/* Error */
	  printf("***********************************************\n");
	  printf("PIF_READ_IMGI4()\n");
	  printf("Can't expand BIN! DiskBIN_STORE==[%d] You want[%d]\n",
		 thisHead->binFactor,
		 thisBin);
	  printf("***********************************************\n");
	  exit(0);
	}/* Error */

      /* We've already taken care of thisBin with binCol/binRow etc... */
      thisBin = 1;

    }/* It's already binned on disk */

  /* Times 4 because we're INT here */
  thisHead->imgOffset = 1024 + (512 * (tmpThisImg-1)) +
    (int)((tmpThisImg - 1) * (binCol * binRow * 4));

#ifdef OLDWAY
  printf("Calling pif_read_map_int_image [%d]\n",thisBin);
  pif_read_map_int_image_(&(thisHead->luNum),
			  float_buffer,
			  &(thisHead->imgOffset),
			  &(thisHead->numCols),
			  &(thisHead->numRows),
			  &numSecs,
			  &(thisHead->recipFloatScaleFactor),
			  &(thisHead->datar4_min),
			  &(thisHead->datar4_max),
			  &thisBin,
			  &(thisHead->mode)
			  );
#endif

  pif_read_float_image_(&(thisHead->luNum),
			float_buffer,
			&(thisHead->imgOffset),
			&(thisHead->numCols),
			&(thisHead->numRows),
			&numSecs,
			&thisBin,
			&(thisHead->recipFloatScaleFactor));


}/* PIF_READ_IMGI4 */



void pif_debug_write_(float *img,int *ncol,int *nrow)
{
  int i,j;
  FILE *fOut;
  fOut = fopen("debug.data","w");
  fwrite(img,sizeof(float),((*ncol)*(*nrow)),fOut);
}

/***************************************************************
 * pif_read_map_int_image() - Read the PIF file's INT image 
 * for a floating point MAP.
*/
void pif_read_map_int_image_(int *luNum,
			     float *float_buffer,
			     int *imgOffset,
			     int *numCols, int *numRows, int *numSecs,
			     float *recipFloatScaleFactor,
			     float *minMapPixel, float *maxMapPixel,
			     int *binSwitch, int *mode)
{/* PIF_READ_MAP_INT_IMAGE */
  int i,j,k,l,m,n,rowCol,rowColBin;
  int *int_buffer[4];         /* Hard code this for only 2x2 binning now */
  int *sec1,*sec2,*sec3;
  int *thisIntBuffer;
  float thisFloat,totFloats;
  FILE *fIn;
  int totPnts, thisOffset,binCol,binRow,binSec,oddFactor;
  int binStep,rowOut,colOut,lastSec,lastRow,lastCol;
  int jM1,j0,jP1,kM1,kP1;
  float binVals[27] = {0.015625, 
			 0.03125, 
			 0.015625, 
			 0.03125, 
			 0.0625,   
			 0.03125,
			 0.015625, 
			 0.03125, 
			 0.015625, 
			 0.03125, 
			 0.0625,   
			 0.03125,
			 0.0625,   
			 0.125,
			 0.0625,   
			 0.03125,
			 0.0625,   
			 0.03125,
			 0.015625, 
			 0.03125, 
			 0.015625, 
			 0.03125, 
			 0.0625,   
			 0.03125,
			 0.015625, 
			 0.03125, 
			 0.015625 };
  float tmpVals[27];
		       
  if(*binSwitch > 2)
    {/* Error */
      printf("PIF_READ_MAP_INT_IMAGE - Cannot bin data like this[%d]\n",
	     *binSwitch);
      exit(0);
    }/* Error */
  else if(*binSwitch==2)
    {/* They want to sample average bin the data */
      binStep = 2;
      lastSec = *numSecs - 2;
      lastRow = *numRows - 2;
      lastCol = *numCols - 2;
      if(*mode==20)
	binSec  = *numSecs/2+1;
      else binSec = *numSecs;
      binRow  = *numRows/2+1;
      binCol  = *numCols/2+1;
      oddFactor = 0;
      if((binSec%2)==0)
	binSec--;
      if((binRow%2)==0)
	binRow--;
      if((binCol%2)==0)
	{/* Make Odd */
	  oddFactor = 1;
	  binCol--;
	}/* Make Odd */
      rowColBin = (binCol) * (binRow);
      binVals[0] = 0.015625;
      binVals[1] = 0.03125; 
      binVals[2] = 0.015625; 
      binVals[3] = 0.03125; 
      binVals[4] = 0.0625;   
      binVals[5] = 0.03125;
      binVals[6] = 0.015625; 
      binVals[7] = 0.03125; 
      binVals[8] = 0.015625; 
      binVals[9] = 0.03125; 
      binVals[10] = 0.0625;   
      binVals[11] = 0.03125;
      binVals[12] = 0.0625;   
      binVals[13] = 0.125;
      binVals[14] = 0.0625;   
      binVals[15] = 0.03125;
      binVals[16] = 0.0625;   
      binVals[17] = 0.03125;
      binVals[18] = 0.015625; 
      binVals[19] = 0.03125; 
      binVals[20] = 0.015625; 
      binVals[21] = 0.03125; 
      binVals[22] = 0.0625;   
      binVals[23] = 0.03125;
      binVals[24] = 0.015625; 
      binVals[25] = 0.03125; 
      binVals[26] = 0.015625; 

    }/* They want to sample average bin the data */
  else
    {/* They DON'T want to bin the data */
      binStep = 1;
      lastSec = *numSecs;
      binRow = lastRow = *numRows;
      binCol = lastCol = *numCols;
      rowColBin = (*numCols/binStep) * (*numRows/binStep);

    }/* They DON'T want to bin the data */

  rowCol    = (*numCols) * (*numRows);
  sec1 = int_buffer[0] = (int *)malloc(sizeof(int)*rowCol);
  sec2 = int_buffer[1] = (int *)malloc(sizeof(int)*rowCol);
  sec3 = int_buffer[2] = (int *)malloc(sizeof(int)*rowCol);

  fIn = lun[*luNum];

  fseek(fIn,*imgOffset,0);

  /**********
  printf("ncol[%d]nr[%d]rowcol[%d]  ns[%d]\n",
	 *numCols,*numRows,rowCol,*numSecs);
  printf("bincol[%d]ro[%d]step[%d]   bsw[%d]\n",
	 binCol,binRow,binStep,*binSwitch);
  ********/

  if(*binSwitch == 2)
    {/*  Sample Average binning */
      int lastSection;
      lastSection = (*numSecs-oddFactor-1);
      if(lastSection <= oddFactor)
	lastSection = oddFactor + 1;  /* Read at least one of these */
      for(i=oddFactor;i<(lastSection);i+=binStep)
	{/* Do these sections */
	  if(i==oddFactor)
	    {/* Read all three */
	      if(oddFactor==0)
		{/* Don't read first section */
		}/* Don't read first section */
	      else
		{/* Read it */
		    OKREAD(fread(int_buffer[0],sizeof(int), /*ROK*/
			       (rowCol),fIn), rowCol,
			 "PIF_READ_MAP_INT_IMAGE");
		  for(m=0;m<rowCol;m++)
		    differentEndian_((BYTE *)&sec1[m],4, endians[*luNum]);
		}/* Read it */
	      OKREAD(fread(int_buffer[1],sizeof(int),(rowCol),fIn),/*ROK*/
		     rowCol,
		     "PIF_READ_MAP_INT_IMAGE");
	      for(m=0;m<rowCol;m++)
		differentEndian_((BYTE *)&sec2[m],4, endians[*luNum]);
	      OKREAD(fread(int_buffer[2],sizeof(int),(rowCol),fIn),/*ROK*/
		     rowCol,
		     "PIF_READ_MAP_INT_IMAGE");
	      for(m=0;m<rowCol;m++)
		differentEndian_((BYTE *)&sec3[m],4, endians[*luNum]);
	    }/* Read all three */
	  else
	    {/* Just read two */
	      /* Stuff from last time */
	      for(m=0;m<rowCol;m++)
		sec1[m]= sec3[m];
	      OKREAD(fread(int_buffer[1],sizeof(int),(rowCol),fIn),/*ROK*/
		     rowCol,
		     "PIF_READ_MAP_INT_IMAGE");
	      for(m=0;m<rowCol;m++)
		differentEndian_((BYTE *)&sec2[m],4, endians[*luNum]);
	      OKREAD(fread(int_buffer[2],sizeof(int),(rowCol),fIn),/*ROK*/
		     rowCol,
		     "PIF_READ_MAP_INT_IMAGE");
	      for(m=0;m<rowCol;m++)
		differentEndian_((BYTE *)&sec3[m],4, endians[*luNum]);
	    }/* Just read two */

	  for(j=oddFactor;j<(*numRows-oddFactor);j+=binStep)
	    {/* Do rows */
	      jM1 = (j-1)*(*numCols);
	      if(jM1 < 0)
		jM1 = 0;
	      j0  = (j)*(*numCols);
	      jP1 = (j+1)*(*numCols);
	      for(k=oddFactor;k<(*numCols-oddFactor);k+=binStep)
		{/* Do columns */
		  kM1 = k - 1;
		  if(kM1 < 0)
		    kM1 = 0;
		  kP1 = k + 1;
		  
		  tmpVals[0] = ((float)sec1[jM1+kM1] );
		  tmpVals[1] = ((float)sec1[jM1+k]   );
		  tmpVals[2] = ((float)sec1[jM1+kP1] );
		  tmpVals[3] = ((float)sec1[j0+kM1]  );
		  tmpVals[4] = ((float)sec1[j0+k]    );
		  tmpVals[5] = ((float)sec1[j0+kP1]  );
		  tmpVals[6] = ((float)sec1[jP1+kM1] );
		  tmpVals[7] = ((float)sec1[jP1+k]   );
		  tmpVals[8] = ((float)sec1[jP1+kP1] );
		  tmpVals[9] = ((float)sec2[jM1+kM1] );
		  tmpVals[10] = ((float)sec2[jM1+k]   );
		  tmpVals[11] = ((float)sec2[jM1+kP1] );
		  tmpVals[12] = ((float)sec2[j0+kM1]  );
		  tmpVals[13] = ((float)sec2[j0+k]    );
		  tmpVals[14] = ((float)sec2[j0+kP1]  );
		  tmpVals[15] = ((float)sec2[jP1+kM1] );
		  tmpVals[16] = ((float)sec2[jP1+k]   );
		  tmpVals[17] = ((float)sec2[jP1+kP1] );
		  tmpVals[18] = ((float)sec3[jM1+kM1] );
		  tmpVals[19] = ((float)sec3[jM1+k]   );
		  tmpVals[20] = ((float)sec3[jM1+kP1] );
		  tmpVals[21] = ((float)sec3[j0+kM1]  );
		  tmpVals[22] = ((float)sec3[j0+k]    );
		  tmpVals[23] = ((float)sec3[j0+kP1]  );
		  tmpVals[24] = ((float)sec3[jP1+kM1] );
		  tmpVals[25] = ((float)sec3[jP1+k]   );
		  tmpVals[26] = ((float)sec3[jP1+kP1] );

		  if(j==0)
		    {/* Off the bottom */
		      tmpVals[0] = 
			tmpVals[1] = 
			  tmpVals[2] = 
			    tmpVals[9] = 
			      tmpVals[10] = 
				tmpVals[11] = 
				  tmpVals[18] = 
				    tmpVals[19] = 
				      tmpVals[20] = 0.0;
		    }/* Off the bottom */
		  if(k==0)
		    {/* Off the bottom */
		      tmpVals[0] = 
			tmpVals[3] = 
			  tmpVals[6] = 
			    tmpVals[9] = 
			      tmpVals[12] = 
				tmpVals[15] = 
				  tmpVals[18] = 
				    tmpVals[21] = 
				      tmpVals[24] = 0.0;
		    }/* Off the bottom */
		  if(j==(*numRows-1))
		    {/* Off the Top */
		      tmpVals[6] = 
			tmpVals[7] = 
			  tmpVals[8] = 
			    tmpVals[15] = 
			      tmpVals[16] = 
				tmpVals[17] = 
				  tmpVals[24] = 
				    tmpVals[25] = 
				      tmpVals[26] = 0.0;
		    }/* Off the Top */
		  if(k==(*numCols-1))
		    {/* Off the Top */
		      tmpVals[2] = 
			tmpVals[5] = 
			  tmpVals[8] = 
			    tmpVals[11] = 
			      tmpVals[14] = 
				tmpVals[17] = 
				  tmpVals[20] = 
				    tmpVals[23] = 
				      tmpVals[26] = 0.0;
		    }/* Off the Top */

		  totFloats = (
			       ((float)tmpVals[0] * binVals[0]) +
			       ((float)tmpVals[1] * binVals[1]) +
			       ((float)tmpVals[2] * binVals[2]) +
			       ((float)tmpVals[3] * binVals[3]) +
			       ((float)tmpVals[4] * binVals[4]) +
			       ((float)tmpVals[5] * binVals[5]) +
			       ((float)tmpVals[6] * binVals[6]) +
			       ((float)tmpVals[7] * binVals[7]) +
			       ((float)tmpVals[8] * binVals[8]) +
			       ((float)tmpVals[9] * binVals[9]) +
			       ((float)tmpVals[10] * binVals[10]) +
			       ((float)tmpVals[11] * binVals[11]) +
			       ((float)tmpVals[12] * binVals[12]) +
			       ((float)tmpVals[13] * binVals[13]) +
			       ((float)tmpVals[14] * binVals[14]) +
			       ((float)tmpVals[15] * binVals[15]) +
			       ((float)tmpVals[16] * binVals[16]) +
			       ((float)tmpVals[17] * binVals[17]) +
			       ((float)tmpVals[18] * binVals[18]) +
			       ((float)tmpVals[19] * binVals[19]) +
			       ((float)tmpVals[20] * binVals[20]) +
			       ((float)tmpVals[21] * binVals[21]) +
			       ((float)tmpVals[22] * binVals[22]) +
			       ((float)tmpVals[23] * binVals[23]) +
			       ((float)tmpVals[24] * binVals[24]) +
			       ((float)tmpVals[25] * binVals[25]) +
			       ((float)tmpVals[26] * binVals[26]) 
			       );

#ifdef DEBUG444
		  /* e.g. Just sample it don't average yet */
		  totFloats = ((float)sec2[j0+k]    );
#endif

		  /* Do we want to convert from TSB format for maps to robem
		     format?? */
		  if(convTsb)
		    rowOut = ((lastRow - j - 1) / binStep)*(binCol);
		  else
		    rowOut = (j / binStep)*(binCol);
		  
		  colOut = (k) / binStep;
		  
		  thisFloat = 
		    float_buffer[((i/binStep)*rowColBin)+
				 rowOut+colOut]= (float)
				   (((float)(totFloats)) *
				    (*recipFloatScaleFactor));
		  
		  *minMapPixel = MIN(*minMapPixel,thisFloat);
		  *maxMapPixel = MAX(*maxMapPixel,thisFloat);
		}/* Do columns */

	    }/* Do rows */

	}/* Do these sections */

    }/*  Sample Average binning */
  else
    {/* Normal */

      for(i=0;i<(lastSec);i+=binStep)
	{/* Do all sections */
	  /* printf("Sec[%d]  rowcol = [%d] bS[%d]\n",
		 i,rowCol,binStep);   */
	  for(l=0;l<binStep;l++)
	    {/* REad all sections */
		OKREAD(fread(int_buffer[l],sizeof(int),(rowCol),fIn), /*ROK*/
		     rowCol,
		     "PIF_READ_MAP_INT_IMAGE");
	    }/* REad all sections */
	  for(j=0;j<(lastRow);j+=binStep)
	    {/* Rows */
	      for(k=0;k<(lastCol);k+=binStep)
		{/* Stuff the float buffer */
		  
		  /* Do we want to convert from TSB format for maps to robem
		     format?? */
		  if(convTsb)
		    rowOut = ((lastRow - j - 1) / binStep)*(*numCols/binStep);
		  else
		    rowOut = ((j) / binStep)*(*numCols/binStep);
		  
		  colOut = (k) / binStep;
		  
		  if(binStep == 1)
		    {/* Just do every pixel */
		      thisOffset = (((j)*(*numCols))+(k));
		      thisIntBuffer = int_buffer[0];

		      differentEndian_((BYTE *)&thisIntBuffer[thisOffset], 
				       4, endians[*luNum]);
		      thisFloat =
			float_buffer[((i/binStep)*rowColBin)+
				     rowOut+colOut]= (float)
				       (float)
				   (((float)(thisIntBuffer[thisOffset])) *
					  (*recipFloatScaleFactor));
		      
		    }/* Just do every pixel */
		  else
		    {/* Average pixels to "bin" them */
		      totFloats = 0.0;
		      totPnts = 0;

		      for(l=0;l<binStep;l++)
			{/* Do both sections */
			  for(m=0;m<binStep;m++)
			    {/* Do both rows */
			      for(n=0;n<binStep;n++)
				{/* Do both cols */
				  
				  thisOffset = (((j+m)*(*numCols))+(k+n));
				  thisIntBuffer = int_buffer[l];
				  
				  differentEndian_((BYTE *)&thisIntBuffer[
							   thisOffset], 
						   4, endians[*luNum]);
				  
				  totFloats +=
				    (float)
				      (((float)(thisIntBuffer[thisOffset])) *
				       (*recipFloatScaleFactor));
				  totPnts++;
				  
				}/* Do both cols */
			    }/* Do both rows */
			}/* Do both sections */
		      
		      /* Debugging stuff:
		      if(i >=6)
			 printf("i[%d]bin[%d]rcb[%d]ro[%d]co[%d]\n",
			 i,binStep,rowColBin,
			 rowOut,colOut);
			 */

		      thisFloat = 
			float_buffer[((i/binStep)*rowColBin)+
				     rowOut+colOut]= (float)
				       (totFloats/(float)totPnts);
		    }/* Average pixels to "bin" them */

		  *minMapPixel = MIN(*minMapPixel,thisFloat);
		  *maxMapPixel = MAX(*maxMapPixel,thisFloat);
		}/* Stuff the float buffer */
	      
	    }/* Rows */
	}/* Do all sections */
      
    }/* Normal */
  
  free(int_buffer[0]);
  free(int_buffer[1]);
  free(int_buffer[2]);

}/* PIF_READ_MAP_INT_IMAGE */


void pif_read_sf(forPifHead_t *sfHead,
		   strucFact_t *sfs)
{/* PIF_READ_SF */
  pif_read_sf_(sfHead, sfs);
}/* PIF_READ_SF */
/**************************************************************
 * pif_read_sf() - Read the Structure factor array to disk.
*/
void pif_read_sf_(forPifHead_t *sfHead,
		   strucFact_t *sfs)
{/* PIF_READ_SF */
  int i,j,k, imgOffset, maxP,localInt;
  FILE *fIn;
  float maxF,minF;

  fIn = lun[sfHead->luNum];

  fseek(fIn,1024,0);  /* SF files always begin at 1024 */

  maxF = -999999.9;
  minF = 999999.9;
  for(i=0;i<sfHead->numRows;i++)
    {/* Read all */


	OKREAD(fread(&(sfs[i].h),sizeof(short),1,fIn),1,"PIF_READ_SF");/*ROK*/
	differentEndian_((BYTE *)&(sfs[i].h),2,  /*RSSSS*/
			 endians[sfHead->luNum]);/*RSSSS*/
	fread(&(sfs[i].k),sizeof(short),1,fIn);  /*ROK*/
	differentEndian_((BYTE *)&(sfs[i].k),2,  /*RSSSS*/
			 endians[sfHead->luNum]);/*RSSSS*/
	fread(&(sfs[i].l),sizeof(short),1,fIn);  /*ROK*/
	differentEndian_((BYTE *)&(sfs[i].l),2,  /*RSSSS*/
			 endians[sfHead->luNum]);/*RSSSS*/
 
      fread(&localInt,sizeof(int),1,fIn); /*ROK*/       
      differentEndian_((BYTE *)&localInt, 4, endians[sfHead->luNum]);
      convertBackFloat_(localInt,&(sfs[i].a_amp),
			sfHead->recipFloatScaleFactor);

      fread(&localInt,sizeof(int),1,fIn); /*ROK*/       
      differentEndian_((BYTE *)&localInt, 4, endians[sfHead->luNum]);
      convertBackFloat_(localInt,&(sfs[i].b_phi),
			sfHead->recipFloatScaleFactor);

      fread(&localInt,sizeof(int),1,fIn); /*ROK*/       
      differentEndian_((BYTE *)&localInt, 4, endians[sfHead->luNum]);
      convertBackFloat_(localInt,&(sfs[i].sig),
			sfHead->recipFloatScaleFactor);

      fread(&localInt,sizeof(int),1,fIn); /*ROK*/       
      differentEndian_((BYTE *)&localInt, 4, endians[sfHead->luNum]);
      convertBackFloat_(localInt,&(sfs[i].dstr),
			sfHead->recipFloatScaleFactor);

      minF = MIN(sfs[i].a_amp,minF);
      minF = MIN(sfs[i].b_phi,minF);

      maxF = MAX(sfs[i].a_amp,maxF);
      maxF = MAX(sfs[i].b_phi,maxF);

    }/* Read all */

}/* PIF_READ_SF */



void pif_init_head(forPifHead_t *thisHead)
{
  pif_init_head_(thisHead);
}
/*************************************************************
 * pif_init_head() - Initialize this PIF header to defautls.
*/
void pif_init_head_(forPifHead_t *thisHead)
{/* PIF_INIT_HEAD */

  thisHead->luNum      = 0;
  thisHead->nimg       = 1;
  thisHead->imgNum     = 1;
  thisHead->floatScaleFactor = 1.0;
  thisHead->recipFloatScaleFactor = 1.0;
  thisHead->int4scaleF = 1.0; /*RSS*/
  thisHead->htype      = 1;
  thisHead->numCols    = 0;
  thisHead->numRows    = 0;
  thisHead->numSecs    = 1;
  thisHead->mode       = -99;    /*20 ==MAP 21 = PRJS 
				   22 = PFTS 6 = PACKED BOXES 
				   1 == Zeiss scan*/
  thisHead->endianNess = Util_IAmBigEndian();

  sprintf(thisHead->versionStr,
"Undefined                                                                    ");
  thisHead->img_bkg    = 0;
  thisHead->datai4_min    = 0;
  thisHead->datai4_max    = 0;
  thisHead->datar4_min    = 0.0;
  thisHead->datar4_max    = 0.0;
  thisHead->datar4_avg    = 0.0;
  thisHead->datar4_std    = 0.0;
  thisHead->imgOffset  = 0;
  sprintf(thisHead->descriptionStr, 
"Undefined                                                                    ");
  sprintf(thisHead->timeStr, 
"Undefined                                                                    ");
  sprintf(thisHead->micrographStr, 
"Undefined                                                                    ");
  sprintf(thisHead->scanNumberStr, 
"Undefined                                                                    ");
  thisHead->aoverb     = 1.0;
  thisHead->map_abang  = 90.0;
  thisHead->a_cell     = 0.0;
  thisHead->b_cell     = 0.0;
  thisHead->c_cell     = 0.0;
  thisHead->alp        = 90.0;
  thisHead->bet        = 90.0;
  thisHead->gam        = 90.0;
  thisHead->dela       = 1.0;
  thisHead->delb       = 1.0;
  thisHead->delc       = 1.0;
  thisHead->a_star     = 0.0;
  thisHead->b_star     = 0.0;
  thisHead->c_star     = 0.0;
  thisHead->alp_star   = 0.0;
  thisHead->bet_star   = 0.0;
  thisHead->gam_star   = 0.0;
  thisHead->pixelSize  = 0.0;
  thisHead->t_matrix[0]= 1;
  thisHead->t_matrix[1]= 0;
  thisHead->t_matrix[2]= 0;
  thisHead->t_matrix[3]= 1;
  thisHead->t_matrix[4]= 0;
  thisHead->t_matrix[5]= 0;
  thisHead->packRadius = 0;
  thisHead->binFactor  = 1;
  thisHead->dthe       = 1.0;
  thisHead->dphi_90    = 1.0;
  thisHead->symmetry   = 1;
  thisHead->pixSizUnits = 0;
  thisHead->res_out = 0.0;
  thisHead->ctfMode = 0;
  thisHead->tempFactor = 0.0;
  thisHead->wiener = 0.0;
  thisHead->borderType = 0;
  thisHead->borderWidth = 0;
  thisHead->rad_lo = 0;
  thisHead->rad_hi = 0;
  thisHead->even   = 0;
  thisHead->mrcX   = 0;
  thisHead->mrcY   = 0;
  thisHead->mrcZ   = 0;
  thisHead->transX   = 0;
  thisHead->transY   = 0;
  thisHead->transZ   = 0;
  thisHead->isTrunc   = 0;
  thisHead->origX    = 0;
  thisHead->origY    = 0;
  thisHead->origZ    = 0;
  
}/* PIF_INIT_HEAD */




/***************************************************************
 * pif_list_head() - Write the PIF file's header.
*/
void pif_list_head_(forPifHead_t *thisHead)
{/* PIF_LIST_HEAD */

  printf("luNum             [%d]\n",thisHead->luNum);
  printf("nimg              [%d]\n",thisHead->nimg);
  printf("version           [%s]\n",thisHead->versionStr);
  printf("int4scaleF        [%f]\n",thisHead->int4scaleF);
  printf("scaleF            [%f]\n",thisHead->floatScaleFactor);
  printf("img_bkg           [%f]\n",thisHead->img_bkg);/*RSS*/
  printf("datai4_min        [%d]\n",thisHead->datai4_min);
  printf("datai4_max        [%d]\n",thisHead->datai4_max);
  printf("datar4_avg        [%f]\n",thisHead->datar4_avg);
  printf("datar4_std        [%f]\n",thisHead->datar4_std);
  printf("numCols           [%d]\n",thisHead->numCols);
  printf("numRows           [%d]\n",thisHead->numRows);
  printf("numSecs           [%d]\n",thisHead->numSecs);
  printf("mode              [%d]\n",thisHead->mode);
  printf("htype             [%d]\n",thisHead->htype);
  printf("description       [%s]\n",thisHead->descriptionStr);
  printf("time              [%s]\n",thisHead->timeStr);        /*RSS*/
  printf("micrograph        [%s]\n",thisHead->micrographStr);  /*RSS*/
  printf("scanNumber        [%s]\n",thisHead->scanNumberStr);  /*RSS*/
  printf("aoverb            [%f]\n",thisHead->aoverb);
  printf("map_abang         [%f]\n",thisHead->map_abang);
  printf("a_cell            [%f]\n",thisHead->a_cell);
  printf("b_cell            [%f]\n",thisHead->b_cell);
  printf("c_cell            [%f]\n",thisHead->c_cell);
  printf("alp               [%f]\n",thisHead->alp);
  printf("bet               [%f]\n",thisHead->bet);
  printf("gam               [%f]\n",thisHead->gam);
  printf("dela              [%f]\n",thisHead->dela);
  printf("delb              [%f]\n",thisHead->delb);
  printf("delc              [%f]\n",thisHead->delc);
  printf("a_star            [%f]\n",thisHead->a_star);
  printf("b_star            [%f]\n",thisHead->b_star);
  printf("c_star            [%f]\n",thisHead->c_star);
  printf("alp_star          [%f]\n",thisHead->alp_star);
  printf("bet_star          [%f]\n",thisHead->bet_star);
  printf("gam_star          [%f]\n",thisHead->gam_star);
  printf("pixelSize         [%f]\n",thisHead->pixelSize);
  printf("packRadius        [%d]\n",thisHead->packRadius);
  printf("binFactor         [%d]\n",thisHead->binFactor);
  printf("dthe              [%f]\n",thisHead->dthe);
  printf("dphi_90           [%f]\n",thisHead->dphi_90);
  printf("symmetry          [%d]\n",thisHead->symmetry);
  printf("pixSizUnits       [%d]\n",thisHead->pixSizUnits);
  printf("res_out           [%f]\n",thisHead->res_out);
  printf("ctfMode           [%d]\n",thisHead->ctfMode);
  printf("tempFactor        [%f]\n",thisHead->tempFactor);
  printf("wiener            [%f]\n",thisHead->wiener);
  printf("borderType        [%d]\n",thisHead->borderType);
  printf("borderWidth       [%d]\n",thisHead->borderWidth);
  printf("rad_hi            [%d]\n",thisHead->rad_hi);
  printf("rad_lo            [%d]\n",thisHead->rad_lo);
  printf("even              [%d]\n",thisHead->even);
  printf("mrcX              [%d]\n",thisHead->mrcX);
  printf("mrcY              [%d]\n",thisHead->mrcY);
  printf("mrcZ              [%d]\n",thisHead->mrcZ);
  printf("transX            [%d]\n",thisHead->transX);
  printf("transY            [%d]\n",thisHead->transY);
  printf("transZ            [%d]\n",thisHead->transZ);
  printf("isTrunc           [%d]\n",thisHead->isTrunc);
  printf("origX             [%d]\n",thisHead->origX);
  printf("origY             [%d]\n",thisHead->origY);
  printf("origZ             [%d]\n",thisHead->origZ);

  
}/* PIF_LIST_HEAD */




/*************************************************************
 * pif_write_gh() - Write the PIF file's global header.
*/
int pif_write_gh_(forPifHead_t *thisHead)
{/* PIF_WRITE_GH */

  if(thisHead->floatScaleFactor != 0.0)
    thisHead->recipFloatScaleFactor = 1.0 / thisHead->floatScaleFactor;

  return(pif_write_header_(&(thisHead->luNum),
			   &(thisHead->nimg),
			   &(thisHead->recipFloatScaleFactor),
			   &(thisHead->int4scaleF),
			   &(thisHead->htype),
			   &(thisHead->numCols),
			   &(thisHead->numRows),
			   &(thisHead->numSecs),
			   &(thisHead->mode),
			   &(thisHead->endianNess),
			   (thisHead->versionStr),
			   &(thisHead->even),
			   &(thisHead->mrcX),
			   &(thisHead->mrcY),
			   &(thisHead->mrcZ))
	 );

}/* PIF_write_GH */



/***************************************************************
 * pif_write_dh() - Write the PIF file's image header.
*/
int pif_write_dh_(forPifHead_t *thisHead)
{/* PIF_WRITE_DH */
  int dataSize;

  if(thisHead->mode==46)
    dataSize = 4;
  else if(thisHead->mode==20)
    dataSize = 2;
  else if(thisHead->mode==1)
    dataSize = 2;
  else if(thisHead->mode==7)
    dataSize = 2;
  else dataSize = 1;
  
  /* Normal */
  thisHead->imgOffset = 512 + /* the Global header */
    /* The data headers */
    (512 * ((thisHead->imgNum -1))) + 
      /* The images */
      ((thisHead->imgNum -1) *
       thisHead->numCols *
       thisHead->numRows *
       thisHead->numSecs *
       dataSize);

  return(pif_write_data_header_(&(thisHead->luNum),
				&(thisHead->floatScaleFactor),
				&(thisHead->int4scaleF),
				&(thisHead->img_bkg),
				&(thisHead->datai4_min),
				&(thisHead->datai4_max),
				&(thisHead->datar4_avg),
				&(thisHead->datar4_std),
				&(thisHead->numCols),
				&(thisHead->numRows),
				&(thisHead->numSecs),
				&(thisHead->imgOffset),
				&(thisHead->mode),
				&(thisHead->htype),
				(thisHead->descriptionStr),
				(thisHead->timeStr),      /*RSS*/
				(thisHead->micrographStr),/*RSS*/
				(thisHead->scanNumberStr),/*RSS*/
				&(thisHead->aoverb),
				&(thisHead->map_abang),
				&(thisHead->a_cell),
				&(thisHead->b_cell),
				&(thisHead->c_cell),
				&(thisHead->alp),
				&(thisHead->bet),
				&(thisHead->gam),
				&(thisHead->dela),
				&(thisHead->delb),
				&(thisHead->delc),
				&(thisHead->a_star),
				&(thisHead->b_star),
				&(thisHead->c_star),
				&(thisHead->alp_star),
				&(thisHead->bet_star),
				&(thisHead->gam_star),
				&(thisHead->pixelSize),
				(thisHead->t_matrix),
				&(thisHead->packRadius),
				&(thisHead->binFactor),
				&(thisHead->dthe),
				&(thisHead->dphi_90),
				&(thisHead->symmetry),
				&(thisHead->pixSizUnits),
				&(thisHead->res_out),
				&(thisHead->ctfMode),
				&(thisHead->tempFactor),
				&(thisHead->wiener),
				&(thisHead->borderType),
				&(thisHead->borderWidth),
				&(thisHead->datar4_min),
				&(thisHead->datar4_max),
				&(thisHead->rad_lo),
				&(thisHead->rad_hi),
				&(thisHead->transX),
				&(thisHead->transY),
				&(thisHead->transZ),
				&(thisHead->isTrunc),
				&(thisHead->origX),
				&(thisHead->origY),
				&(thisHead->origZ)
				));

}/* PIF_WRITE_DH */




/*************************************************************
 * pif_write_header() - Write the PIF file's global header.
*/
int pif_write_header_(int *luNum,
		      int *nimg, 
		      float *recipFloatScaleFactor,
		      float *int4scaleF,
		      int *htype, int *numCols, int *numRows, int *numSecs,
		      int *mode, int *endianNess, char *versionStr,
		      int *even, int *mrcX, int *mrcY, int *mrcZ)
{/* PIF_WRITE_HEADER */
  int file_id;
  char floatScaleStr[18];
  char junkChar[512];
  char *s1;
  int localInt, i;
  float localFloat;
  FILE *fPIF;

  fPIF = lun[*luNum];
  endians[*luNum] = *endianNess;

  if(fPIF != NULL)
    {/* It is writeable */

      if(fseek(fPIF,0,0)==-1)
	{/* Error */
	  printf("fseek() error.  If VMS have the file attributes changed?");
	  exit(0);
	}/* Error */

      /* [1] Write global file header */
      file_id = 8;                 /* means PIF EMPORTABLE*/
      differentEndian_((BYTE *)&file_id, 4, 0); /* force little-endian */
      fwrite(&file_id,sizeof(int),1,fPIF);
      fwrite(&file_id,sizeof(int),1,fPIF);
      

      sprintf(floatScaleStr,"%16.9E",(*recipFloatScaleFactor));
      fwrite(floatScaleStr,sizeof(char),16,fPIF);
#ifdef DEBUG_THIS
      printf("HEADER [%f][%.12E][%s]\n",
	     *recipFloatScaleFactor,
	     *recipFloatScaleFactor,
	     floatScaleStr);
#endif
      localInt = *nimg;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
/* 28 */      
      /* This is VAX or DECstation */
      localInt = *endianNess;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
      
      sprintf(versionStr,"%s",VERSION); 
      fwrite(versionStr,sizeof(char),32,fPIF);
/* 64 */      
      
      localInt = *htype;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
      localInt = *numCols;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
      localInt = *numRows;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
      localInt = *numSecs;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             

      localInt = *mode;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
      if(*mode==20)
	i2Map = TRUE;
      else i2Map = FALSE;
/* 84 */      

      localInt = *even;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
      localInt = *mrcX;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
      localInt = *mrcY;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
      localInt = *mrcZ;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
/* 100 */      

      sprintf(floatScaleStr,"%16.9E",(*int4scaleF));
      fwrite(floatScaleStr,sizeof(char),16,fPIF);
/* 116 */

      /* Is this really necessary??? */
      for(i=0;i<512;i++)
	junkChar[i]= (char)0;
      fwrite(junkChar,sizeof(char),396,fPIF);

    }/* It is writeable */

  return(0);

}/* PIF_WRITE_HEADER */





/*************************************************************
 * pif_write_image_header() - Write the PIF file's header for a 
 * specific image.
*/
void pif_write_image_header_(int *luNum,
			    int *imgNum,
			    float *floatScaleFactor,
			    float *img_bkg, int *img_min, int *img_max,
			    float *img_avg, float *img_std,
			    int *numCols, int *numRows, int *numSecs,
			    int *imgOffset, int *mode,
			    char *descriptionStr, char *timeStr,
			    char *micrographStr, char *scanNumberStr)
{/* PIF_WRITE_IMAGE_HEADER */
  int file_id,endianNess,htype,i,j,k,l,m,x,y;
  int seekPos,thisNcols,thisNrows,thisNsecs;
  char  floatScaleStr[57];
  char versionStr[33];
  int scanNum;
  char *s1;
  FILE *fPIF;
  int localInt;
  float junkFloat,localFloat;
  int mrc_nxstart,mrc_nystart,mrc_nzstart,
      mrc_mx,mrc_my,mrc_mz,mrc_xlength,mrc_ylength,
      mrc_zlength,mrc_alpha,mrc_beta,mrc_gamma,
      mrc_mapc,mrc_mapr,mrc_maps,mrc_min,mrc_max,
      mrc_mean,mrc_std,mrc_ispg,mrc_nsymbt,mrc_xorigin,
      mrc_yorigin, packRadius;
  char junkChar[512];

  fPIF = lun[*luNum];

  if(fPIF != NULL)
    {/* It is writeable */
      
      if(fseek(fPIF,*imgOffset,0)==-1)
	{/* Error */
	  printf(
	    "IH fseek() error.  If VMS have the file attributes changed?");
	  exit(0);
	}/* Error */

      /* [2] Write header for each image */

      localInt = *numCols;                   /* mrc_mx */
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             

      localInt = *numRows;                   /* mrc_my */
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             

      localInt = *numSecs;                   /* mrc_mz */
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             

      localInt = *mode; 		    /* pifmode */
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
      if(*mode==20)
	i2Map = TRUE;
      else i2Map = FALSE;
/* 16 */
      convertFloat_(&localInt,*img_bkg,*floatScaleFactor);
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(float),1,fPIF);             
/* 20 */
      /* Need a packRadius */
      packRadius = (*numCols)/2;   /* find a better place for this ... */
      localInt = packRadius;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
/* 24 */

      fwrite(&mrc_nxstart,sizeof(int),1,fPIF);        
      fwrite(&mrc_nystart,sizeof(int),1,fPIF);        
      fwrite(&mrc_nzstart,sizeof(int),1,fPIF);        
/* 36 */

      mrc_xlength= 
	mrc_ylength= 
	  mrc_zlength= 
	    mrc_alpha= 
	      mrc_beta= 
		mrc_gamma= 
		  mrc_mapc= 
		    mrc_mapr= 
		      mrc_maps= 0;

      fwrite(&mrc_xlength,sizeof(int),1,fPIF);      
      fwrite(&mrc_ylength,sizeof(int),1,fPIF);      
      fwrite(&mrc_zlength,sizeof(int),1,fPIF);      
      fwrite(&mrc_alpha,sizeof(int),1,fPIF);        
      fwrite(&mrc_beta,sizeof(int),1,fPIF);         
      fwrite(&mrc_gamma,sizeof(int),1,fPIF);        
      fwrite(&mrc_mapc,sizeof(int),1,fPIF);           
      fwrite(&mrc_mapr,sizeof(int),1,fPIF);           
      fwrite(&mrc_maps,sizeof(int),1,fPIF); 
/* 72 */
          
      localInt = *img_min;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
      localInt = *img_max;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
/* 80 */

      convertFloat_(&mrc_mean,*img_avg,*floatScaleFactor);
      convertFloat_(&mrc_std,*img_std,*floatScaleFactor);
      localInt = mrc_mean;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
      localInt = mrc_std;
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fPIF);             
/* 88 */
      
      mrc_ispg= 
	mrc_nsymbt= 0;
      mrc_xorigin= 
	mrc_yorigin= 0.0;
      fwrite(&mrc_ispg,sizeof(int),1,fPIF);           
      fwrite(&mrc_nsymbt,sizeof(int),1,fPIF);         
      fwrite(&mrc_xorigin,sizeof(float),1,fPIF);      
      fwrite(&mrc_yorigin,sizeof(float),1,fPIF);      
/* 104 */
      
      fwrite(descriptionStr,sizeof(char),80,fPIF);
      fwrite(timeStr,sizeof(char),32,fPIF);
      fwrite(micrographStr,sizeof(char),16,fPIF);
      fwrite(scanNumberStr,sizeof(char),8,fPIF);
/* 240 */
      
      for(i=0;i<512;i++)
	junkChar[i]= (char)0;
      fwrite(junkChar,sizeof(char),260,fPIF); /* Do we need this? */
      
    }/* It is writeable */
  else
    {/* error */
	printf("[PIF_WRITE_IMAGE_HEADER] Image [%d] unwriteable.\n",*imgNum);/*RSS*/
    }/* error */
	  
}/* PIF_WRITE_IMAGE_HEADER */



/*************************************************************
 * pif_write_data_header() - Write the PIF file's header for a 
 * specific image/MAP.
*/
int pif_write_data_header_(int *luNum,
			   float *floatScaleFactor,
			   float *int4scaleF,
			   float *img_bkg, int *img_min, int *img_max,
			   float *img_avg, float *img_std,
			   int *numCols, int *numRows, 
			   int *numSecs,
			   int *imgOffset, int *pifMode, int *htype,
			   char *descriptionStr,
			   char *timeStr,       /*RSS*/
			   char *micrographStr, /*RSS*/
			   char *scanNumberStr, /*RSS*/
			   float *aoverb, float *map_abang,
			   float *a_cell,float *b_cell,
			   float *c_cell, float *alp, float *bet,
			   float *gam, float *dela, float *delb, float *delc,
			   float *a_star, float *b_star, float *c_star,
			   float *alp_star,float *bet_star,float *gam_star,
			   float *pixelSize,
			   int *t_matrix, int *packRadius, int *binFactor,
			   float *dthe, float *dphi_90, int *symmetry,
			   int *pixSizUnits, float *res_out,
			   int *ctfMode, float *tempFactor,
			   float *wiener,
			   int *borderType, int *borderWidth,
			   float *r4_min, float *r4_max,
			   int *rad_lo, int *rad_hi,
			   int *transX, int *transY, int *transZ, int *isTrunc,
			   int *origX, int *origY, int *origZ)
{/* PIF_WRITE_DATA_HEADER */
  int  i,j,k;
  long seekPos, imageBytes;  
  int  thisNcols,thisNrows,thisNsecs;
  char floatScaleStr[57];
  char versionStr[33];
  /*char timeStr[36];*/ /*RSS*/
  /*char micrographStr[200];*/ /*RSS*/
  int  scanNum;
  /*char scanNumberStr[210];*/ /*RSS*/
  char descStr[100];
  char *s1;
  float junkFloat, thisFSF;
  int  junkInt,localInt;
  int  numRead;
  int mrc_nxstart,mrc_nystart,mrc_nzstart,
      mrc_mx,mrc_my,mrc_mz,mrc_xlength,mrc_ylength,
      mrc_zlength,mrc_alpha,mrc_beta,mrc_gamma,
      mrc_mapc,mrc_mapr,mrc_maps,mrc_min,mrc_max,
      mrc_mean,mrc_std,mrc_ispg,mrc_nsymbt,mrc_xorigin,
      mrc_yorigin;
  int aoverbI, map_abangI, delaI, delbI, delcI;
  FILE *fOut;

  /* Init these out */
  mrc_nxstart = mrc_nystart = mrc_nzstart = 
    mrc_mx = mrc_my = mrc_mz = mrc_xlength = mrc_ylength = 
      mrc_zlength = mrc_alpha = mrc_beta = mrc_gamma = 
	mrc_mapc = mrc_mapr = mrc_maps = mrc_min = mrc_max = 
	  mrc_mean = mrc_std = mrc_ispg = mrc_nsymbt = mrc_xorigin = 
	    mrc_yorigin = 0;
  aoverbI = map_abangI = delaI = delbI = delcI = 0;

  fOut = lun[*luNum];    /* a */

  seekPos = *imgOffset;
      
  /* fseek(fOut,seekPos,0); */
  if(fseek(fOut,seekPos,0)==-1)
    {/* Error */
      printf(
	     "DH fseek() error.  If VMS have the file attributes changed?");
      exit(0);
    }/* Error */

  thisFSF = *floatScaleFactor;
  if(*pifMode == 20)
    {/* We've got to get header values like pixelSize in greater precision */
      thisFSF = *int4scaleF;
    }/* We've got to get header values like pixelSize in greater precision */

  thisNcols = *numCols;
  thisNrows = *numRows;
  thisNsecs = *numSecs;
  differentEndian_((BYTE *)&thisNcols, 4, endians[*luNum]);
  OKREAD(fwrite(&thisNcols, sizeof(int),1,fOut),1,"PIF_WRITE_DATA_HEADER");
  differentEndian_((BYTE *)&thisNrows, 4, endians[*luNum]);
  OKREAD(fwrite(&thisNrows, sizeof(int),1,fOut),1,"PIF_WRITE_DATA_HEADER");
  differentEndian_((BYTE *)&thisNsecs, 4, endians[*luNum]);
  OKREAD(fwrite(&thisNsecs, sizeof(int),1,fOut),1,"PIF_WRITE_DATA_HEADER");
/* 12 */

  /* [2] write header for this image */
  localInt = *pifMode;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  if(*pifMode==20)
    i2Map = TRUE;
  else i2Map = FALSE;
  convertFloat_(&localInt,*img_bkg,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
/* 20 */

  localInt = *packRadius;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             

  fwrite(&mrc_nxstart,sizeof(int),1,fOut);  /* Not used as of 12-17-97 */      
  fwrite(&mrc_nystart,sizeof(int),1,fOut);        
  fwrite(&mrc_nzstart,sizeof(int),1,fOut);        
  fwrite(&mrc_mx,sizeof(int),1,fOut);             
  fwrite(&mrc_my,sizeof(int),1,fOut);             
  fwrite(&mrc_mz,sizeof(int),1,fOut);             
/* 48 */

  /*fwrite(&mrc_xlength,sizeof(int),1,fOut);      
  fwrite(&mrc_ylength,sizeof(int),1,fOut);      
  fwrite(&mrc_zlength,sizeof(int),1,fOut);       */
  convertFloat_(&localInt,*a_cell,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*b_cell,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*c_cell,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
/* 68 */

  /* fwrite(&mrc_alpha,sizeof(int),1,fOut);        
     fwrite(&mrc_beta,sizeof(int),1,fOut);         
     fwrite(&mrc_gamma,sizeof(int),1,fOut);         */
  convertFloat_(&localInt,*alp,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*bet,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*gam,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             

  fwrite(&mrc_mapc,sizeof(int),1,fOut);          /* Not used as of 12-17-97 */ 
  fwrite(&mrc_mapr,sizeof(int),1,fOut);           
  fwrite(&mrc_maps,sizeof(int),1,fOut);           
/* 92 */

  /* fwrite(&mrc_min,sizeof(int),1,fOut);          
     fwrite(&mrc_max,sizeof(int),1,fOut);          
     fwrite(&mrc_mean,sizeof(int),1,fOut);         
     fwrite(&mrc_std,sizeof(int),1,fOut);        */
#ifdef NEEDTOWRITE_THESE_AS_FLOAT
  convertFloat_(&localInt,*img_min,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*img_max,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);       
#else
  localInt = *img_min;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  localInt = *img_max;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);       
#endif /* #ifdef NEEDTOWRITE_THESE_AS_FLOAT */
  convertFloat_(&localInt,*img_avg,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*img_std,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             

  fwrite(&mrc_ispg,sizeof(int),1,fOut);          /* Not used as of 12-17-97 */ 
  fwrite(&mrc_nsymbt,sizeof(int),1,fOut);         
  fwrite(&mrc_xorigin,sizeof(float),1,fOut);      
  fwrite(&mrc_yorigin,sizeof(float),1,fOut);      
/* 124 */
      
  memcpy(descStr,descriptionStr,72);
  descStr[72] = '\0';
  fwrite(descStr,sizeof(char),80,fOut);

  memcpy(descStr,timeStr,32);           /*RSS*/
  descStr[31] = '\0';                   /*RSS*/
  fwrite(descStr,sizeof(char),32,fOut); /*RSS*/

  memcpy(descStr,micrographStr,16);     /*RSS*/
  descStr[15] = '\0';                   /*RSS*/
  fwrite(descStr,sizeof(char),16,fOut); /*RSS*/

  memcpy(descStr,scanNumberStr,8);      /*RSS*/
  descStr[7] = '\0';                    /*RSS*/
  fwrite(descStr,sizeof(char),8,fOut);  /*RSS*/
/* 260 */
      
  /* fwrite(&aoverbI,sizeof(int),1,fOut);        
     fwrite(&map_abangI,sizeof(int),1,fOut);         
     fwrite(&delaI,sizeof(int),1,fOut);        
     fwrite(&delbI,sizeof(int),1,fOut);        
     fwrite(&delcI,sizeof(int),1,fOut);         */
  convertFloat_(&localInt,*aoverb,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*map_abang,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*dela,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*delb,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*delc,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
/* 280 */


#ifdef OLDWAY
  fwrite(t_matrix,sizeof(int),6,fOut);        
#else
  for(i=0;i<6;i++)
    {
      localInt = t_matrix[i];
      differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
      fwrite(&localInt,sizeof(int),1,fOut);        
    }
#endif

  convertFloat_(&localInt,*dthe,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*dphi_90,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             


  localInt = *symmetry;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             

  localInt = *binFactor;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             

  convertFloat_(&localInt,*a_star,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*b_star,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*c_star,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*alp_star,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*bet_star,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*gam_star,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*pixelSize,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             

  localInt = *pixSizUnits;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*res_out,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  localInt = *ctfMode;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*tempFactor,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*wiener,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  localInt = *borderType;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  localInt = *borderWidth;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             

  convertFloat_(&localInt,*r4_min,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,*r4_max,thisFSF);
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);       

  localInt = *rad_lo;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  localInt = *rad_hi;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             

  localInt = *transX;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  localInt = *transY;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  localInt = *transZ;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  localInt = *isTrunc;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  localInt = *origX;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  localInt = *origY;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  localInt = *origZ;
  differentEndian_((BYTE *)&localInt, 4, endians[*luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             

  return(0); 		/* no error	   */

}/* PIF_WRITE_DATA_HEADER */



/**************************************************************
 * pif_write_debugi4() - Take a int*4 image buffer and write a
 * PIF image file from it.
*/
void pif_write_debugi4_(char *fName,
			float *float_buffer,
			int *ncol,
			int *nrow)
{/* PIF_WRITE_DEBUGI4 */
  int i,j,k, imgOffset;
  float maxP,minP,scaleF;
  forPifHead_t thisHead;
  BYTE *byte_buffer;

  byte_buffer = malloc(sizeof(BYTE)*(*ncol)*(*nrow));

  minP = maxP = float_buffer[0];
  for(j=0;j<(*nrow);j++)
    {
      for(k=0;k<(*ncol);k++)
	{
	  maxP = MAX(maxP,float_buffer[(j*(*ncol))+k]);
	  minP = MIN(minP,float_buffer[(j*(*ncol))+k]);
	}
    }
  scaleF = 254.0 / (maxP-minP);
  printf("min/max/scale [%f/%f/%f]\n",minP,maxP,scaleF);

  maxP = 0.0;
  for(j=0;j<(*nrow);j++)
    {
      for(k=0;k<(*ncol);k++)
	{
	  byte_buffer[(j*(*ncol))+k]= (BYTE)((float_buffer[(j*(*ncol))+k]-
					      minP)*scaleF);
	  maxP = MAX(maxP,float_buffer[(j*(*ncol))+k]);
	}
    }
  printf("max[%f]\n",maxP);
  
  pif_init_head(&thisHead);
  thisHead.luNum = 8;
  thisHead.numCols = *ncol;
  thisHead.numRows = *nrow;
  thisHead.numSecs = 1;
  thisHead.mode = 0;
  thisHead.datai4_max = (int)(maxP+0.5);;
  pif_open(fName,&thisHead.luNum, "w");
  pif_write_gh(&thisHead);
  pif_write_dh(&thisHead);
  imgOffset = 1024;
  pif_write_byte_image(&(thisHead.luNum),
		       byte_buffer,
		       &imgOffset,
		       &(thisHead.numCols),			   
		       &(thisHead.numRows),			   
		       &(thisHead.numSecs)
		       );
  pif_close(&(thisHead.luNum));

}/* PIF_WRITE_DEBUGI4 */
void pif_write_debugi4(char *fName,
		       float *float_buffer,
		       int *ncol,
		       int *nrow)
{
  pif_write_debugi4_(fName,
		     float_buffer,
		     ncol,
		     nrow);
}



void pif_write_prji1_(forPifHead_t *prjHead,
		      float *float_buffer,
		      int *id,
		      int *numCols, int *numRows,
		      float *sFac, float *min)
{/* PIF_WRITE_PRJI1 */
  int i,j,k,imgOffset,yOut,offset,numSecs,ncol,nrow;
  BYTE *byte_buffer;
  float minP,maxP,thisFac;

  thisFac = *sFac;
  minP = *min;
  ncol = *numCols;
  nrow = *numRows;
  byte_buffer = (BYTE *)malloc((nrow * ncol)*sizeof(BYTE));
  if(!byte_buffer)
    {/* CAnt go on */
      printf("[PIF_WRITE_PRJI1] Cannot allocate [%d]\n",(nrow * ncol));
      exit(0);
    }/* CAnt go on */

  if(prjHead->imgNum==1)
    {/* Do this from the first one */
      minP = maxP = float_buffer[0];
      for(i=0;i<nrow;i++)
	{
	  yOut = i * (ncol);
	  for(j=0;j<ncol;j++)
	    {/* Do all */
	      offset = yOut+j;
	      minP = MIN(float_buffer[offset],minP);
	      maxP = MAX(float_buffer[offset],maxP);
	    }/* Do all */
	}
#ifdef USETHIS
      if(fabs(minP) > maxP)
	maxP = 0.0 - minP;
      else
	minP = 0.0 - maxP;
      minP *= 1.3;
      maxP *= 1.3;
#endif
      thisFac = *sFac = 250.0 / (maxP-minP);
      *min = minP;
    }/* Do this from the first one */

  for(i=0;i<nrow;i++)
    {
      yOut = i * (ncol);
      for(j=0;j<ncol;j++)
	{/* Do all */
	  offset = yOut+j;
	  byte_buffer[offset]= (BYTE)((float_buffer[offset]-minP)
				      * thisFac);
	}/* Do all */
    }
  
  prjHead->imgNum    = *id;
  prjHead->img_bkg   = byte_buffer[0];
  /* Need a packRadius */
  prjHead->packRadius = (*numCols)/2;
  prjHead->numCols   = ncol;
  prjHead->numRows   = nrow;
  pif_write_dh (prjHead);
  
  prjHead->imgOffset += 512;
  numSecs = 1;
  pif_write_byte_image_(&(prjHead->luNum),
			byte_buffer,
			&(prjHead->imgOffset),
			numCols, numRows, &numSecs);

  free(byte_buffer);

}/* PIF_WRITE_PRJI1 */



/*************************************************************
 * pif_write_byte_image() - Write the PIF file's BYTE image for a 
 * specific image.
*/
void pif_write_byte_image_(int *luNum,
			  BYTE *byte_buffer,
			  int *imgOffset,
			  int *numCols, int *numRows, int *numSecs)
{/* PIF_WRITE_BYTE_IMAGE */
  FILE *fPIF;
  int i;

  fPIF = lun[*luNum];

  fseek(fPIF,*imgOffset,0);
  fwrite(byte_buffer,sizeof(BYTE),
	 ((*numCols)*(*numRows)*(*numSecs)),
	 fPIF);

}/* PIF_WRITE_BYTE_IMAGE */




/*************************************************************
 * pif_write_byteshort_image() - Write the PIF file's SHORT image 
 * in BYTE on disk for a specific image.
*/
void pif_write_byteshort_image_(int *luNum,
				short *short_buffer,
				int *imgOffset,
				int *numCols, int *numRows, int *numSecs)
{/* PIF_WRITE_BYTESHORT_IMAGE */
  FILE *fPIF;
  BYTE *byte_buffer;
  int i,j,k,ncol,nrow,nsec,rowOut;
  float scaleFactor,minF,maxF;

  nsec = *numSecs;
  nrow = *numRows;
  ncol = *numCols;
  minF = maxF = (float)short_buffer[0];
  for(i=0;i<nsec;i++)
    {
      for(j=0;j<nrow;j++)
	{
	  rowOut = (i*ncol*nrow)+(j*ncol);
	  for(k=0;k<ncol;k++)
	    {
	      minF = MIN(minF,(float)short_buffer[rowOut+k]);
	      maxF = MAX(maxF,(float)short_buffer[rowOut+k]);
	    }
	}
    }
  if(minF != maxF)
    scaleFactor = 254.0 / (maxF-minF);
  else scaleFactor = 1.0;

  byte_buffer = (BYTE *)malloc(sizeof(BYTE)*(ncol*nrow*nsec));
  for(i=0;i<nsec;i++)
    {
      for(j=0;j<nrow;j++)
	{
	  rowOut = (i*ncol*nrow)+(j*ncol);
	  for(k=0;k<ncol;k++)
	    {
	      byte_buffer[rowOut+k]= (BYTE)(((float)short_buffer[rowOut-k]-
					     minF) *
					    scaleFactor);
	    }
	}
    }

  fPIF = lun[*luNum];

  fseek(fPIF,*imgOffset,0);
  fwrite(byte_buffer,sizeof(BYTE),
	 ((*numCols)*(*numRows)*(*numSecs)),
	 fPIF);

  free(byte_buffer);

}/* PIF_WRITE_BYTESHORT_IMAGE */




/*******************************************************************
 * pif_write_short_image() - Write the PIF file's SHORT image for a 
 * specific image.
*/
void pif_write_short_image_(int *luNum,
			    unsigned short *short_buffer,
			    int *imgOffset,
			    int *numCols, int *numRows, int *numSecs)
{/* PIF_WRITE_SHORT_IMAGE */
  int i,j,k;
  unsigned short *local_short_buffer;
  FILE *fPIF;

  fPIF = lun[*luNum];

  fseek(fPIF,*imgOffset,0);

  local_short_buffer = (unsigned  short *)malloc(sizeof(unsigned short)*
						(
						 (*numCols) *
						 (*numRows) *
						 (*numSecs) 
						 ));
  for(i=0;i<(*numSecs);i++)
    for(j=0;j<(*numRows);j++)
      for(k=0;k<(*numCols);k++)
	{/* Do each one */
	  local_short_buffer[((i*(*numRows)*(*numCols))+
			      (j*(*numCols))+k)] =
				short_buffer[((i*(*numRows)*(*numCols))+
					      (j*(*numCols))+k)];
	  differentEndian_((BYTE *)&local_short_buffer[(
						(i*(*numRows)*(*numCols))+
						       (j*(*numCols))+k)], 2, endians[*luNum]);
	}/* Do each one */

  fwrite(local_short_buffer,sizeof(unsigned short),
	 ((*numCols)*(*numRows)*(*numSecs)),
	 fPIF);

  free(local_short_buffer);

}/* PIF_WRITE_SHORT_IMAGE */



void pif_write_rowi4(forPifHead_t *thisHead,  float *float_buffer, 
		      int *nrot,  int *nannuli,
		      int *whichAnnulus, int *whichImg)
{
  pif_write_rowi4_(thisHead, float_buffer, nrot, nannuli, 
		   whichAnnulus, whichImg);
}

/*******************************************************************
 * pif_write_rowi4() - Write this INTEGER*4 float to disk one 
 * row at a time.
*/
void pif_write_rowi4_(forPifHead_t *thisHead, 
		      float *float_buffer, 
		      int *nrot,
		      int *nannuli,
		      int *whichAnnulus,
		      int *whichImg)
{/* PIF_WRITE_ROWI4 */
  int imgOffset,thisAnnulus,nsec,thisNrot;

  imgOffset =  1024 + (((*whichAnnulus)-1)*(*nrot/4)*4) + 
    (((*whichImg)-1)*(*nrot/4)*(*nannuli)*4);

  nsec = 1;
  thisAnnulus = 1;
  thisNrot = *nrot/4;
  pif_write_map_fltint (&(thisHead->luNum), 
			float_buffer, 
			&imgOffset, 
			&thisNrot, 
			&thisAnnulus, 
			&nsec, 
			&(thisHead->floatScaleFactor));

}/* PIF_WRITE_ROWI4 */



void pif_write_seci4(forPifHead_t *thisHead,
			  float *float_buffer,
			  int *ncol,
			  int *nrow,
			  int *secNum)
{
  pif_write_seci4_(thisHead, float_buffer, ncol, nrow, secNum);
}
#ifdef OLDWAY_BEFORE_YJI
/********************************************************************
 * pif_write_seci4_() - Just write this section of the I4 map.
*/
void pif_write_seci4_(forPifHead_t *thisHead,
		      float *float_buffer,
		      int *ncol,
		      int *nrow,
		      int *secNum)
{/* PIF_WRITE_SECI4_ */
  int thisBin,nsec;

  /* Check here does the ncol/nrow passed in agree with
   * the thisHead->binFactor??? if not pass a new bin factor to 
   * the writing rotuines... */
  if(thisHead->numCols == *ncol &&
     thisHead->numRows == *nrow )
    thisBin = 1;    /* All is at largest size */
  else   if(thisHead->numCols < *ncol ||
	    thisHead->numRows < *nrow )
    {/* Error */
      printf(
       "PIF_WRITE_SEC4 - I don't know how to do this [%d][%d][%d][%d][%d]\n",
	     thisHead->numCols,
	     thisHead->numRows,
	     *ncol,*nrow,
	     thisHead->binFactor);
      exit(0);
    }/* Error */
  else
    thisBin = 2; /* They want to make it smaller */

  if(thisHead->binFactor==2)
    {/* They want to sample average bin the data */

      if(thisBin == 1)
	thisBin = 2;        /* Input array is large but let's BIN it down */
      else
	thisBin = 1;        /* Input array is small already */

    }/* They want to sample average bin the data */

  if(thisBin == 2)
    {/* Don't know how to do this yet ... */
      printf("PIF_WRITE_SECI4() BIN not implemented on write yet... \n");
      exit(0);
    }/* Don't know how to do this yet ... */

  nsec = 1;  /* Just one section */
  pif_write_mapi4_(&(thisHead->luNum),
		   float_buffer,
		   ncol,
		   nrow,
		   &nsec,
		   secNum,
		   &(thisHead->floatScaleFactor),
		   &thisBin);

}/* PIF_WRITE_SECI4_ */
#endif  /* OLDWAY_BEFORE_YJI*/
/********************************************************************
 * pif_write_seci4_() - Just write this section of the I4 map.
*/
void pif_write_seci4_(forPifHead_t *thisHead,
		      float *float_buffer,
		      int *ncol,
		      int *nrow,
		      int *secNum)
{/* PIF_WRITE_SECI4_ */
  int thisBin;

  /* Check here does the ncol/nrow passed in agree with
   * the thisHead->binFactor??? if not pass a new bin factor to 
   * the writing rotuines... */
  if(thisHead->numCols == *ncol &&
     thisHead->numRows == *nrow )
    thisBin = 1;    /* All is at largest size */
  else   if(thisHead->numCols < *ncol ||
	    thisHead->numRows < *nrow )
    {/* Error */
      printf(
       "PIF_WRITE_SEC4 - I don't know how to do this [%d][%d][%d][%d][%d]\n",
	     thisHead->numCols,
	     thisHead->numRows,
	     *ncol,*nrow,
	     thisHead->binFactor);
      exit(1);
    }/* Error */
  else
    thisBin = 2; /* They want to make it smaller */

  if(thisHead->binFactor==2)
    {/* They want to sample average bin the data */

      if(thisBin == 1)
	thisBin = 2;        /* Input array is large but let's BIN it down */
      else
	thisBin = 1;        /* Input array is small already */

    }/* They want to sample average bin the data */

  if(thisBin == 2)
    {/* Don't know how to do this yet ... */
      printf("PIF_WRITE_SECI4() BIN not implemented on write yet... \n");
      exit(1);
    }/* Don't know how to do this yet ... */

  pif_write_map_section4_(&(thisHead->luNum),
                         float_buffer,
                         secNum,
                         ncol,
                         nrow,
                         &(thisHead->floatScaleFactor));
}/* PIF_WRITE_SECI4_ */





void pif_write_wmapi4(forPifHead_t *thisHead,
		      float *float_buffer, 
		      int *ncol, 
		      int *nrow, 
		      int *nsec)
{
  pif_write_wmapi4_(thisHead, float_buffer, ncol, nrow, nsec);
}

/****************************************************************
 * pif_write_wmapi4() - Write entire I2 map.
*/
void pif_write_wmapi4_(forPifHead_t *thisHead,
		       float *float_buffer, 
		       int *ncol, 
		       int *nrow, 
		       int *nsec)
{/* PIF_WRITE_WMAPI4 */
  int thisBin,startSec;
  if(*ncol == thisHead->numCols &&
     *nrow == thisHead->numRows &&
     *nsec == thisHead->numSecs )
    thisBin = 1;
  else if(*ncol > thisHead->numCols &&
	  *nrow > thisHead->numRows &&
	  *nsec > thisHead->numSecs )
    {/* Error */
      printf(
       "PIF_WRITE_WMAPI4 - I don't know how to do this [%d][%d][%d][%d][%d]\n",
	     thisHead->numCols,
	     thisHead->numRows,
	     *ncol,*nrow,
	     thisHead->binFactor);
      exit(0);
    }/* Error */
  else thisBin = 2;


  if(thisHead->binFactor==2)
    {/* They want to sample average bin the data */

      if(thisBin == 1)
	thisBin = 2;        /* Input array is large but let's BIN it down */
      else
	thisBin = 1;        /* Input array is small already */

    }/* They want to sample average bin the data */

  if(thisBin == 2)
    {/* Don't know how to do this yet ... */
      printf("PIF_WRITE_WMAPI4() BIN not implemented on write yet... \n");
      exit(0);
    }/* Don't know how to do this yet ... */

  startSec = 1;
  pif_write_mapi4_(&(thisHead->luNum),
		   float_buffer, 
		   ncol,nrow,nsec,
		   /* 		   &startSec, */
		   &(thisHead->floatScaleFactor),
		   &thisBin);
}/* PIF_WRITE_WMAPI4 */



#ifdef OLDWAY_BEFORE_YJI
/****************************************************************
 * pif_write_mapi4() - Write I4 map.
*/
void pif_write_mapi4_(int *luNum,
		      float *float_buffer, 
		      int *numCols, 
		      int *numRows, 
		      int *numSecs,
		      int *sectionStart,
		      float *floatScaleFactor,
		      int *bin)
{/* PIF_WRITE_MAPI4 */
  int mapOffset;
  
  /* Times 4 because we're INT here */
  mapOffset = 1024 + ((*sectionStart-1)*(*numCols)*(*numRows)*4); 
  
  pif_write_map_fltint_(luNum,
			float_buffer,
			&mapOffset,
			numCols,
			numRows,
			numSecs,
			floatScaleFactor);

}/* PIF_WRITE_MAPI4 */
#endif  /* OLDWAY_BEFORE_YJI  */
/****************************************************************
 * pif_write_mapi4() - Write I4 map. added by yji
*/
void pif_write_mapi4_(int *luNum,
                      float *float_buffer,
                      int *numCols,
                      int *numRows,
                      int *numSecs,
                      float *floatScaleFactor,
                      int *bin)
{/* PIF_WRITE_MAPI4 */
  int mapOffset;

  mapOffset = 1024;  /* Hard code for now... */

  pif_write_map_fltint_(luNum,
                        float_buffer,
                        &mapOffset,
                        numCols,
                        numRows,
                        numSecs,
                        floatScaleFactor);
}/* PIF_WRITE_MAPI4 */









void pif_write_wmapi2(forPifHead_t *thisHead,
		       float *float_buffer, 
		       int *ncol, 
		       int *nrow, 
		       int *nsec)
{
  pif_write_wmapi2_(thisHead, float_buffer, ncol, nrow, nsec);
}

/****************************************************************
 * pif_write_wmapi2() - Write entire I2 map.
*/
void pif_write_wmapi2_(forPifHead_t *thisHead,
		       float *float_buffer, 
		       int *ncol, 
		       int *nrow, 
		       int *nsec)
{/* PIF_WRITE_WMAPI2 */
  int thisBin;
  if(*ncol == thisHead->numCols &&
     *nrow == thisHead->numRows &&
     *nsec == thisHead->numSecs )
    thisBin = 1;
  else if(*ncol > thisHead->numCols &&
	  *nrow > thisHead->numRows &&
	  *nsec > thisHead->numSecs )
    {/* Error */
      printf(
       "PIF_WRITE_WMAPI2 - I don't know how to do this [%d][%d][%d][%d][%d]\n",
	     thisHead->numCols,
	     thisHead->numRows,
	     *ncol,*nrow,
	     thisHead->binFactor);
      exit(0);
    }/* Error */
  else thisBin = 2;


  if(thisHead->binFactor==2)
    {/* They want to sample average bin the data */

      if(thisBin == 1)
	thisBin = 2;        /* Input array is large but let's BIN it down */
      else
	thisBin = 1;        /* Input array is small already */

    }/* They want to sample average bin the data */

  if(thisBin == 2)
    {/* Don't know how to do this yet ... */
      printf("PIF_WRITE_WMAPI2() BIN not implemented on write yet... \n");
      exit(0);
    }/* Don't know how to do this yet ... */

  pif_write_mapi2_(&(thisHead->luNum),
		    float_buffer, 
		    ncol,nrow,nsec,
		    &(thisHead->floatScaleFactor),
		    &thisBin);
}/* PIF_WRITE_WMAPI2 */


/****************************************************************
 * pif_write_mapi2() - Write entire I2 map.
*/
void pif_write_mapi2_(int *luNum,
		      float *float_buffer, 
		      int *numCols, 
		      int *numRows, 
		      int *numSecs,
		      float *floatScaleFactor,
		      int *bin)
{/* PIF_WRITE_MAPI2 */
  int mapOffset;

  mapOffset = 1024;  /* Hard code for now... */

  pif_write_map_float_image_(luNum,
			     float_buffer,
			     &mapOffset,
			     numCols,
			     numRows,
			     numSecs,
			     floatScaleFactor);
}/* PIF_WRITE_MAPI2 */





void pif_write_seci2(forPifHead_t *thisHead,
			  float *float_buffer,
			  int *ncol,
			  int *nrow,
			  int *secNum)
{
  pif_write_seci2_(thisHead, float_buffer, ncol, nrow, secNum);
}
/********************************************************************
 * pif_write_seci2_() - Just write this section of the map.
*/
void pif_write_seci2_(forPifHead_t *thisHead,
			  float *float_buffer,
			  int *ncol,
			  int *nrow,
			  int *secNum)
{/* PIF_WRITE_SECI2_ */
  int thisBin;

  /* Check here does the ncol/nrow passed in agree with
   * the thisHead->binFactor??? if not pass a new bin factor to 
   * the writing rotuines... */
  if(thisHead->numCols == *ncol &&
     thisHead->numRows == *nrow )
    thisBin = 1;    /* All is at largest size */
  else   if(thisHead->numCols < *ncol ||
	    thisHead->numRows < *nrow )
    {/* Error */
      printf(
       "PIF_WRITE_SEC2 - I don't know how to do this [%d][%d][%d][%d][%d]\n",
	     thisHead->numCols,
	     thisHead->numRows,
	     *ncol,*nrow,
	     thisHead->binFactor);
      exit(0);
    }/* Error */
  else
    thisBin = 2; /* They want to make it smaller */

  if(thisHead->binFactor==2)
    {/* They want to sample average bin the data */

      if(thisBin == 1)
	thisBin = 2;        /* Input array is large but let's BIN it down */
      else
	thisBin = 1;        /* Input array is small already */

    }/* They want to sample average bin the data */

  if(thisBin == 2)
    {/* Don't know how to do this yet ... */
      printf("PIF_WRITE_SECI2() BIN not implemented on write yet... \n");
      exit(0);
    }/* Don't know how to do this yet ... */

  pif_write_map_section_(&(thisHead->luNum),
			 float_buffer,
			 secNum,
			 ncol,
			 nrow,
			 &(thisHead->floatScaleFactor));

}/* PIF_WRITE_SECI2_ */





/***************************************************************
 * pif_write_map_short_image() - Write the PIF file's SHORT image 
 * for a floating point MAP.
*/
void pif_write_map_short_image_(int *luNum,
				short *short_buffer,
				int *mapOffset,
				int *numCols)
{/* PIF_WRITE_MAP_SHORT_IMAGE */
  int i;
  FILE *fIn;

  fIn = lun[*luNum];

  fseek(fIn,*mapOffset,0);
  for(i=0;i<(*numCols);i++)
      differentEndian_((BYTE *)&short_buffer[i], 2, endians[*luNum]);
  OKREAD(fwrite(short_buffer,sizeof(short),(*numCols),fIn),
	     *numCols,
	     "PIF_WRITE_MAP_SHORT_IMAGE");

}/* PIF_WRITE_MAP_SHORT_IMAGE */



/***************************************************************
 * pif_write_map_section() - Write the PIF file's FLOAT image 
 * for this section.  NOTE:  This rtn assumes there is only 
 * 1 map image in the entire PIF file and this one's it.  Also
 * assumes you're writing in int*2 on disk. 
*/
void pif_write_map_section_(int *luNum,
			    float *float_buffer,
			    int *secNum,
			    int *numCols,
			    int *numRows,
			    float *floatScaleFactor)
{/* PIF_WRITE_MAP_SECTION */
  int mapOffset, numSecs, thisSecNum;

  thisSecNum = *secNum - 1;  /* Fortran numbers from 1 not 0 */
  numSecs = 1;
  /* Times 2 because they're shorts here */
  mapOffset = 1024 + (2 * thisSecNum * (*numRows) * (*numCols));

  pif_write_map_float_image_(luNum,
			     float_buffer,
			     &mapOffset,
			     numCols,
			     numRows,
			     &numSecs,
			     floatScaleFactor);  

}/* PIF_WRITE_MAP_SECTION */



/***************************************************************
 * pif_write_map_float_image() - Write the PIF file's FLOAT image 
 * for entire floating point MAP.
*/
void pif_write_map_float_image_(int *luNum,
				float *float_buffer,
				int *mapOffset,
				int *numCols,
				int *numRows,
				int *numSecs,
				float *floatScaleFactor)
{/* PIF_WRITE_MAP_FLOAT_IMAGE */
  int i,j,k,thisSec,thisRow;
  FILE *fOut;
  short *thisShort;
  int thisInt;
  float thisFloat;
  double thisDiff;

  fOut = lun[*luNum];

  thisShort = (short *)malloc(sizeof(short)*(*numCols+2));

  fseek(fOut,(*mapOffset),0);

  for(i=0;i<*numSecs;i++)
    {/* secs */
      thisSec = i * (*numRows) * (*numCols);
      for(j=0;j<*numRows;j++)
	{/* rows */
	  /* It's already in TSB format  */
	  thisRow = (j) * (*numCols);

	  for(k=0;k<*numCols;k++)
	    {/* Do all */
	      thisFloat = (float)(float_buffer[(thisSec + 
						thisRow + k)]*
				  (*floatScaleFactor));
	      /* Mimic the fortran nint() rtn */
	      thisInt = (int)thisFloat;
	      thisDiff = (double)((double)thisFloat - (double)(thisInt));

	      if(thisDiff >= 0.5)
		{/* rounnd it */
		  thisShort[k] = (short)thisInt + 1;
		}/* rounnd it */
	      else if(thisDiff <= -0.5)
		{/* rounnd it */
		  thisShort[k] = (short)thisInt - 1;
		}/* rounnd it */
	      else thisShort[k] = (short)thisInt;
	      differentEndian_((BYTE *)&thisShort[k], 2, endians[*luNum]);
	    }/* Do all */

	  OKREAD(fwrite(thisShort,sizeof(short),(*numCols),fOut),
		 *numCols,
		 "PIF_WRITE_MAP_SHORT_IMAGE");
	}/* rows */
    }/* secs */

  free(thisShort);

}/* PIF_WRITE_MAP_FLOAT_IMAGE */



/***************************************************************
 * pif_write_map_fltint() - Write the PIF file's FLOAT image 
 * for entire floating point MAP.  Convert floats to ints not
 * shorts as is done in pif_write_map_float_image() above.
*/
void pif_write_map_fltint_(int *luNum,
			   float *float_buffer,
			   int *mapOffset,
			   int *numCols,
			   int *numRows,
			   int *numSecs,
			   float *floatScaleFactor)
{/* PIF_WRITE_MAP_FLTINT */
  int i,j,k,thisSec,thisRow;
  FILE *fOut;
  int *thisInt;
  int tempInt,minInt,maxInt;
  float thisFloat;
  double thisDiff;

  fOut = lun[*luNum];

  thisInt = (int *)malloc(sizeof(int)*(*numCols+2));

  fseek(fOut,(*mapOffset),0);

  /* minInt = maxInt = (int)float_buffer[0]; */
  for(i=0;i<*numSecs;i++)
    {/* secs */
      thisSec = i * (*numRows) * (*numCols);
      for(j=0;j<*numRows;j++)
	{/* rows */
	  /* It's already in TSB format  */
	  thisRow = (j) * (*numCols);

	  for(k=0;k<*numCols;k++)
	    {/* Do all */
	      thisFloat = (float)(float_buffer[(thisSec + 
						thisRow + k)]*
				  (*floatScaleFactor));
	      /* Mimic the fortran nint() rtn */
	      tempInt = (int)thisFloat;
	      thisDiff = (double)((double)thisFloat - (double)(tempInt));
	      if(thisDiff >= 0.5)
		{/* rounnd it */
		  thisInt[k] = (int)tempInt + 1;
		}/* rounnd it */
	      else if(thisDiff <= -0.5)
		{/* rounnd it */
		  thisInt[k] = (int)tempInt - 1;
		}/* rounnd it */
	      else thisInt[k] = (int)tempInt;
	      if(thisFloat > 2147483647.0)
		{/* Overflow */
		  printf("[PIF_WRITE_MAP_FLTINT()] Overflow of int.[%f][%f]\n",
			 float_buffer[(thisSec + thisRow + k)],
			 *floatScaleFactor);
		  exit(0);
		}/* Overflow */
	      /*minInt = MIN(minInt,thisInt[k]);
	      maxInt = MAX(maxInt,thisInt[k]); */
	      differentEndian_((BYTE *)&thisInt[k], 4, endians[*luNum]);
	    }/* Do all */

	  OKREAD(fwrite(thisInt,sizeof(int),(*numCols),fOut),
		 *numCols,
		 "PIF_WRITE_MAP_FLTINT");
	}/* rows */
    }/* secs */

  free(thisInt);

}/* PIF_WRITE_MAP_FLTINT */



void pif_write_sf(forPifHead_t *sfHead,
		   strucFact_t *sfs)
{/* PIF_WRITE_SF */
  pif_write_sf_(sfHead, sfs);
}/* PIF_WRITE_SF */
/**************************************************************
 * pif_write_sf() - Write the Structure factor array to disk.
*/
void pif_write_sf_(forPifHead_t *sfHead,
		   strucFact_t *sfs)
{/* PIF_WRITE_SF */
  int i,j,k, imgOffset, maxP,localInt;
  FILE *fOut;
  float minF,maxF;

  fOut = lun[sfHead->luNum];

  fseek(fOut,1024,0);  /* SF files always begin at 1024 */

  maxF= minF = sfs[0].a_amp;
  for(i=0;i<sfHead->numRows;i++)
    {/* Write all */

      OKREAD(fwrite(&(sfs[i].h),sizeof(short),1,fOut),1,"PIF_WRITE_SF");
      fwrite(&(sfs[i].k),sizeof(short),1,fOut);
      fwrite(&(sfs[i].l),sizeof(short),1,fOut);
 
      convertFloat_(&localInt,sfs[i].a_amp,sfHead->floatScaleFactor);
      differentEndian_((BYTE *)&localInt, 4, endians[sfHead->luNum]);
      fwrite(&localInt,sizeof(int),1,fOut);             

      convertFloat_(&localInt,sfs[i].b_phi,sfHead->floatScaleFactor);
      differentEndian_((BYTE *)&localInt, 4, endians[sfHead->luNum]);
      fwrite(&localInt,sizeof(int),1,fOut);             

      convertFloat_(&localInt,sfs[i].sig,sfHead->floatScaleFactor);
      differentEndian_((BYTE *)&localInt, 4, endians[sfHead->luNum]);
      fwrite(&localInt,sizeof(int),1,fOut);             

      convertFloat_(&localInt,sfs[i].dstr,sfHead->floatScaleFactor);
      differentEndian_((BYTE *)&localInt, 4, endians[sfHead->luNum]);
      fwrite(&localInt,sizeof(int),1,fOut);             

      minF = MIN(sfs[i].a_amp,minF);
      minF = MIN(sfs[i].b_phi,minF);

      maxF = MAX(sfs[i].a_amp,maxF);
      maxF = MAX(sfs[i].b_phi,maxF);

    }/* Write all */

  fseek(fOut,(512+92),0);  /* Seek the data min/max fields */
  convertFloat_(&localInt,minF,sfHead->floatScaleFactor);
  differentEndian_((BYTE *)&localInt, 4, endians[sfHead->luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);             
  convertFloat_(&localInt,maxF,sfHead->floatScaleFactor);
  differentEndian_((BYTE *)&localInt, 4, endians[sfHead->luNum]);
  fwrite(&localInt,sizeof(int),1,fOut);


}/* PIF_WRITE_SF */



/* *************************************** *
 *  floatInt*4 Packed boxes.
 * *************************************** */

void pif_write_boxi4(forPifHead_t *boxHead,
		      float *float_buffer)
{
  pif_write_boxi4_(boxHead,float_buffer);
}
void pif_read_boxi4(forPifHead_t *boxHead,
		     float *float_buffer) 
{
  pif_read_boxi4_(boxHead,float_buffer);
}

/****************************************************************
 * pif_write_boxi4() - Write packed box in I4 floatInt format.
*/
void pif_write_boxi4_(forPifHead_t *boxHead,
		      float *float_buffer)
{/* PIF_WRITE_BOXI4 */
  int boxOffset;
  int luNum,numCols, numRows, numSecs,imgNum;
  float floatScaleFactor;


  luNum   = boxHead->luNum;
  numCols = boxHead->numCols;
  numRows = boxHead->numRows;
  numSecs = 1;
  imgNum  = boxHead->imgNum;
  floatScaleFactor  = boxHead->floatScaleFactor;
  
  /* Times 4 because we're INT here */
  boxOffset = 512 + (imgNum * 512) + 
    ((imgNum-1)*(numCols)*(numRows)*4); 
  
  pif_write_map_fltint_(&luNum,
			float_buffer,
			&boxOffset,
			&numCols,
			&numRows,
			&numSecs,
			&floatScaleFactor);

}/* PIF_WRITE_BOXI4 */




/****************************************************************
 * OLD_pif_read_boxi4() - Read packed box in I4 floatInt format.
*/
void OLD_pif_read_boxi4_(forPifHead_t *boxHead,
		     float *float_buffer) 
{/* OLD_PIF_READ_BOXI4 */
  int boxOffset,i,j;
  int luNum,numCols, numRows, numSecs,imgNum;
  float recipFloatScaleFactor;
  int localInt;
  int *rowInts;
  float localFloat;
  FILE *fIn;

  luNum   = boxHead->luNum;
  numCols = boxHead->numCols;
  numRows = boxHead->numRows;
  numSecs = 1;
  imgNum  = boxHead->imgNum;
  recipFloatScaleFactor  = boxHead->recipFloatScaleFactor;
  
/* OLD_PIF_READ_BOXI4 */
  /* Times 4 because we're INT here */
  boxOffset = 512 + (imgNum * 512) + 
    ((imgNum-1)*(numCols)*(numRows)*4); 

  rowInts = (int *)malloc(sizeof(int)*numCols);

  fIn = lun[luNum];

  fseek(fIn,boxOffset,0);

/* OLD_PIF_READ_BOXI4 */
  for(i=0;i<numRows;i++)
    {/* rows */

	fread(&localInt,sizeof(int),numCols,fIn); /*ROK*/
      for(j=0;j<numCols;j++)
	{/* cols */
	  
	  localInt = rowInts[i];
	  differentEndian_((BYTE *)&localInt, 4, endians[luNum]);
	  convertBackFloat_(localInt,&localFloat,
			    recipFloatScaleFactor);
	  float_buffer[((i*numCols)+j)]= localFloat;

	}/* cols */
/* OLD_PIF_READ_BOXI4 */

    }/* rows */

  free(rowInts);

/* OLD_PIF_READ_BOXI4 */
}/* OLD_PIF_READ_BOXI4 */


/********************************************************************
 * pif_read_boxi4() - Gotten from Misha Sherman Jan 2003
 */
void pif_read_boxi4_(forPifHead_t *boxHead,
		     float *float_buffer) 
{/* PIF_READ_BOXI4 */
  int boxOffset,i,j;
  int luNum,numCols, numRows, numSecs,imgNum;
  float recipFloatScaleFactor;
  int localInt;
  int *rowInts;
  float localFloat;
  FILE *fIn;

  luNum   = boxHead->luNum;
  numCols = boxHead->numCols;
  numRows = boxHead->numRows;
  numSecs = 1;
  imgNum  = boxHead->imgNum;
  recipFloatScaleFactor  = boxHead->recipFloatScaleFactor;
  
  /* Times 4 because we're INT here */
  boxOffset = 512 + (imgNum * 512) + 
    ((imgNum-1)*(numCols)*(numRows)*4); 

  rowInts = (int *)malloc(sizeof(int)*numCols);

  fIn = lun[luNum];

  fseek(fIn,boxOffset,0);

  for(i=0;i<numRows;i++)
    {/* rows */

	fread(rowInts,sizeof(int),numCols,fIn); /*ROK*/
      for(j=0;j<numCols;j++)
	{/* cols */
	  
	  localInt = rowInts[j];
	  differentEndian_((BYTE *)&localInt, 4, endians[luNum]);
	  convertBackFloat_(localInt,&localFloat,
			    recipFloatScaleFactor);
	  float_buffer[((i*numCols)+j)]= localFloat;

	}/* cols */

    }/* rows */

  free(rowInts);

}/* PIF_READ_BOXI4 */






/* *************************************** *
 *  Debugging routines
 * *************************************** */

void debug_fft_buffer(float *fft_buffer, int *idim1, int *idim2, int *bufNum)
{
  debug_fft_buffer_(fft_buffer, idim1, idim2, bufNum);
}

/****************************************************************
 * debug_fft_buffer() - Write fft buffer for reading by robem.
*/
void debug_fft_buffer_(float *fft_buffer, int *idim1, int *idim2, int *bufNum)
{/* DEBUG_FFT_BUFFER */
  char outFile[256];
  FILE *fOut;

  sprintf(outFile,"fft_buffer_%03d.dat",*bufNum);
  fOut = fopen(outFile,"w");
  if(fOut)
    {/* Write it */
      fwrite(fft_buffer,sizeof(float),(((*idim1)*2)*(*idim2)),fOut);
      fclose(fOut);
    }/* Write it */
  else
    printf("DEBUG_FFT_BUFFER - Can't open [%s]\n",outFile);

}/* DEBUG_FFT_BUFFER */

void debug_fft_buffer_read(float *fft_buffer, 
			   int *idim1, int *idim2, int *bufNum)
{
  debug_fft_buffer_read_(fft_buffer, idim1, idim2, bufNum);
}

/****************************************************************
 * debug_fft_buffer_read() - Read fft buffer file.
*/
void debug_fft_buffer_read_(float *fft_buffer, 
			    int *idim1, int *idim2, int *bufNum)
{/* DEBUG_FFT_BUFFER_READ */
  char outFile[256];
  FILE *fOut;

  sprintf(outFile,"fft_buffer_%03d.dat",*bufNum);
  fOut = fopen(outFile,"r");
  if(fOut)
    {/* Write it */
	fread(fft_buffer,sizeof(float),(((*idim1)*2)*(*idim2)),fOut);/*ROK*/
      fclose(fOut);
    }/* Write it */
  else
    printf("DEBUG_FFT_BUFFER_READ - Can't open [%s]\n",outFile);

}/* DEBUG_FFT_BUFFER_READ */


/*RSS Changed return type to void */
void pif_copy_head(forPifHead_t *fromHead,forPifHead_t *toHead)
{/* PIF_COPY_HEAD */
  pif_copy_head_(fromHead,toHead);
}/* PIF_COPY_HEAD */

/***************************************************************
 * pif_copy_head() - Copy a PIF file's header.
*/
/*RSS Changed return type to void */
void pif_copy_head_(forPifHead_t *fromHead,forPifHead_t *toHead)
{/* PIF_COPY_HEAD */

  int i;

  toHead->luNum       = fromHead->luNum;
  toHead->imgNum      = fromHead->imgNum;
  toHead->nimg        = fromHead->nimg;
  toHead->floatScaleFactor = fromHead->floatScaleFactor;
  toHead->recipFloatScaleFactor = fromHead->recipFloatScaleFactor;
  toHead->img_bkg     = fromHead->img_bkg;
  toHead->datai4_min  = fromHead->datai4_min;
  toHead->datai4_max  = fromHead->datai4_max;
  toHead->datar4_avg  = fromHead->datar4_avg;
  toHead->datar4_std  = fromHead->datar4_std;
  toHead->numCols     = fromHead->numCols;
  toHead->numRows     = fromHead->numRows;
  toHead->numSecs     = fromHead->numSecs;
  toHead->imgOffset   = fromHead->imgOffset;
  toHead->mode        = fromHead->mode;
  toHead->htype       = fromHead->htype;
  toHead->aoverb      = fromHead->aoverb;
  toHead->map_abang   = fromHead->map_abang;
  toHead->a_cell      = fromHead->a_cell;
  toHead->b_cell      = fromHead->b_cell;
  toHead->c_cell      = fromHead->c_cell;	 
  toHead->alp         = fromHead->alp;	 
  toHead->bet         = fromHead->bet;	 
  toHead->gam         = fromHead->gam;	 
  toHead->dela        = fromHead->dela;	 
  toHead->delb        = fromHead->delb;	 
  toHead->delc        = fromHead->delc;	 
  toHead->a_star      = fromHead->a_star;	 
  toHead->b_star      = fromHead->b_star;	 
  toHead->c_star      = fromHead->c_star;	 
  toHead->alp_star    = fromHead->alp_star;	 
  toHead->bet_star    = fromHead->bet_star;	 
  toHead->gam_star    = fromHead->gam_star;	 
  toHead->pixelSize   = fromHead->pixelSize; 
  toHead->t_matrix[0] = fromHead->t_matrix[0];
  toHead->t_matrix[1] = fromHead->t_matrix[1];
  toHead->t_matrix[2] = fromHead->t_matrix[2];
  toHead->t_matrix[3] = fromHead->t_matrix[3];
  toHead->t_matrix[4] = fromHead->t_matrix[4];
  toHead->t_matrix[5] = fromHead->t_matrix[5];
  toHead->binFactor   = fromHead->binFactor; 
  toHead->dthe        = fromHead->dthe;	 
  toHead->dphi_90     = fromHead->dphi_90;	 
  toHead->symmetry    = fromHead->symmetry;	 
  toHead->pixSizUnits = fromHead->pixSizUnits;
  toHead->res_out     = fromHead->res_out;	 
  toHead->ctfMode     = fromHead->ctfMode;	 
  toHead->tempFactor  = fromHead->tempFactor;
  toHead->wiener      = fromHead->wiener;	 
  toHead->packRadius  = fromHead->packRadius;
  toHead->borderType  = fromHead->borderType;
  toHead->borderWidth = fromHead->borderWidth;
  for(i=0;i<80;i++)
    {
      toHead->descriptionStr[i] = fromHead->descriptionStr[i];
      toHead->timeStr[i]        = fromHead->timeStr[i];       /*RSS*/
      toHead->micrographStr[i]  = fromHead->micrographStr[i];
      toHead->scanNumberStr[i]  = fromHead->scanNumberStr[i];
    }
  toHead->rad_lo = fromHead->rad_lo;
  toHead->rad_hi = fromHead->rad_hi;
  toHead->pifVersion = fromHead->pifVersion;
  toHead->even = fromHead->even;
  toHead->mrcX = fromHead->mrcX;
  toHead->mrcY = fromHead->mrcY;
  toHead->mrcZ = fromHead->mrcZ;
  toHead->transX = fromHead->transX;
  toHead->transY = fromHead->transY;
  toHead->transZ = fromHead->transZ;
  toHead->isTrunc = fromHead->isTrunc;
  toHead->origX = fromHead->origX;
  toHead->origY = fromHead->origY;
  toHead->origZ = fromHead->origZ;


}/* PIF_COPY_HEAD */


void chk_file_exists_(char *fname)
{
  chk_file_exists(fname);
}
/*****************************************************************
 * chk_file_exists() -
 */
void chk_file_exists(char *fname)
{/* CHK_FILE_EXISTS */
  FILE *fIn;
  char inChar,sysCall[1024];
  int done,totBytes;

  fIn = fopen(fname,"r");
  if(fIn)
    {/* there */
      printf("\n\nFound file [%s]\n",fname);
      done = FALSE;
      totBytes = 0;
      while(!feof(fIn) && !done)
	{/* Read it all */
	    fread(&inChar,sizeof(BYTE),1,fIn);/*ROK*/
	  totBytes++;
	  if(totBytes > 10)
	    done = TRUE;
	}/* Read it all */
      fclose(fIn);
      if(totBytes > 8)
	{/* WE've got to keep */
	  printf("\n\n\
***********************************************************\n\
You've already got a file called [%s] with data in it.\n\
Aborting program ...\n\
***********************************************************\n\n",
		 fname);
	  exit(0);
	}/* WE've got to keep */
      else
	{/* Blow it away */
	  printf("\n\n\
***********************************************************\n\
You've got a file called [%s] with minimal data in it.\n\
I'm blowing that file away ...\n\
***********************************************************\n\n",
		 fname);
	  sprintf(sysCall,"rm %s",fname);
	  system(sysCall);
	}/* Blow it away */
      
    }/* there */


}/* CHK_FILE_EXISTS */






/***************************************************************
 * pif_write_map_section4_() - Write the PIF file's FLOAT image 
 * for this section.  NOTE:  This rtn assumes there is only 
 * 1 map image in the entire PIF file and this one's it.  Also
 * assumes you're writing in int*2 on disk. 
*/
void pif_write_map_section4_(int *luNum,
                            float *float_buffer,
                            int *secNum,
                            int *numCols,
                            int *numRows,
                            float *floatScaleFactor)
{/* PIF_WRITE_MAP_SECTION */
  int mapOffset, numSecs, thisSecNum;

  thisSecNum = *secNum - 1;  /* Fortran numbers from 1 not 0 */
  numSecs = 1;
  /* Times 2 because they're shorts here */
  mapOffset = 1024 + (4 * thisSecNum * (*numRows) * (*numCols));

  pif_write_map_fltint_(luNum,
                        float_buffer,
                        &mapOffset,
                        numCols,
                        numRows,
                        &numSecs,
                        floatScaleFactor);

}/* PIF_WRITE_MAP_SECTION4_ */






/* End of libPIF.c */


